/* Copyright: 1997 by
 * 
 *            Rolf Rabenseifner
 *            Computing Center University of Stuttgart
 *            rabenseifner@rus.uni-stuttgart.de 
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 *
 * Design:         Rolf Rabenseifner on the base of lat_bnd.c
 * Implementation: Rolf Rabenseifner,
 *                 Christoph Grunwald  *cg*
 */ 

/*

 SX:
   compile:
   mpicc -o mpi_bench5 mpi_bench5.c 

     additional compile time options: -DSXGMEM 
                                      all buffers are located in global memory 

   start program:
   mpirun -host host1 -np #NumPEonHost1 {-host host2 -np #NumPEonHost2} ./mpi_bench5

 Paragon:
   compile:
   cc -nx -o mpi_bench5 mpi_bench5.c -I/usr/local/mpi/include -L/usr/local/mpi/lib/paragon/nx -lmpi

   start program:
   isub -sz 2 ./mpi_bench5

 CRAY T3E:
   environment:
   rus00598 hwwt3e 157$ module list
   Currently Loaded Modulefiles:
        1) modules        3) craylibs       5) cf90           7) cam
        2) nqe.3.1.0.1    4) craytools      6) CC             8) PrgEnv
        9) mpt.1.1.0.0

   compile:
   cc -o mpi_bench5 mpi_bench5.c

     additional compile time options: -DSHMEM_GET or -DSHMEM_PUT 
                                      substitutes MPI_Sendrecv by shmem_put or shmem_get
                                      in all MPI_Sendrecv test cases 

   start program:
   mpirun -np 2 ./mpi_bench5


 Hitachi (operating system HI-UX/MPP):

   initialization at RUS:
     USE_MPI=1  in .profile

   compile:
     cc  -I/usr/local/mpi/include  -c mpi_bench5.c
     cc  -o mpi_bench5 mpi_bench5.o -L/usr/local/mpi/lib -lmpi   -lm

   start program:
     mpirun -np 2 ./mpi_bench5


 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
/* for sqrt: */
#include <math.h> 

/*rolf for sysconf*/
#include <unistd.h> 

#if defined(SHMEM_PUT) || defined(SHMEM_GET)
#include <mpp/shmem.h>
#define MPI_Sendrecv MPISHMEM_Sendrecv 
int MPISHMEM_Sendrecv(void *sendbuf, int sendcount, MPI_Datatype sendtype,
    int dest, int sendtag, void *recvbuf, int recvcount,
    MPI_Datatype recvtype, int source, int recvtag, MPI_Comm comm,
    MPI_Status *status);
#endif

#if defined(MPI_INFO_NULL) && defined(MPI_WIN_NULL)
  /* # define WITH_ONESIDED 
     (not yet part of the libmpi.a in CRAYs mpt/1.2.0.1)*/
#endif 

/* NEC HB   
   to enable Global Shared Memoy usage, which speeds up IXS transfers
   on multinode SX systems
*/
#if defined(SXGMEM)
 #define malloc(x) gmalloc(x)
 #define free(x)   gfree(x)

 void *gmalloc(size_t size) 
 {
   char * ptr;
   MPI_Alloc_mem(size,MPI_INFO_NULL,&ptr);
   return ptr;
 }

 void gfree(void *ptr); 
 {
   MPI_Free_mem(ptr); 
 } 
#endif



#define   MAX_PROCESSES  2048 
 /*  ___________________________________________________  */
   static void band_s ( int partner );
   static void band_r ( int partner );
   static void coll   ( );
 /*  ___________________________________________________  */
static void  print_status (void)
{
   time_t   zeit;

   zeit = time (NULL);

   printf ( "%s", ctime (&zeit) );
   system ( "uname -a" );
   printf ( "\n" );
}

 /*  ___________________________________________________  */
static double time_base; 
static int use_uncorrected_clocks; 
static int band_tests;  /* # measurement for bandwith averaging */ 
static int num_tests;   /* # tests for one measurement */ 
static int max_len;     /* maximum buffer length */ 
static unsigned char *Send_Buf;
static unsigned char *Recv_Buf;
static double        *SEND_Buf;
static double        *RECV_Buf;
typedef struct {double v; int i;} double_int; 
static double_int    *SEND_LOC_Buf; 
static double_int    *RECV_LOC_Buf; 
static int     Numnode;
static int     Mynode;
static int     test_case;
static char   *operation1_str, *operation2_str; 
static char    empty1_str[160], empty2_str[80]; 
static int     send_method, recv_method, rpc_or_crossed;
static int     client_rank, server_rank;
static int     coll_operation, group_method;
static int     root_in_comm_world, root_in_coll_comm; 
static MPI_Comm coll_comm; 
#ifdef WITH_ONESIDED
static MPI_Win  coll_win; 
#endif 
static int     *is_in_coll_comm; 
static int     coll_size, coll_myrank;
static double  sum_nodes1; 
static int     min_lng, max_lng, step_method, step_lng;
static double  expected_latency, expected_bandwidth;
static int     CPUs_per_Node, num_vert_rings, stride_vert_rings; 
static MPI_Request array_of_requests[MAX_PROCESSES]; 
static MPI_Status  array_of_statuses[MAX_PROCESSES]; 
static MPI_Status  local_array_of_statuses[MAX_PROCESSES]; 
static int         array_of_indices[MAX_PROCESSES];
 /*  ___________________________________________________  */
 
int MYallgather(void* s_buf, int s_cnt, MPI_Datatype s_type, 
            void* r_buf, int r_cnt, MPI_Datatype r_type, MPI_Comm comm);
int CHallgather(void* s_buf, int s_cnt, MPI_Datatype s_type, 
            void* r_buf, int r_cnt, MPI_Datatype r_type, MPI_Comm comm);
 /*  ___________________________________________________  */
 
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
int MPI_MYallreduce(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, 
                 MPI_Op op, MPI_Comm comm);
int MPI_MYreduce(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, 
                 MPI_Op op, int root, MPI_Comm comm);
#endif  
 
 /*  ___________________________________________________  */
 
static void print_description ( void )
{
  char *text1[] = {
"This program tries to measure the bandwidth and latency of this system",
"using the MPI library. The program needs to run on at least two nodes.",
"All timings are in milliseconds, bandwidth in MBytes per second",
#ifdef NORM_MB 
 "with 1 MBytes/s = 1,000,000 Bytes/s.",
#else
 "with 1 MBytes/s = 1,048,576 Bytes/s = 2**20 Bytes/s.",
#endif 
  NULL };

  char *text2[] = {
   "\007", /*BEL*/ 
   NULL           };

   int x;

   for ( x = 0; text1[x] != NULL; x++ )
     printf ( "  %s\n" , text1[x] );
   for ( x = 0; text2[x] != NULL; x++ )
     printf ( "  %s\n" , text2[x] );
}

static void time_sync(int use_uncorrected_clocks)
{
  int i, flag; 
  void* wtime_is_global; 
 
  MPI_Attr_get( MPI_COMM_WORLD, MPI_WTIME_IS_GLOBAL, &wtime_is_global, &flag); 
  if ((((int)wtime_is_global) == 1) || (use_uncorrected_clocks == 1))
  { if (Mynode==0) printf("\nThis system has a global time\n"); 
    time_base = 0; 
  } 
  else
  { double t, tsum, tmin, tmax;
    if (Mynode==0) printf("\nThis system has only local clocks, computing a global base...\n");
    tsum = 0; 
    for (i=0; i<10; i++)
    { MPI_Barrier(MPI_COMM_WORLD);
      t = MPI_Wtime(); tsum = tsum + t; 
    } 
    time_base = tsum / 10; 
    MPI_Reduce( &time_base, &tmin, 1, MPI_DOUBLE, MPI_MIN,
                0, MPI_COMM_WORLD);
    MPI_Reduce( &time_base, &tmax, 1, MPI_DOUBLE, MPI_MAX,
                0, MPI_COMM_WORLD);
    if (Mynode==0) printf("\ntime base maximum - minimum = %9.6lf ms\n",
                          1000 * (tmax-tmin));
  } 
}


#ifdef MPX_SUCCESS
/*rolf*/void dump_global(int level); void dump_local(int level); void dump_listener(int level); 
#endif 

int main ( int argc, char *argv[] )
{
  size_t size; int i;
  char nl; 

  MPI_Init ( &argc, &argv );
#ifdef MPX_SUCCESS
/*rolf   dump_global(0); dump_local(0); dump_listener(0); */
#endif 
  MPI_Comm_size ( MPI_COMM_WORLD, &Numnode );
  MPI_Comm_rank ( MPI_COMM_WORLD, &Mynode );

  if ( Mynode == 0 )
    {
      print_status ();
      print_description ();
    }

  time_sync(0); 

#ifdef SHMEM_PUT
  printf("\n\n-DSHMEM_PUT --> all MPI_Sendrecv are substituted by shmem_put+shmem_quiet");
  printf(  "\n                (only MPI_COMM_WORLD can be used)\n");
#endif 
#ifdef SHMEM_GET
  printf("\n\n-DSHMEM_GET --> all MPI_Sendrecv are substituted by shmem_get");
  printf(  "\n                (only MPI_COMM_WORLD can be used; MPI_ANY_SOURCE is not allowed)\n");
#endif 

  if ( Mynode == 0 )
    {
      printf("\nMaximum buffer length in bytes (e.g. %1d, %1d, %1d, %1d, %1d)? ",
          512*1024, 1*1024*1024, 2*1024*1024, 4*1024*1024, 8*1024*1024);
      scanf("%d",&max_len);
      printf("\nmax buffer length = %1d\n", max_len);
    }
  MPI_Bcast ( &max_len, 1, MPI_INT, 0, MPI_COMM_WORLD);
 
/*rolf printf("OPEN_MAX on node %3d = %4d\n",Mynode,sysconf(_SC_OPEN_MAX)); */

  /* two times for test_case==2 && method==4 */
  MPI_Buffer_attach( malloc(2*(MPI_BSEND_OVERHEAD+max_len)),
                            2*(MPI_BSEND_OVERHEAD+max_len) );

 
  size = max_len;
#if defined(SHMEM_PUT) || defined(SHMEM_GET)
  if ( ( Send_Buf = shmalloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
  if ( ( Recv_Buf = shmalloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
  if ( ( SEND_Buf = shmalloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
  if ( ( SEND_LOC_Buf = shmalloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
#else
  if ( ( Send_Buf = malloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
# ifdef WITH_ONESIDED
   if ( MPI_Alloc_mem(size, MPI_INFO_NULL, &Recv_Buf) ) 
    { fprintf ( stderr, " can not alloc recv window memory\n" ); exit ( 1 ); }
# else 
   if ( ( Recv_Buf = malloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
# endif 
  if ( ( SEND_Buf = malloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
  if ( ( SEND_LOC_Buf = malloc ( size ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); }
#endif
  RECV_Buf = (double *) Recv_Buf;
  RECV_LOC_Buf = (double_int *) Recv_Buf;
  if ( ( is_in_coll_comm = malloc ( sizeof(int)*Numnode ) ) == NULL )
    { fprintf ( stderr, " can not alloc memory \n" ); exit ( 1 ); } 

  for ( i=0; i < max_len; i++) Send_Buf[i]=(i*(Mynode+1)) % 256;
  for ( i=0; i < max_len/sizeof(double); i++) SEND_Buf[i]=i*(Mynode+1);
  srand48( (long)Mynode ); 
  for ( i=0; i < max_len/sizeof(double_int); i++) 
   { SEND_LOC_Buf[i].v = -(1+drand48()); SEND_LOC_Buf[i].i = Mynode;
     if ((i % Numnode) == Mynode) SEND_LOC_Buf[i].v = Mynode;
   }

 
  do{

    operation1_str = empty1_str; operation2_str = empty2_str; 
    if ( Mynode == 0 )
    {
      printf("\nWhich test case (0=quit, 1=point-to-point, 2=collective, 3=internal_dump,\n");
      printf("\n                 4=time_sync, 5=getcwd)\n? ");
      scanf("%d%c",&test_case,&nl);
      if (test_case==1 || test_case==2) 
      { if (nl=='\n') printf("\nHow many measurements should be done for the averaging (e.g. 8)? ");
        scanf("%d%c",&band_tests,&nl);
        if (nl=='\n') printf("\nLoop length for one measurement (e.g. 3)? ");
        scanf("%d%c",&num_tests,&nl);
      } else {band_tests=0; num_tests=0;} 
    }

    MPI_Bcast ( &band_tests, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast ( &num_tests,  1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast ( &test_case, 1, MPI_INT, 0, MPI_COMM_WORLD);

    switch (test_case) {

      case 0: /* quit */  break;

      case 1: /* point-to-point benchmarks */

        if ( Mynode == 0 )
        {
          if (nl=='\n') printf("\nWhich send method (1=Send,  2=Bsend,  3=Ssend,");
          if (nl=='\n') printf("\n                   4=Isend+Wait, 5=Ibsend+Wait, 6=Issend+Wait ");
          if (nl=='\n') printf("\n                   7=Isend+Request_free "); /*cg*/
          if (nl=='\n') printf("\n                   8=Send_init+Start+Wait+Request_free"); /*cg*/
          if (nl=='\n') printf("\n                   9=Bsend_init+Start+Wait+Request_free"); /*cg*/
          if (nl=='\n') printf("\n                  10=Ssend_init+Start+Wait+Request_free"); /*cg*/
          if (nl=='\n') printf("\n                  11=Rsend_init+Start+Wait+Request_free"); /*cg*/
          if (nl=='\n') printf("\n                  12=Send_init+Startall+Wait+Request_free"); /*cg*/
          if (nl=='\n') printf(")? ");
          scanf("%d%c",&send_method,&nl);
          switch (send_method)
           { 
             case 1: operation1_str = "MPI_Send"; break; 
             case 2: operation1_str = "MPI_Bsend"; break; 
             case 3: operation1_str = "MPI_Ssend"; break; 
             case 4: operation1_str = "MPI_Isend + MPI_Wait"; break; 
             case 5: operation1_str = "MPI_Ibsend + MPI_Wait"; break; 
             case 6: operation1_str = "MPI_Issend + MPI_Wait"; break; 
 
             case 7: operation1_str = "MPI_Isend + MPI_Request_free"; break;
             case 8: operation1_str = "MPI_Send_init+Start+Wait+Request_free"; break;
             case 9: operation1_str = "MPI_Bsend_init+Start+Wait+Request_free"; break;
             case 10: operation1_str = "MPI_Ssend_init+Start+Wait+Request_free"; break;
             case 11: operation1_str = "MPI_Rsend_init+Start+Wait+Request_free"; break;   
             case 12: operation1_str = "MPI_Send_init+Startall+Wait+Request_free"; break;
            } 
          if (nl=='\n') printf("\nWhich recv method (1=Recv,  2=Irecv, 3=Irecv before send, 4=Probe+Recv ");
          if (nl=='\n') printf("\n                   5=Recv_init+Start+Wait+Request_free,");
          if (nl=='\n') printf("\n                   6=Recv(MPI_ANY_SOURCE)) ? ");
          scanf("%d%c",&recv_method,&nl);
          switch (recv_method) { 
           case 1: operation2_str = "/  MPI_Recv"; break; 
           case 2: operation2_str = "/  MPI_Irecv"; break; 
           case 3: operation2_str = "/  MPI_Irecv before send"; break; 
           case 4: operation2_str = "/  MPI_Probe + MPI_Recv"; break; 
           case 5: operation2_str = "/  MPI_Recv_init+Start+Wait+Request_free"; break; 
           case 6: operation2_str = "/  MPI_Recv(MPI_ANY_SOURCE)"; break; 
          } 
          if (nl=='\n') printf("\nRPC or crossed? (1=RPC ping-pong, 2=client and server sends in parallel,");
          if (nl=='\n') printf("\n                 i.e. deadlock possible with Send, guaranteed with Ssend!)? ");
          scanf("%d%c",&rpc_or_crossed,&nl);
          if (nl=='\n') printf("\nWhich should be the rank of the client node (0-%1d)? ",Numnode-1);
          scanf("%d%c",&client_rank,&nl);
          if (nl=='\n') printf("\nWhich should be the rank of the server node (");
          if (client_rank>0)
          { if (nl=='\n') printf("0");
            if (client_rank>1) if (nl=='\n') printf("-%1d",client_rank-1);
          }
          if (client_rank>0 && client_rank<(Numnode-1)) if (nl=='\n') printf(", ");
          if (client_rank<(Numnode-1))
          { if (nl=='\n') printf("%1d",client_rank+1);
            if (client_rank<(Numnode-2)) if (nl=='\n') printf("-%1d",Numnode-1);
          }
          if (nl=='\n') printf(")? ");
          scanf("%d%c",&server_rank,&nl);
          if (nl=='\n') printf("\nMinimal transfer length in bytes (0-%1d)? ",max_len);
          scanf("%d%c",&min_lng,&nl);
          if (nl=='\n') printf("\nMaximal transfer length in bytes (%1d-%1d)? ",min_lng,max_len);
          scanf("%d%c",&max_lng,&nl);
          if (nl=='\n')
          { printf("\nStep method (");
            if (min_lng>0) printf("1=multiplication by factor, ");
            printf("2=addition of a steplength)? ");
          } 
          scanf("%d%c",&step_method,&nl);
          if (nl=='\n')
          { if (step_method==1)
                 printf("\nStep-multiplication-factor (>=2)? ");
            else printf("\nStep-addition-length (>=1)? ");
          } 
          scanf("%d%c",&step_lng,&nl);
          if (nl=='\n') printf("Expected latency in usec (e.g.30)? ");
          scanf("%lf%c",&expected_latency,&nl);
          if (nl=='\n') printf("Expected bandwidth in Mbyte/s (e.g.200)? ");
          scanf("%lf%c",&expected_bandwidth,&nl);
        } 

        MPI_Bcast ( &send_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &recv_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &rpc_or_crossed,1,MPI_INT,0, MPI_COMM_WORLD);
        MPI_Bcast ( &client_rank, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &server_rank, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &min_lng    , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &max_lng    , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &step_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &step_lng   , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &expected_latency, 1, MPI_DOUBLE,0,MPI_COMM_WORLD);
        MPI_Bcast ( &expected_bandwidth,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

        if ( Mynode == client_rank )
        {
          printf("\ntest    # loop   send   recv  rpc or client server min_lng max_lng   step   step expected  expected");
          printf("\ncase meas lng. method method crossed   rank   rank                 method length  latency bandwidth");
          printf("\n%4d %4d %4d %6d %6d %7d %6d %6d %7d %7d %6d %6d %8.3lf %9.3lf\n", 
               test_case, band_tests, num_tests,
               send_method, recv_method, rpc_or_crossed, client_rank,
               server_rank, min_lng, max_lng, step_method, step_lng,
               expected_latency, expected_bandwidth);
        }

        MPI_Barrier ( MPI_COMM_WORLD );

        if ( Mynode == client_rank ) band_s ( server_rank );
        if ( Mynode == server_rank ) band_r ( client_rank );

      break /*case 1*/;

      case 2: /* collective benchmarks */
        if ( Mynode == 0 )
        {
          if (nl=='\n')
          { 
          printf("\nWhich operation (0=Barrier, 1=Bcast, 2=Reduce(sum)");
          printf("\n                 3=Bcast+Barrier, 4=Allreduce(sum)");
          printf("\n                 5=Reduce+Bcast,  6=Reduce+Barrier");
          printf("\n                 7=Send+Recv+Waitany 8=...+Waitall");
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
          printf("\n                 9=MYreduce(sum)  10=MYre.+Barrier");
          printf("\n                 11=MYreduce(max) 12=MYre.+Barrier");
#endif 
          printf("\n                 13=Reduce(max)  14=Re.(max)+Barr.");
          printf("\n                 15=NOOP    16=Sendrecv(in a ring)");
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
          printf("\n                 17=MYallreduce(dbl,sum) 18=...max");
#endif 
          printf("\n                 19=Allreduce(double,max) ");
          printf("\n                 20=Send+Recv+Waitsome ");    /*cg*/
          printf("\n                 21=Send+Recv+Testany ");     /*cg*/
          printf("\n                 22=Send+Recv+Testsome ");    /*cg*/
          printf("\n                 23=Sendrecv(only any_source cases)"); /*cg*/
          printf("\n                 24=Sendrecv_replace(in a ring   "); /*cg*/
          printf("\n                 25=Sendrecv_replace(any_source) "); /*cg*/
          printf("\n                 26=Alltoall (sendlng = lng) ");
          printf("\n                 27=Alltoall (sendlng = lng/#PE) ");
          printf("\n                 28=Alltoall (sendlng = lng/#PE^2) ");
          printf("\n                 29=Allgather (sendlng = lng/#PE) ");
          printf("\n                 30=Allgather (sendlng = lng/#PE^2) ");
          printf("\n                 31=Send+Recv+Testall (no progress) ");
          printf("\n                 32=half<->half left/right ISend/Irecv");
          printf("\n                 33=Scatter(cnt=lng) 34=Scatter(cnt=lng/#PE)");
          printf("\n                 35=Gather (cnt=lng) 36=Gather (cnt=lng/#PE)");
          printf("\n                 37=Allgather(c=lng) 29=Allgather(c=lng/#PE)");
          printf("\n                 38=Allgather(c=lng) simulating an optimization");
          printf("\n                 39=MYallgather(c=lng) 40=CHallgather(c=lng)");
#ifdef WITH_ONESIDED
          printf("\n                 41=Win_fence Put(c=lng, alltoall) Win_fence");
#endif
/* 
          if (CPUs_per_Node == 4) { 
            printf("\n    _ _                                     ");
            printf("\n   |    |                                  ");
            printf("\n   | V  0 --- 1--- 2--- 3   ");
            printf("\n   | E  |     |    |    |   "); 
            printf("\n   | R  4 --- 5--- 6--- 7   ");
            printf("\n   | T  |     |    |    |   ");
            printf("\n   | I  8---- 9---10---11   ");
            printf("\n   | C  |     |    |    |   ");
            printf("\n   | A  12---13---14---15   ");
            printf("\n   | L  |     |    |    |   ");
            printf("\n    ----16---17---18---19-- "); 
            printf("\n   |__H_O_R_I_Z_O_N_T_A_L__|");
            printf("\n                                                  ");   
          } else if (CPUs_per_Node == 8) { 
*/ 
            printf("\n ");   
            printf("\n    _ _      e.g. if CPUs_per_Node==8       ");
            printf("\n   |    |                                  ");
            printf("\n   | V  0 --- 1--- 2--- 3--- 4 --- 5 ---6--- 7   ");
            printf("\n   | E  |     |    |    |    |     |    |    |   "); 
            printf("\n   | R  8 --- 9---10---11---12--- 13---14---15   ");
            printf("\n   | T  |     |    |    |    |     |    |    |   ");
            printf("\n   | I  16---17---18---19 ---20---21---22---23   ");
            printf("\n   | C  |     |    |    |    |     |    |    |   ");
            printf("\n   | A  24---25---26---27--- 28---29---30---31   ");
            printf("\n   | L  |     |    |    |    |     |    |    |   ");
            printf("\n    ----32---33---34---35---36---37--- 38---39 -- "); 
            printf("\n   |_ _ _ _ _ _ _H_O_R_I_Z_O_N_T_A_L_ _ _ _ _ _ _|");
            printf("\n                                                  ");   
/* 
          } else if (CPUs_per_Node == 16) { 
            printf("\n    _ _                                     ");
            printf("\n   | V  |                                  ");
            printf("\n   | E  0 - 1- 2- 3- 4- 5- 6- 7- 8- 9-10-11-12-13-14-15 ");
            printf("\n   | R  |   |  |  |  |  |  |  |  |  |  |  |  |  |  |  | "); 
            printf("\n   | T  8 - 9-10-11-12-13-14-15-24-25-26-27-28-29-30-31 ");
            printf("\n   | I  |   |  |  |  |  |  |  |  |  |  |  |  |  |  |  | "); 
            printf("\n   | C  32-33-34-35-36-37-38-39-40-41-42-43-44-45-46-47 ");
            printf("\n   | A  |   |  |  |  |  |  |  |  |  |  |  |  |  |  |  | "); 
            printf("\n   | L  48-49-50-51-52-53-54-55-56-57-58-59-60-61-62-63 "); 
            printf("\n   |_ _ _ _ _ _ _ _ _ _H_O_R_I_Z_O_N_T_A_L_ _ _ _ _ _ _ ");
          } 
*/ 
          printf("\n                                                  ");   
/* 
          printf("\n                 42=Sendrecv(one vertical ring out of %1d rings)",CPUs_per_Node);
          printf("\n                 43=Sendrecv(%1d vertical rings)",CPUs_per_Node);
          printf("\n                 44=Sendrecv(%1d vertical rings + horizontal ring inside nodes)",CPUs_per_Node);
*/ 
          printf("\n                 42=Sendrecv(one vertical ring out of CPUs_per_Node rings)");
          printf("\n                 43=Sendrecv(CPUs_per_Node vertical rings)");
          printf("\n                 44=Sendrecv(CPUs_per_Node vertical rings + horizontal ring inside nodes)");
          printf("\n                 45=Sendrecv(horizontal ring inside nodes)");
          printf("\n                 46=Sendrecv(some vertical rings out of CPUs_per_Node rings)");
          printf("\n                 47=Allreduce(double,maxloc)    48=reduce(dbl,maxloc)");
#ifdef WITH_MYREDUCE
          printf("\n                 49=MYallreduce(double,maxloc)  50=MYreduce(dbl,maxloc)");
#endif 

          printf(")? ");
          } 
          scanf("%d%c",&coll_operation,&nl);
          if (nl=='\n') printf("\nMinimal transfer length in bytes (0-%1d)? ",max_len);
          scanf("%d%c",&min_lng,&nl);
          if (nl=='\n') printf("\nMaximal transfer length in bytes (%1d-%1d)? ",min_lng,max_len);
          scanf("%d%c",&max_lng,&nl);
          if (nl=='\n') printf("\nStep method (");
          if (min_lng>0) if (nl=='\n') printf("1=multiplication by factor, ");
          if (nl=='\n') printf("2=addition of a steplength)? ");
          scanf("%d%c",&step_method,&nl);
          if (nl=='\n')
          { if (step_method==1)
                 printf("\nStep-multiplication-factor (>=2)? ");
            else printf("\nStep-addition-length (>=1)? ");
          } 
          scanf("%d%c",&step_lng,&nl);
          if (nl=='\n') printf("Expected latency in usec (e.g.30)? ");
          scanf("%lf%c",&expected_latency,&nl);
          if (nl=='\n') printf("Expected bandwidth in Mbyte/s (e.g.200)? ");
          scanf("%lf%c",&expected_bandwidth,&nl);
          if (nl=='\n') printf("\nWhich node should be the root for coll.op and timing (use COMM_WORLD ranking! 0-%1d)? ",Numnode-1);
          scanf("%d%c",&root_in_comm_world,&nl);
          if (nl=='\n') printf("\nWhich grouping method (0=COMM_WORLD, 1=Comm_dup(COMM_WORLD),");
          if (nl=='\n') printf("\n                       2=Comm_create, 3=Comm_split,");
          if (nl=='\n') printf("\n                       4=Comm_create + Cart_create,"); /*cg*/
          if (nl=='\n') printf("\n                       5=Comm_create + Graph_create,"); /*cg*/
          if (nl=='\n') printf("\n                       6=Comm_create+Cart_create+Cart_sub) ? "); /*cg*/
          if (nl=='\n') printf("\n                       7=Comm_create+Cart_create(SQRT*SQRT)) ? "); /*rab*/
          scanf("%d%c",&group_method,&nl);
          if (group_method >= 2)
          { int i, beg_rank, end_rank; char c;
            for (i=0; i<Numnode; i++) is_in_coll_comm[i] = 0;
            sum_nodes1 = 0; 
            if (nl=='\n') printf("\nGive the ranks (in COMM_WORLD, i.e 0..%1d) of all members (e.g. 0,2-5,13)? ",Numnode-1); 
            do { 
              scanf("%d",&beg_rank); 
              scanf("%c",&c); 
              if (c == '-')
              { scanf("%d",&end_rank);
                scanf("%c",&c); 
              } else {
                end_rank = beg_rank; 
              } 
              for (i=beg_rank; i<=end_rank; i++) 
              { 
                is_in_coll_comm[i] = 1;
                sum_nodes1 = sum_nodes1 + (i+1);
              } 
            } while (c == ','); 
            nl = c;
          }
          else
          { int i;
            for (i=0; i<Numnode; i++) is_in_coll_comm[i] = 1;
            sum_nodes1 = /* sum of all rank_in_comm_world+1 */ 
                         1.0 * Numnode * (Numnode + 1.0) / 2.0;
          } 
          if ((coll_operation >=42) && (coll_operation <= 46)) 
          {
            if (nl=='\n') printf("\nFor Operations 42-46: How many CPUs are on one SMP node? ");
            scanf("%d%c",&CPUs_per_Node,&nl);
          }
          else CPUs_per_Node = 1; 
          if ((coll_operation >=46) && (coll_operation <= 46)) 
          {
            if (nl=='\n') printf("\nFor Operations 46: How many vertical rings out of %1d? ", CPUs_per_Node);
            scanf("%d%c",&num_vert_rings,&nl);
            if (nl=='\n') printf("\nFor Operations 46: Stride between the %1d vertical rings? ", num_vert_rings);
            scanf("%d%c",&stride_vert_rings,&nl);
          }
          else { num_vert_rings=1; stride_vert_rings=1; } 
 
          switch (coll_operation) { 
           case  0: operation1_str = "MPI_Barrier"; break;
           case  1: operation1_str = "MPI_Bcast"; break;
           case  2: operation1_str = "MPI_Reduce(double,sum)"; break;
           case  3: operation1_str = "MPI_Bcast + MPI_Barrier"; break;
           case  4: operation1_str = "MPI_Allreduce(double,sum)"; break;
           case  5: operation1_str = "MPI_Reduce + MPI_Bcast"; break;
           case  6: operation1_str = "MPI_Reduce + MPI_Barrier"; break;
           case  7: operation1_str = "MPI_Send + Recv + Waitany"; break;
           case  8: operation1_str = "MPI_Send + Recv + Waitall"; break;
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
           case  9: operation1_str = "MPI_MYreduce(double,sum)"; break;
           case 10: operation1_str = "MPI_MYreduce(sum)+Barrier"; break;
           case 11: operation1_str = "MPI_MYreduce(double,max)"; break;
           case 12: operation1_str = "MPI_MYreduce(max)+Barrier"; break;
#endif 
           case 13: operation1_str = "MPI_Reduce(double,max)"; break;
           case 14: operation1_str = "MPI_Reduce(max)+Barrier"; break;
           case 15: operation1_str = "NOOP"; break;
           case 16: operation1_str = "Sendrecv in a ring"; break;
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
           case 17: operation1_str = "MPI_MYallreduce(double,sum)"; break;
           case 18: operation1_str = "MPI_MYallreduce(double,max)"; break;
#endif 
           case 19: operation1_str = "MPI_Allreduce(double,max)"; break;
           case 20: operation1_str = "MPI_Send+Recv+MPI_Waitsome"; break;
           case 21: operation1_str = "MPI_Send+Recv+MPI_Testany"; break;
           case 22: operation1_str = "MPI_Send+Recv+MPI_Testsome"; break; 
           case 23: operation1_str = "MPI_Sendrecv with MPI_ANY_SOURCE";
                    break;
           case 24: operation1_str = "MPI_Sendrecv_replace"; break;
           case 25: operation1_str = "MPI_Sendrecv_replace with MPI_ANY_SOURCE"; break; 
           case 26: operation1_str = "MPI_Alltoall (sendlng=lng)"; break; 
           case 27: operation1_str = "MPI_Alltoall (sendlng=lng/#PE)"; break; 
           case 28: operation1_str = "MPI_Alltoall (sendlng=lng/#PE^2)"; break; 
           case 29: operation1_str = "MPI_Allgather (sendlng=lng/#PE)"; break; 
           case 30: operation1_str = "MPI_Allgather (sendlng=lng/#PE^2)"; break;
           case 31: operation1_str = "MPI_Send+Recv+Testall (no progress)"; break;
           case 32: operation1_str = "half<->half left/right ISend/Irecv"; break;
           case 33: operation1_str = "MPI_Scatter (sendlng=lng)"; break; 
           case 34: operation1_str = "MPI_Scatter (sendlng=lng/#PE)"; break; 
           case 35: operation1_str = "MPI_Gather (sendlng=lng)"; break; 
           case 36: operation1_str = "MPI_Gather (sendlng=lng/#PE)"; break; 
           case 37: operation1_str = "MPI_Allgather (sendlng=lng)"; break; 
           case 38: operation1_str = "MPI_Allgather (sendlng=lng) simulation"; break; 
           case 39: operation1_str = "MYallgather (sendlng=lng)"; break; 
           case 40: operation1_str = "CHallgather (sendlng=lng)"; break; 
#ifdef WITH_ONESIDED
           case 41: operation1_str = "Fence Put(c=lng, alltoall) Fence"; break; 
#endif
           case 42: operation1_str = "MPI_Sendrecv (one ring vertical out of CPUs_per_Node)"; break;
           case 43: operation1_str = "MPI_Sendrecv (CPUs_per_Node rings vertical)"; break;
           case 44: operation1_str = "MPI_Sendrecv (CPUs_per_Node rings vertical + ring inside of nodes)"; break;
/* 
           case 42: sprintf(operation1_str, "MPI_Sendrecv (one ring vertical out of %1d)", CPUs_per_Node); break;
           case 43: sprintf(operation1_str, "MPI_Sendrecv (%1d rings vertical)", CPUs_per_Node); break;
           case 44: sprintf(operation1_str, "MPI_Sendrecv (%1d rings vertical + ring inside of nodes)", CPUs_per_Node); break;
*/ 
           case 45: operation1_str = "MPI_Sendrecv (ring inside of nodes)"; break;
           case 46: sprintf(operation1_str, "MPI_Sendrecv (%1d rings vertical with stride %1d out of %1d)", num_vert_rings, stride_vert_rings, CPUs_per_Node); break;
           case 47: operation1_str = "MPI_Allreduce(double,maxloc)"; break;
           case 48: operation1_str = "MPI_Reduce(double,maxloc)"; break;
#ifdef WITH_MYREDUCE
           case 49: operation1_str = "MPI_MYallreduce(double,maxloc)"; break;
           case 50: operation1_str = "MPI_MYreduce(double,maxloc)"; break;
#endif 
          }
          operation2_str = ""; 
        } 
 
        MPI_Bcast ( &coll_operation,     1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &min_lng           , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &max_lng           , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &step_method       , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &step_lng          , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &expected_latency  , 1,MPI_DOUBLE,0,MPI_COMM_WORLD);
        MPI_Bcast ( &expected_bandwidth, 1,MPI_DOUBLE,0,MPI_COMM_WORLD);
        MPI_Bcast ( &root_in_comm_world, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &group_method,       1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( is_in_coll_comm,Numnode,MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &sum_nodes1        , 1,MPI_DOUBLE,0,MPI_COMM_WORLD);
        MPI_Bcast ( &CPUs_per_Node     , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &num_vert_rings    , 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast ( &stride_vert_rings , 1, MPI_INT, 0, MPI_COMM_WORLD);

        if ( Mynode == root_in_comm_world )
        {
          if (nl=='\n') printf("\ntest    # loop coll min_lng max_lng   step   step expected  expected root  group");
          if (group_method >= 2) if (nl=='\n') printf(" group   "); 
          if ((coll_operation >=42) && (coll_operation <= 46)) printf(" CPUs/");
          if ((coll_operation >=46) && (coll_operation <= 46)) printf(" #vert strid");
 
          printf("\ncase meas lng. oper                 method length  latency bandwidth rank method");
          if (group_method >= 2) printf(" elements"); 
          if ((coll_operation >=42) && (coll_operation <= 46)) printf(" Node ");
          if ((coll_operation >=46) && (coll_operation <= 46)) printf(" rings v.ri.");
 
          printf("\n%4d %4d %4d %4d %7d %7d %6d %6d %8.3lf %9.3lf %4d %6d", 
               test_case, band_tests, num_tests, coll_operation,
               min_lng, max_lng, step_method, step_lng,
               expected_latency, expected_bandwidth, 
               root_in_comm_world, group_method);
          if (group_method >= 2)
          { int beg_rank, end_rank; char delim;
            beg_rank = -1; delim = ' ';
            while (beg_rank < Numnode) {
              do beg_rank++; while ((beg_rank < Numnode) && !is_in_coll_comm[beg_rank]);
              if (beg_rank < Numnode)
              { end_rank = beg_rank;
                do end_rank++; while ((end_rank < Numnode) && is_in_coll_comm[end_rank]);
                end_rank--; 
                printf("%c%1d",delim,beg_rank); delim = ',';
                if (end_rank>beg_rank) printf("-%1d",end_rank); 
                beg_rank = end_rank; 
              } 
            } 
          } 
          if ((coll_operation >=42) && (coll_operation <= 46)) printf(" %5d",CPUs_per_Node);
          if ((coll_operation >=46) && (coll_operation <= 46)) printf(" %5d %5d",num_vert_rings, stride_vert_rings);
          printf("\n"); 
        }

        switch(group_method) {
         case 0:
           coll_comm = MPI_COMM_WORLD; 
           root_in_coll_comm = root_in_comm_world; 
          break; 
         case 1:
           MPI_Comm_dup(MPI_COMM_WORLD, &coll_comm); 
           root_in_coll_comm = root_in_comm_world; 
          break; 
         case 2:
         case 4:                                                  /*cg*/
         case 5:                                                  /*cg*/
         case 6:                                                  /*cg*/
         case 7:                                                 /*rab*/
          { int *ranks; MPI_Group world_group, coll_group; int Numcoll, i;
            ranks = (int *)malloc(Numnode*sizeof(int));
            Numcoll = 0;
            for (i=0; i<Numnode; i++) 
              if (is_in_coll_comm[i]) { ranks[Numcoll]=i; Numcoll++; }
            MPI_Comm_group(MPI_COMM_WORLD, &world_group);
            MPI_Group_incl(world_group, Numcoll, ranks, &coll_group);
            free(ranks); 
            MPI_Comm_create(MPI_COMM_WORLD, coll_group, &coll_comm); 
            MPI_Group_free(&coll_group);
            if( ((group_method==4)||(group_method==6)) && (is_in_coll_comm[Mynode]) )
              { MPI_Comm coll_comm_without_topol;
                int coll_size_without_topol;
                int dims[1];
                int periods[1];
                coll_comm_without_topol = coll_comm; /*moving handle into ..._topol*/
                coll_comm = MPI_COMM_NULL;
                periods[0]=0;  
                MPI_Comm_size(coll_comm_without_topol,
                              &coll_size_without_topol);
                dims[0]=coll_size_without_topol;
                MPI_Cart_create(coll_comm_without_topol,1,dims,
                                periods , 1 , &coll_comm);
                MPI_Comm_free(&coll_comm_without_topol);
              }  
            if((group_method==6)&&(is_in_coll_comm[Mynode]))
              { 
                 MPI_Comm comm_scratch;
                 int remain_dims[1];
                 comm_scratch=coll_comm;
                 coll_comm = MPI_COMM_NULL;
                 remain_dims[0]=1;
                 MPI_Cart_sub(comm_scratch,remain_dims,&coll_comm);
                 MPI_Comm_free(&comm_scratch);
             }
            if((group_method==5)&&(is_in_coll_comm[Mynode])) 
             {
                MPI_Comm coll_comm_without_topol;
                int coll_size_without_topol,i;
                int *index, *edges;
                coll_comm_without_topol = coll_comm; /*moving handle into ..._topol*/
                coll_comm = MPI_COMM_NULL;
                MPI_Comm_size(coll_comm_without_topol,
                              &coll_size_without_topol);
                index=(int*) malloc(coll_size_without_topol*sizeof(int));
                edges=(int*)malloc(2*coll_size_without_topol*sizeof(int));  
                for(i=0;i<coll_size_without_topol;i++)
                  index[i]=(i+1)*2;
                 
                for(i=0;i<coll_size_without_topol;i++){
                  edges[i*2]=(i-1+coll_size_without_topol)%coll_size_without_topol;
                  edges[i*2+1]=(i+1)%coll_size_without_topol; 
                }   
                MPI_Graph_create(coll_comm_without_topol,coll_size_without_topol,
                                 index, edges, 1, &coll_comm);
                MPI_Comm_free(&coll_comm_without_topol);
             } 
            if( (group_method==7) && (is_in_coll_comm[Mynode]) ) /*rab*/
              { MPI_Comm coll_comm_without_topol;
                MPI_Group coll_group_without_topol, coll_group; 
                int coll_size_without_topol;
                int dims[2];
                int periods[2];
                int d0,d1,sq,i,cmp; 
                coll_comm_without_topol = coll_comm; /*moving handle into ..._topol*/
                coll_comm = MPI_COMM_NULL;
                periods[0]=0;  periods[1]=0;  
                MPI_Comm_size(coll_comm_without_topol,
                              &coll_size_without_topol);
                sq=(int)(sqrt((double)coll_size_without_topol)+0.01);
                for (d1 = sq;  d1 > 0;  d1--)
                { d0 = coll_size_without_topol / d1;
                  if (d0*d1 == coll_size_without_topol) break;
                } 
                dims[0] = d0; dims[1] = d1;
                MPI_Cart_create(coll_comm_without_topol,2,dims,
                                periods , 1 , &coll_comm);
                MPI_Comm_rank(coll_comm,&coll_myrank); 
                if (coll_myrank==0)
                {
                  MPI_Comm_group(coll_comm, &coll_group);
                  MPI_Comm_group(coll_comm_without_topol, &coll_group_without_topol);
                  MPI_Group_compare(coll_group, coll_group_without_topol, &cmp);
                  printf("cartesian topology %d*%d, %s\n",
                   d0,d1,(cmp==MPI_IDENT ? "same order" : "reordered"));
                  MPI_Group_free(&coll_group); 
                  MPI_Group_free(&coll_group_without_topol); 
                } 
                MPI_Comm_free(&coll_comm_without_topol);
              }  
            if(is_in_coll_comm[Mynode])
             {
               MPI_Comm_group(coll_comm, &coll_group);
               MPI_Group_translate_ranks(world_group,1,&root_in_comm_world,
                                         coll_group, &root_in_coll_comm); 
               MPI_Group_free(&coll_group); 
             }
            MPI_Group_free(&world_group);
          }break;
      
         case 3:
          { 
            MPI_Comm_split(MPI_COMM_WORLD, 
                           (is_in_coll_comm[Mynode] ? 0:MPI_UNDEFINED),
                           Mynode, &coll_comm); 
            if (is_in_coll_comm[Mynode])
            { MPI_Group world_group, coll_group;
              MPI_Comm_group(MPI_COMM_WORLD, &world_group);
              MPI_Comm_group(coll_comm, &coll_group);
              MPI_Group_translate_ranks(world_group,1,&root_in_comm_world,
                                        coll_group, &root_in_coll_comm); 
              MPI_Group_free(&world_group); MPI_Group_free(&coll_group); 
            } 
            else root_in_coll_comm = MPI_UNDEFINED; 
          }break; 

        } /*group_method*/

        if (is_in_coll_comm[Mynode])
        { 
#ifdef WITH_ONESIDED
          MPI_Info info;
#  ifdef WITH_INFO 
          MPI_Info_create(&info);
          MPI_Info_set(info,"no_locks","true"); 
#  else 
          info = MPI_INFO_NULL; 
#  endif 
          MPI_Win_create(RECV_Buf, max_len, sizeof(double), info,
                         coll_comm, &coll_win);
#  ifdef WITH_INFO 
          MPI_Info_free(&info);
#  endif 
#endif 
          MPI_Comm_size(coll_comm,&coll_size);
          MPI_Comm_rank(coll_comm,&coll_myrank); 
        } 

        MPI_Barrier ( MPI_COMM_WORLD );
 
        coll(); 

        if (group_method == 0) coll_comm = MPI_COMM_NULL;
        else if (is_in_coll_comm[Mynode]) MPI_Comm_free(&coll_comm); 
#ifdef WITH_ONESIDED
        if (is_in_coll_comm[Mynode]) MPI_Win_free(&coll_win);
#endif 

      break /*case 2, i.e. collective benchmarks*/;


      case 3: /* internal dump */
        if ( Mynode == 0 )
        {
          if (nl=='\n') printf("\nWhich should be the rank of the node (0-%1d)? ",Numnode-1);
          scanf("%d%c",&client_rank,&nl);
        }
        MPI_Bcast ( &client_rank, 1, MPI_INT, 0, MPI_COMM_WORLD);
        if (Mynode == client_rank)
        {
#ifdef MPX_SUCCESS
/*rolf   dump_global(0); dump_local(0); dump_listener(0); */
/*rolf*/ dump_local(0);
#endif 
        } 
      break /*case 3*/;

 
      case 4: /* time_sync */
        if ( Mynode == 0 )
        {
          if (nl=='\n') printf("Should I choose uncorrected clocks (0/1)? ");
          scanf("%d%c",&use_uncorrected_clocks,&nl);
        } 
        MPI_Bcast ( &use_uncorrected_clocks, 1, MPI_INT, 0, MPI_COMM_WORLD);
        time_sync(use_uncorrected_clocks); 
        if (use_uncorrected_clocks<0)
        { printf("[%3d] time_corr=%20.9lf sec\n",Mynode, time_base);
          fflush(stdout);
        } 
      break /*case 4*/;

 
      case 5: /* pwd */
        {  char *cwd; int i;
           printf("[%2d]", Mynode); 
           if ((cwd = getcwd(NULL, 64)) == NULL)
                { printf(" error in getcwd");}
           else { printf(" cwd=%s", cwd); free(cwd); }
           printf(" args="); 
           for (i=0; i < argc; i++) printf(" %s",argv[i]);  
           printf("\n"); fflush(stdout);
        } 
      break /*case 4*/;
    } /*switch (test_case)*/

    MPI_Barrier ( MPI_COMM_WORLD );

  }while (test_case != 0);

  { char *buff; int size;
    MPI_Buffer_detach( &buff, &size);  free(buff);
  }
  free(Send_Buf); free(SEND_Buf);
#ifdef WITH_ONESIDED
  MPI_Free_mem(Recv_Buf);
#else 
  free(Recv_Buf);
#endif 
  free(is_in_coll_comm); 
  MPI_Finalize ();
  return ( 0 );

}

/* _________________________ BANDWIDTH  __________________________ */

void print_head (int used_nodes)
{
  printf ( "\n%s  %s  (expected bandwidth = %4.0lf MB/s)\n", 
           operation1_str, operation2_str, expected_bandwidth );
  printf (
"#PE=%4d of %4d buflen/   latency for 1 message or coll.  allowed t. avr./    (msg len)/\n"
"     buf len     lat.avg.     min        max      average  for 1 msg  allowd  (lat.-time-%2.0lfus)\n"
"      bytes      mbyte/s  milli sec  milli sec  milli sec  milli sec  -1 [%%]      mbyte/s\n", used_nodes, Numnode, expected_latency
         );
} 
 
void print_band ( int len, char gl, double sum, double max, double min, int numtests)
{
  double  n_mb, avg_time;

#ifdef NORM_MB 
  n_mb = ( (double)len ) / ((double)(1000000L));
#else
  n_mb = ( (double)len ) / ((double)(1024L*1024L));
#endif 
  avg_time = sum / ( band_tests * numtests );

  printf( "  % 10d %c %9.5lf  %9.5lf  %9.5lf  %9.5lf  %9.5lf  %6.0lf",
      len, gl,
      n_mb / avg_time ,
      min / numtests *1000.0,
      max / numtests *1000.0,
      avg_time * 1000.0,
      (n_mb / expected_bandwidth + expected_latency*1e-6) * 1000.0,
      (avg_time / (n_mb / expected_bandwidth + expected_latency*1e-6) - 1) * 100.0 /*[%]*/
        );
  if (avg_time > 1.3*expected_latency*1e-6 ) printf("    %9.5lf",n_mb / (avg_time-expected_latency*1e-6 ));
  else                        printf("             ");
  printf("\n");
  fflush(stdout); 

}

static void band_s_one ( int p, int len)
{
  double   beg, end, min, max, s_time;
  long     cnt;
  MPI_Status  MPI_stat, MPI_Probe_stat;                           /*cg*/
  MPI_Request send_rq, recv_rq;
  long x;
  double  z;

      s_time = max = 0.0;
      min = 99999000.0;

      if(recv_method == 5)                                        /*cg*/
        { MPI_Recv_init( Recv_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &recv_rq );}
      switch(send_method)                                         /*cg*/
        { case 8: 
          case 12:
            MPI_Send_init(Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq); break; 
          case 9: MPI_Bsend_init(Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq); break;
          case 10: MPI_Ssend_init(Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq); break;
          case 11: MPI_Rsend_init(Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq);  break;
        }

      for ( x = -1; x < band_tests; x++ )
        {
          beg = MPI_Wtime () - time_base;
          for ( cnt = 0; cnt < num_tests; cnt++ )
           {
             if(recv_method==3) MPI_Irecv ( Recv_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &recv_rq );

             switch(send_method) {
              case 1: MPI_Send ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD ); break;
              case 2: MPI_Bsend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD ); break;
              case 3: MPI_Ssend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD ); break;
              case 4: MPI_Isend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
              case 5: MPI_Ibsend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
              case 6: MPI_Issend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
              case 7: MPI_Isend ( Send_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &send_rq ); MPI_Request_free(&send_rq); break; /*cg*/
              case  8: /* Send_init */                            /*cg*/
              case  9: /* Bsend_init */                           /*cg*/
              case 10: /* Ssend_init */                           /*cg*/
              case 11: /* Rsend_init */                           /*cg*/
                 MPI_Start ( &send_rq ); 
                 MPI_Wait (&send_rq, &MPI_stat);
                 break;
              case 12: /* Startall */                             /*cg*/
                 MPI_Startall(1, &send_rq); 
                 MPI_Wait(&send_rq, &MPI_stat); 
                 break;                
             }
             switch(recv_method) {
              case 1: MPI_Recv ( Recv_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &MPI_stat ); break;
              case 2: MPI_Irecv ( Recv_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &recv_rq );
                      MPI_Wait  ( &recv_rq, &MPI_stat ); break;
              case 3: MPI_Wait  ( &recv_rq, &MPI_stat ); break;

              case 4: MPI_Probe ( p, 4714, MPI_COMM_WORLD, &MPI_Probe_stat); /*cg*/
                      MPI_Recv ( Recv_Buf, len, MPI_BYTE, p, 4714,
                                 MPI_COMM_WORLD, &MPI_stat); break; 
              case 5: MPI_Start(&recv_rq);                        /*cg*/
                      MPI_Wait(&recv_rq, &MPI_stat); break;
              case 6: MPI_Recv ( Recv_Buf, len, MPI_BYTE, MPI_ANY_SOURCE, 4714, MPI_COMM_WORLD, &MPI_stat ); break;
             }
           }
          end = MPI_Wtime () - time_base;

          if (x >= 0) /* -1 is used for a first non-measured round */
          { 
            z = end - beg; s_time += z;
            if ( z > max ) max = z;
            if ( z < min ) min = z;
          } 
        }
 
      if((send_method >= 8)&&(send_method <= 12))                 /*cg*/
        MPI_Request_free(&send_rq);

      if(recv_method == 5) MPI_Request_free(&recv_rq);            /*cg*/
         
      { /* checking the message in Recv_Buf */
        int i;
        for (i=0; i<len; i++)
           if (Recv_Buf[i] != (i*(p+1)) % 256 ) 
           {
             printf("Myrank = %1d, Recv_Buf[%d]=%3.0lf != %1d\n",
                     Mynode, i, Recv_Buf[i], (i*(p+1))%256);  
           } 
      }

      if ( recv_method == 4 )                                     /*cg*/
      { /* checking statuses of MPI_Probe+MPI_Recv */
        int probe_count, recv_count; 
        MPI_Get_count(&MPI_Probe_stat, MPI_BYTE, &probe_count);
        MPI_Get_count(&MPI_stat, MPI_BYTE, &recv_count);
        if ( ( probe_count != recv_count )||
             ( recv_count < 1 )||
             ( MPI_Probe_stat.MPI_SOURCE != p )||
             ( MPI_stat.MPI_SOURCE != p )||
             ( MPI_Probe_stat.MPI_TAG != 4714 )||
             ( MPI_stat.MPI_TAG != 4714 )
           ) 
          {
            printf( "\n Watching Recv+Probe statuses: \n" );
            printf( "===============================\n\n" );  
            printf( "Sources ( expected / Probe / Recv ): %d / %d / %d\n",
                      p, MPI_Probe_stat.MPI_SOURCE, MPI_stat.MPI_SOURCE );
            printf( "Tags    ( expected / Probe / Recv ): 4714 / %d / %d\n",
                       MPI_Probe_stat.MPI_TAG, MPI_stat.MPI_TAG);
            printf( "Message Lengths    ( Probe / Recv ):  %d / %d\n", 
                      probe_count, recv_count ); 
                        
          }  
      }

      print_band (  len, ' ', s_time, max, min, 2*num_tests );
}

static void band_s ( int p )
{
  int len;
  print_head (2); 
  for(len = min_lng; len <= max_lng;
      len = (step_method==1 ? len*step_lng : len+step_lng)) band_s_one(p,len);
}

static void band_r_one ( int p, int len)
{
  long x, cnt;
  MPI_Status  MPI_stat, MPI_Probe_stat;                           /*cg*/
  MPI_Request send_rq, recv_rq; 
 
      if(recv_method == 5)                                        /*cg*/
        { MPI_Recv_init(Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &recv_rq );} 
      switch(send_method)                                         /*cg*/
        { case 8: 
          case 12:
            MPI_Send_init(Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq); break;
          case 9: MPI_Bsend_init(Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq); break;
          case 10: MPI_Ssend_init(Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq); break;
          case 11: MPI_Rsend_init(Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq); break;
        } 

      for ( x = -1; x < band_tests; x++ )
        {
          for ( cnt = 0; cnt < num_tests; cnt++ )
           {
             if (rpc_or_crossed==1) switch(recv_method) {
              case 1: MPI_Recv ( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &MPI_stat ); break;
              case 2:
              case 3: MPI_Irecv ( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &recv_rq );
                      MPI_Wait  ( &recv_rq, &MPI_stat ); break;

              case 4: MPI_Probe( p, 4713, MPI_COMM_WORLD, &MPI_Probe_stat); /*cg*/
                      MPI_Recv( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &MPI_stat );
                      break;  
              case 5: MPI_Start( &recv_rq);                       /*cg*/
                      MPI_Wait( &recv_rq, &MPI_stat); break;   
              case 6: MPI_Recv ( Recv_Buf, len, MPI_BYTE, MPI_ANY_SOURCE, 4713, MPI_COMM_WORLD, &MPI_stat ); break;
             }
 
             if ((rpc_or_crossed!=1) && (recv_method==3))
                      MPI_Irecv ( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &recv_rq );
             switch(send_method) {
              case 1: MPI_Send ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD ); break;
              case 2: MPI_Bsend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD ); break;
              case 3: MPI_Ssend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD ); break;
              case 4: MPI_Isend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
              case 5: MPI_Ibsend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
              case 6: MPI_Issend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq ); MPI_Wait  ( &send_rq, &MPI_stat ); break;
             case 7: MPI_Isend ( Send_Buf, len, MPI_BYTE, p, 4714, MPI_COMM_WORLD, &send_rq ); MPI_Request_free ( &send_rq); break; /*cg*/
             case 8:                                              /*cg*/
             case 9:
             case 10:
             case 11:  MPI_Start (&send_rq); MPI_Wait(&send_rq, &MPI_stat); break;
             case 12:  MPI_Startall (1, &send_rq);                /*cg*/
                       MPI_Wait (&send_rq, &MPI_stat); break;  
             }
 
             if (rpc_or_crossed!=1) switch(recv_method) {
              case 1: MPI_Recv ( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &MPI_stat ); break;
              case 2: MPI_Irecv ( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &recv_rq );
                      MPI_Wait  ( &recv_rq, &MPI_stat ); break;
              case 3: MPI_Wait  ( &recv_rq, &MPI_stat ); break;
              case 4: MPI_Probe( p, 4713, MPI_COMM_WORLD, &MPI_Probe_stat); /*cg*/
                      MPI_Recv( Recv_Buf, len, MPI_BYTE, p, 4713, MPI_COMM_WORLD, &MPI_stat );
                      break;  
              case 5: MPI_Start(&recv_rq);                        /*cg*/
                      MPI_Wait(&recv_rq, &MPI_stat); break;
              case 6: MPI_Recv ( Recv_Buf, len, MPI_BYTE, MPI_ANY_SOURCE, 4713, MPI_COMM_WORLD, &MPI_stat ); break;
             }
           }
        }
       
      if((send_method >= 8)&&(send_method <= 12))                 /*cg*/
        MPI_Request_free(&send_rq);

      if(recv_method == 5) MPI_Request_free(&recv_rq);            /*cg*/
 
      { /* checking the message in Recv_Buf */
        int i;
        for (i=0; i<len; i++)
           if (Recv_Buf[i] != (i*(p+1)) % 256) 
           {
             printf("Myrank = %1d, Recv_Buf[%d]=%3.0lf != %1d\n",
                     Mynode, i, Recv_Buf[i], (i*(p+1))%256);  
           } 
      }
 
      if ( recv_method == 4 )                                     /*cg*/
      { /* checking statuses of MPI_Probe+MPI_Recv */
        int probe_count, recv_count; 
        MPI_Get_count(&MPI_Probe_stat, MPI_BYTE, &probe_count);
        MPI_Get_count(&MPI_stat, MPI_BYTE, &recv_count);
        if ( ( probe_count != recv_count )||
             ( recv_count < 1 )||
             ( MPI_Probe_stat.MPI_SOURCE != p )||
             ( MPI_stat.MPI_SOURCE != p )||
             ( MPI_Probe_stat.MPI_TAG != 4713 )||
             ( MPI_stat.MPI_TAG != 4713 )
           ) 
          {
            printf( "\n Watching Recv+Probe statuses: \n" );
            printf( "===============================\n\n" );  
            printf( "Sources ( expected / Probe / Recv ): %d / %d / %d\n",
                      p, MPI_Probe_stat.MPI_SOURCE, MPI_stat.MPI_SOURCE );
            printf( "Tags    ( expected / Probe / Recv ): 4713 / %d / %d\n",
                       MPI_Probe_stat.MPI_TAG, MPI_stat.MPI_TAG);
            printf( "Message Lengths    ( Probe / Recv ):  %d / %d\n", 
                      probe_count, recv_count ); 
                        
          } 
      } 
}

static void band_r ( int p )
{
  int len;
  for(len = min_lng; len <= max_lng;
      len = (step_method==1 ? len*step_lng : len+step_lng)) band_r_one(p,len);
}

/* _________________________ COLLECTIVE _________________________ */

static void coll_one (int len)
{
  /* measurement at each CPU, written to protocol only by root: */ 
  double   beg,       end,    min,    max,    s_time;
  /* compute globally first begin and globally last end: */
  double   gl_beg, gl_end, gl_min, gl_max, gl_s_time;
  /* compute Maximum of time Difference over all CPUs: */ 
  double   md_diff,        md_min, md_max, md_s_time;
  long     x, cnt;
  int      i, LEN, LEN_LOC, coll_from, from, flag,coll_my_rank_in_node,dest1,source1,
           dest2,source2; 
  double   z;
  double   *Buf;
  MPI_Status MPI_stat;

    if (is_in_coll_comm[Mynode])
    { 
      gl_s_time = s_time = gl_max = max = md_s_time = md_max = 0.0;
      gl_min = min = md_min = 99999000.0;
      Buf = (Mynode==root_in_comm_world ? SEND_Buf : RECV_Buf); 
      LEN = len / sizeof (double);
      LEN_LOC = len / sizeof (double_int);
      if((coll_operation==24)||(coll_operation==25))              /*cg*/
        /* Initialization of Recv_Buf */
          for(i=0;i<max_len;i++) Recv_Buf[i]=(i*(coll_myrank+1))%256;

      for ( x = -1; x < band_tests; x++ )
        {
          MPI_Barrier ( coll_comm ); 
          beg = MPI_Wtime () - time_base;
          for ( cnt = 0; cnt < num_tests; cnt++ )
            switch (coll_operation) {
             case 0 /*Barrier*/: 
               MPI_Barrier ( coll_comm ); break; 
             case 1 /*Bcast*/: 
               MPI_Bcast ( Buf, LEN, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 2 /*Reduce(sum)*/: 
               MPI_Reduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_SUM, root_in_coll_comm, coll_comm); break; 
             case 3 /*Bcast+Barrier*/: 
               MPI_Bcast ( Buf, LEN, MPI_DOUBLE, root_in_coll_comm, coll_comm);
               MPI_Barrier ( coll_comm ); break; 
             case 4 /*Allreduce(sum)*/: 
               MPI_Allreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_SUM, coll_comm); break; 
             case 5 /*Reduce(sum)+Bcast*/: 
               MPI_Reduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_SUM, root_in_coll_comm, coll_comm);
               MPI_Bcast ( RECV_Buf, LEN, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 6 /*Reduce(sum)+Barrier*/: 
               MPI_Reduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_SUM, root_in_coll_comm, coll_comm);
               MPI_Barrier ( coll_comm ); break; 
             case 7  /*Send+Recv+Waitany*/: 
             case 8  /*Send+Recv+Waitall*/: 
             case 20 /*Send+Recv+Waitsome*/:                      /*cg*/
             case 21 /*Send+Recv+Testany*/:                       /*cg*/
             case 22 /*Send+Recv+Testsome*/:                      /*cg*/
             case 31 /*Send+Recv+Testall*/:                       /*cg*/
              if (Mynode == root_in_comm_world)
               { int flag; 
                 int Num_partner = 0; 
                 for (i=0; i<Numnode; i++)
                   if (is_in_coll_comm[i] && (i != Mynode)) 
                   { MPI_Irecv(Recv_Buf, len, MPI_BYTE, 
                               ((i % 2) == 0 ? i : MPI_ANY_SOURCE), 
                               5000+i,
                               MPI_COMM_WORLD, &(array_of_requests[i]));
                     Num_partner++; 
                   } else  array_of_requests[i] = MPI_REQUEST_NULL;  
                 if((coll_operation==7) 
                     || (coll_operation==21))   
                 { for (i=0; i<Numnode; i++) 
                   { array_of_statuses[i].MPI_SOURCE = -9999;
                     array_of_statuses[i].MPI_TAG = -9999;
                   } 
                   for (i=0; i<Num_partner; ) 
                   { int index; MPI_Status status; 
                     if (coll_operation == 7)
                     {  MPI_Waitany(Numnode, array_of_requests,
                                   &index,&status);
                        i++;
                     } 
                     else /* coll_operation == 21 */              /*cg*/
                     { MPI_Testany(Numnode, array_of_requests,
                                   &index, &flag, &status);
                       if (flag) i++;
                     }
                     if (index != MPI_UNDEFINED )
                       memcpy(array_of_statuses+index, &status,
                              (size_t)(sizeof(MPI_Status)));
                   }
                 }
                 else if ( (coll_operation == 20)||               /*cg*/
                           (coll_operation == 22) )               /*cg*/
                 { 
                   int outcount, j;
                   for ( i=0; i<Numnode; i++ ) 
                   { array_of_statuses[i].MPI_SOURCE = -9999;
                     array_of_statuses[i].MPI_TAG = -9999;
                   } 
                   do 
                   {
                     if ( coll_operation == 20 )
                       MPI_Waitsome( Numnode, array_of_requests,
                                     &outcount, array_of_indices,
                                     local_array_of_statuses );
                     else if ( coll_operation == 22 )
                       MPI_Testsome( Numnode, array_of_requests,
                                     &outcount, array_of_indices,
                                     local_array_of_statuses ); 
                     for ( j=0; j<outcount; j++ ) 
                         { memcpy( array_of_statuses + 
                                     array_of_indices[j],
                                   local_array_of_statuses + j,
                                 ( size_t )( sizeof ( MPI_Status ) ) );
                         }
                   } while(outcount!=MPI_UNDEFINED);
                 } /* else if */
                 else if ( coll_operation == 31 )
                   { for(i=0; i<Numnode; i++ ) 
                       { array_of_statuses[i].MPI_SOURCE = -9999;
                         array_of_statuses[i].MPI_TAG = -9999;
                       } 
               do{
                MPI_Testall( Numnode, array_of_requests,
                             &flag, array_of_statuses );
                                         
                  } while(!flag);
                } 
                 else /* case 8 */
                 { MPI_Waitall(Numnode,array_of_requests,
                   array_of_statuses);
                 }
               } else MPI_Ssend ( Send_Buf, len, MPI_BYTE, root_in_comm_world, 5000+Mynode, MPI_COMM_WORLD );
               break;
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
             case 9 /*MYreduce(sum)*/: 
               MPI_MYreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, 
                             MPI_SUM, root_in_coll_comm, coll_comm); break; 
             case 10 /*MYreduce(sum)+Barrier*/: 
               MPI_MYreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, 
                             MPI_SUM, root_in_coll_comm, coll_comm);
               MPI_Barrier ( coll_comm ); break; 
             case 11 /*MYreduce(max)*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_MYreduce( SEND_Buf, RECV_Buf, LEN,MPI_DOUBLE,MPI_MAX,
                             root_in_coll_comm, coll_comm);
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
             case 12 /*MYreduce(max)+Barrier*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_MYreduce( SEND_Buf, RECV_Buf, LEN,MPI_DOUBLE,MPI_MAX,
                             root_in_coll_comm, coll_comm);
               MPI_Barrier ( coll_comm );
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
#endif 
             case 13 /*Reduce(max)*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_Reduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_MAX,
                           root_in_coll_comm, coll_comm);
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
             case 14 /*Reduce(max)+Barrier*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_Reduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_MAX,
                           root_in_coll_comm, coll_comm);
               MPI_Barrier ( coll_comm );
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
             case 15 /*NOOP*/:
               break; 
             case 16 /*Sendrecv(in a ring)*/:
               coll_from = (coll_myrank ? coll_myrank-1 : coll_size-1);
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                             (coll_myrank+1) % coll_size, 5700, 
                             Recv_Buf, len, MPI_BYTE, coll_from, 5700,
                             coll_comm, &MPI_stat );
               break; 
             case 42 /*Sendrecv (one ring vertical out of CPUs_per_Node)*/:
               coll_my_rank_in_node = coll_myrank % CPUs_per_Node;
               if(coll_my_rank_in_node == 0) 
                { 
                  dest1 = (coll_myrank + CPUs_per_Node) % coll_size;
                  source1 = (coll_myrank - CPUs_per_Node + coll_size) % coll_size;
                  MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                                dest1, 5700, 
                                Recv_Buf, len, MPI_BYTE, source1, 5700,
                                coll_comm, &MPI_stat );
                } 
               break; 
             case 46 /*Sendrecv (one ring vertical out of CPUs_per_Node)*/:
               coll_my_rank_in_node = coll_myrank % CPUs_per_Node;
               if( (coll_my_rank_in_node%stride_vert_rings == 0) && 
                   (coll_my_rank_in_node/stride_vert_rings < num_vert_rings) )
                { 
                  dest1 = (coll_myrank + CPUs_per_Node) % coll_size;
                  source1 = (coll_myrank - CPUs_per_Node + coll_size) % coll_size;
                  MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                                dest1, 5700, 
                                Recv_Buf, len, MPI_BYTE, source1, 5700,
                                coll_comm, &MPI_stat );
                } 
               break; 
             case 43 /*Sendrecv (CPUs_per_Node rings vertical)*/:
               dest1 = (coll_myrank + CPUs_per_Node) % coll_size;
               source1 = (coll_myrank - CPUs_per_Node + coll_size) % coll_size;
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                             dest1, 5700, 
                             Recv_Buf, len, MPI_BYTE, source1, 5700,
                             coll_comm, &MPI_stat );
               break; 
             case 44 /*Sendrecv (CPUs_per_Node rings vertical + ring inside of nodes)*/:
               coll_my_rank_in_node = coll_myrank % CPUs_per_Node;
               dest1 = (coll_myrank + CPUs_per_Node) % coll_size;
               source1 = (coll_myrank - CPUs_per_Node + coll_size) % coll_size;
               dest2 = coll_myrank + 1;
               if(coll_my_rank_in_node == (CPUs_per_Node-1)) dest2 = dest2 - CPUs_per_Node;
               source2 = coll_myrank - 1;
               if(coll_my_rank_in_node == 0) 
               source2 = source2 + CPUs_per_Node;
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                             dest1, 5700, 
                             Recv_Buf, len, MPI_BYTE, source1, 5700,
                             coll_comm, &MPI_stat );
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                             dest2, 5700, 
                             Recv_Buf, len, MPI_BYTE, source2, 5700,
                             coll_comm, &MPI_stat );
               break;
             case 45 /*Sendrecv (rings inside of nodes)*/:
               coll_my_rank_in_node = coll_myrank % CPUs_per_Node;
               dest2 = coll_myrank + 1;
               if(coll_my_rank_in_node == (CPUs_per_Node-1)) 
               dest2 = dest2 - CPUs_per_Node;
               source2 = coll_myrank - 1;
               if(coll_my_rank_in_node == 0) source2 = source2 + CPUs_per_Node;
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE, 
                             dest2, 5700, 
                             Recv_Buf, len, MPI_BYTE, source2, 5700,
                             coll_comm, &MPI_stat );
               break;
             case 23 /*Sendrecv with MPI_ANY_SOURCE */:           /*cg*/
               coll_from = (coll_myrank ? coll_myrank-1 : coll_size-1);
               MPI_Sendrecv( Send_Buf, len, MPI_BYTE,
                             (coll_myrank+1) % coll_size, 5700,
                             Recv_Buf, len, MPI_BYTE, MPI_ANY_SOURCE, 
                             5700, coll_comm, &MPI_stat );
               break;
             case 24 /*Sendrecv_replace(in a ring)*/:             /*cg*/
               coll_from = (coll_myrank-1+coll_size) % coll_size;
               MPI_Sendrecv_replace(Recv_Buf, len, MPI_BYTE,
                                    (coll_myrank+1) % coll_size,
                                    5700, coll_from, 5700,
                                    coll_comm, &MPI_stat);
               break; 
             case 25 /*Sendrecv_replace with MPI_ANY_SOURCE */:   /*cg*/
               coll_from = (coll_myrank-1+coll_size) % coll_size;
               MPI_Sendrecv_replace(Recv_Buf, len, MPI_BYTE,
                                    (coll_myrank+1) % coll_size,
                                    5700, MPI_ANY_SOURCE, 5700,
                                    coll_comm, &MPI_stat);
               break;
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
             case 17 /*MYallreduce(sum)*/: 
               MPI_MYallreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_SUM, coll_comm); break; 
             case 18 /*MYallreduce(max)*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_MYallreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_MAX, coll_comm);
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
#endif 
             case 19 /*Allreduce(max)*/: 
               SEND_Buf[Mynode] = 1234567;
               MPI_Allreduce( SEND_Buf, RECV_Buf, LEN, MPI_DOUBLE, MPI_MAX, coll_comm);
               SEND_Buf[Mynode] = Mynode * (Mynode+1); break; 
             case 26 /*Alltoall (sendlng=lng)*/: 
               MPI_Alltoall ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, coll_comm); break; 
             case 27 /*Alltoall (sendlng=lng/n)*/: 
               MPI_Alltoall ( SEND_Buf, LEN/coll_size, MPI_DOUBLE, RECV_Buf, LEN/coll_size, MPI_DOUBLE, coll_comm); break; 
             case 28 /*Alltoall (sendlng=lng/n^2)*/: 
               MPI_Alltoall ( SEND_Buf, LEN/(coll_size*coll_size), MPI_DOUBLE, RECV_Buf, LEN/(coll_size*coll_size), MPI_DOUBLE, coll_comm); break; 
             case 29 /*Allgather (sendlng=lng/n)*/: 
               MPI_Allgather ( SEND_Buf, LEN/coll_size, MPI_DOUBLE, RECV_Buf, LEN/coll_size, MPI_DOUBLE, coll_comm); break; 
             case 30 /*Allgather (sendlng=lng/n^2)*/: 
               MPI_Allgather ( SEND_Buf, LEN/(coll_size*coll_size), MPI_DOUBLE, RECV_Buf, LEN/(coll_size*coll_size), MPI_DOUBLE, coll_comm); break; 
             case 32 /*half<->half left/right ISend/Irecv*/:
              {MPI_Request send_rq, recv_rq;  int diff, index;
               diff = (coll_size+1)/2; 
               if (coll_myrank < coll_size/2) 
               { /* first half */
                 MPI_Isend ( Send_Buf, len, MPI_BYTE, coll_myrank+diff, 1201, coll_comm, &send_rq ); 
                 MPI_Irecv ( Recv_Buf, len, MPI_BYTE, coll_myrank+diff, 1202, coll_comm, &recv_rq );
                 do { MPI_Waitany(1, &recv_rq, &index, &MPI_stat);} 
                  while(index != MPI_UNDEFINED);
                 MPI_Waitall (1, &send_rq, &MPI_stat); 
               } 
               if (coll_myrank >= diff) 
               { /* second half */
                 MPI_Isend ( Send_Buf, len, MPI_BYTE, coll_myrank-diff, 1202, coll_comm, &send_rq ); 
                 MPI_Irecv ( Recv_Buf, len, MPI_BYTE, coll_myrank-diff, 1201, coll_comm, &recv_rq );
                 do { MPI_Waitany(1, &recv_rq, &index, &MPI_stat);} 
                  while(index != MPI_UNDEFINED);
                 MPI_Waitall (1, &send_rq, &MPI_stat); 
               } 
              } break; 
             case 33 /*Scatter (sendlng=lng)*/: 
               MPI_Scatter ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 34 /*Scatter (sendlng=lng/n)*/: 
               MPI_Scatter ( SEND_Buf, LEN/coll_size, MPI_DOUBLE, RECV_Buf, LEN/coll_size, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 35 /*Gather (sendlng=lng)*/: 
               MPI_Gather ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 36 /*Gather (sendlng=lng/n)*/: 
               MPI_Gather ( SEND_Buf, LEN/coll_size, MPI_DOUBLE, RECV_Buf, LEN/coll_size, MPI_DOUBLE, root_in_coll_comm, coll_comm); break; 
             case 37 /*Allgather (sendlng=lng)*/: 
               MPI_Allgather ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, coll_comm); break;
             case 38 /*Allgather (sendlng=lng) simulating an optimization*/: 
               /* MPI_Allgather ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, coll_comm); */
               { int i, diff, partner;
                 for (diff=1; diff<coll_size; diff=diff*2 ) {}
                 if (diff==coll_size)
                 { 
                   MPI_Sendrecv( SEND_Buf, LEN, MPI_DOUBLE,
                                 coll_myrank,5800, 
                                 RECV_Buf, LEN, MPI_DOUBLE, 
                                 coll_myrank,5800, coll_comm,&MPI_stat);
                   for (i=1, diff=1; diff<coll_size; i++, diff=diff*2 ) 
                   { partner = ((coll_myrank % (2*diff)) < diff 
                                   ? coll_myrank + diff
                                   : coll_myrank - diff);
                     MPI_Sendrecv( SEND_Buf, LEN*diff, MPI_DOUBLE,
                                   partner,5800+i, 
                                   RECV_Buf, LEN*diff, MPI_DOUBLE, 
                                   partner,5800+i, coll_comm,&MPI_stat);
                   } 
                 } 
               } 
              break; 
             case 39 /* MYallgather (sendlng=lng)*/: 
               MYallgather ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, coll_comm); break;
             case 40 /* CHallgather (sendlng=lng)*/: 
               CHallgather ( SEND_Buf, LEN, MPI_DOUBLE, RECV_Buf, LEN, MPI_DOUBLE, coll_comm); break;
#ifdef WITH_ONESIDED
             case 41 /* Fence Put(c=lng, alltoall) Fence */: 
               /* local access to RECV_buffers */ 
               MPI_Win_fence(MPI_MODE_NOPRECEDE,coll_win); 
               /* RMA access to RECV_buffers */ 
               for (i=0; i<coll_size; i++) 
                 MPI_Put(SEND_Buf,LEN,MPI_DOUBLE, 
                         i,LEN*coll_myrank,LEN,MPI_DOUBLE, coll_win);
               MPI_Win_fence(MPI_MODE_NOSTORE|MPI_MODE_NOPUT|
                             MPI_MODE_NOSUCCEED,coll_win); 
              break;
#endif 
             case 47 /*Allreduce(maxloc)*/: 
               MPI_Allreduce( SEND_LOC_Buf, RECV_LOC_Buf, LEN_LOC, MPI_DOUBLE_INT,
                             MPI_MAXLOC, coll_comm); break;
             case 48 /*Reduce(maxloc)*/: 
               MPI_Reduce( SEND_LOC_Buf, RECV_LOC_Buf, LEN_LOC, MPI_DOUBLE_INT,
                             MPI_MAXLOC, root_in_coll_comm, coll_comm); break;
#ifdef WITH_MYREDUCE
             case 49 /*MYallreduce(maxloc)*/: 
               MPI_MYallreduce( SEND_LOC_Buf, RECV_LOC_Buf, LEN_LOC, MPI_DOUBLE_INT,
                             MPI_MAXLOC, coll_comm); break;
             case 50 /*MYreduce(maxloc)*/: 
               MPI_MYreduce( SEND_LOC_Buf, RECV_LOC_Buf, LEN_LOC, MPI_DOUBLE_INT,
                             MPI_MAXLOC, root_in_coll_comm, coll_comm); break;
#endif 
            } /*coll_operation*/ 
          end = MPI_Wtime () - time_base;

          if (x >= 0) /* -1 is used for a first non-measured round */
          { 
            z = end - beg; s_time += z;
            if ( z > max ) max = z;
            if ( z < min ) min = z;
   
            MPI_Reduce( &beg, &gl_beg, 1, MPI_DOUBLE, MPI_MIN,
                        root_in_coll_comm, coll_comm);
            MPI_Reduce( &end, &gl_end, 1, MPI_DOUBLE, MPI_MAX,
                        root_in_coll_comm, coll_comm);
            MPI_Reduce( &z, &md_diff, 1, MPI_DOUBLE, MPI_MAX,
                        root_in_coll_comm, coll_comm);

            if (Mynode == root_in_comm_world)  
            { z = gl_end - gl_beg; gl_s_time += z;
              if ( z > gl_max ) gl_max = z;
              if ( z < gl_min ) gl_min = z;
              md_s_time += md_diff; 
              if ( md_diff > md_max ) md_max = md_diff;
              if ( md_diff < md_min ) md_min = md_diff;
            } 
          } 
        }

      switch (coll_operation) { /* checking the last result */
       case 0 /*Barrier*/: 
         len = LEN * sizeof (double); 
        break; 
       case 2 /*Reduce(sum)*/: 
       case 6 /*Reduce(sum)+Barrier*/: 
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
       case 9 /*MYreduce(sum)*/: 
       case 10 /*MYreduce(sum)+Barrier*/: 
#endif 
         if (root_in_comm_world == Mynode)
          for (i=0; i<LEN; i++)
           if (RECV_Buf[i] != i * sum_nodes1) 
           {
             printf("Myrank = %1d, RECV_Buf[%d]=%3.0lf != %1d\n",
                     Mynode, i, RECV_Buf[i], i * sum_nodes1);  
             /* CAUTION: The comparison should be exact until a mantissa
                         with 52 bits (IEEE 64-bit-double), i.e., until
                         i=2**20 (LEN=1048576, len=8388608) and #PEs = 2895 */
           } 
         len = LEN * sizeof (double); 
        break; 
       case 1 /*Bcast*/: 
       case 3 /*Bcast+Barrier*/: 
         for (i=0; i<LEN; i++) 
           if (Buf[i] != i*(root_in_comm_world+1)) 
           {
             printf("Myrank = %1d, Buf[%d]=%3.0lf != %1d\n",
                     Mynode, i, Buf[i], i*(root_in_comm_world+1));  
           } 
         len = LEN * sizeof (double); 
        break; 
       case 4 /*Allreduce(sum)*/: 
       case 5 /*Reduce(sum)+Bcast*/: 
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
       case 17 /*MYallreduce(sum)*/: 
#endif 
         for (i=0; i<LEN; i++)
           if (RECV_Buf[i] != i * sum_nodes1) 
           {
             printf("Myrank = %1d, RECV_Buf[%d]=%3.0lf != %1d\n",
                     Mynode, i, RECV_Buf[i], i * sum_nodes1);  
             /* CAUTION: The comparison should be exact until a mantissa
                         with 52 bits (IEEE 64-bit-double), i.e., until
                         i=2**20 (LEN=1048576, len=8388608) and #PEs = 2895 */
           } 
         len = LEN * sizeof (double); 
        break; 
       case 7  /* Send+Recv+Waitany  */: 
       case 8  /* Send+Recv+Waitall  */: 
       case 20 /* Send+Recv+Waitsome */: 
       case 21 /* Send+Recv+Testany  */:
       case 22 /* Send+Recv+Testsome */:
       case 31 /* SendRecv+Testall   */:
         if (Mynode == root_in_comm_world)
         { int val;
           printf("Source:"); 
           for (i=0; i<Numnode; i++)
           { val = array_of_statuses[i].MPI_SOURCE;
             if (val==-9999) printf(" ----"); else printf(" %4d",val); }
           printf("\n"); 
           printf("Tag:   "); 
           for (i=0; i<Numnode; i++)
           { val = array_of_statuses[i].MPI_TAG;
             if (val==-9999) printf(" ----"); else printf(" %4d",val); }
           printf("\n"); 
         } 
        break; 
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
       case 11 /*MYreduce(max)*/: 
       case 12 /*MYreduce(max)+Barrier*/: 
#endif 
       case 13 /*Reduce(max)*/: 
       case 14 /*Reduce(max)+Barrier*/: 
         if (root_in_comm_world == Mynode)
         { int max_node = 0;
           for (i=0; i<Numnode; i++) if(is_in_coll_comm[i]) max_node=i; 
           for (i=0; i<LEN; i++)
           { double cmp_val;
             cmp_val = i * (max_node+1);
             if ((i<Numnode) && is_in_coll_comm[i] && (cmp_val<1234567))
               cmp_val = 1234567;
             if (RECV_Buf[i] != cmp_val) 
             {
               printf("Myrank = %1d, RECV_Buf[%d]=%3.0lf != %3.0lf\n",
                       Mynode, i, RECV_Buf[i], cmp_val );  
             } 
           } 
         } 
         len = LEN * sizeof (double); 
        break; 
       case 15 /*NOOP*/:
        break; 
       case 16 /*Sendrecv(in a ring)*/:
       case 23 /*Sendrecv with MPI_ANY_SOURCE */: 
         { /* checking the message in Recv_Buf */
           int i;
           MPI_Group world_group, coll_group;
           MPI_Comm_group(MPI_COMM_WORLD, &world_group);
           MPI_Comm_group(coll_comm, &coll_group);
           MPI_Group_translate_ranks(coll_group, 1, &coll_from,
                                     world_group, &from); 
           MPI_Group_free(&world_group); MPI_Group_free(&coll_group); 
           
           for (i=0; i<len; i++)
             if (Recv_Buf[i] != (i*(from+1)) % 256) 
              {
                printf("Myrank = %1d, Recv_Buf[%d]=%3d != %3d\n",
                        Mynode, i, Recv_Buf[i], (i*(from+1))%256);  
              }
         } 
        break;
       case 24 /*Sendrecv_replace(in a ring)*/:                   /*cg*/
       case 25 /*Sendrecv_replace with MPI_ANY_SOURCE */: 
         { /* checking the message in Recv_Buf */
           int i;
          /*   MPI_Group world_group, coll_group;
           *   MPI_Comm_group(MPI_COMM_WORLD, &world_group);
           *   MPI_Comm_group(coll_comm, &coll_group);
           */ 
           coll_from=((coll_myrank-(((band_tests+1)*num_tests) % coll_size))+coll_size) %  coll_size;
           /*  MPI_Group_free(&world_group); MPI_Group_free(&coll_group); 
            */ 
           for (i=0; i<len; i++)
             if ( (Recv_Buf[i]) != ( (i*(coll_from+1))%256) ) 
               {
                 printf( "Myrank = %1d, Recv_Buf[%d]=%3d != %3d\n",
                          Mynode, i, Recv_Buf[i],( (i*(coll_from+1))%256) ); 
               }
         }
        break; 
#if defined(MPX_SUCCESS) || defined(WITH_MYREDUCE)
       case 18 /*MYallreduce(max)*/: 
#endif 
       case 19 /*Allreduce(max)*/: 
         { int max_node = 0;
           for (i=0; i<Numnode; i++) if(is_in_coll_comm[i]) max_node=i; 
           for (i=0; i<LEN; i++)
           { double cmp_val;
             cmp_val = i * (max_node+1);
             if ((i<Numnode) && is_in_coll_comm[i] && (cmp_val<1234567))
               cmp_val = 1234567;
             if (RECV_Buf[i] != cmp_val) 
             {
               printf("Myrank = %1d, RECV_Buf[%d]=%3.0lf != %3.0lf\n",
                       Mynode, i, RECV_Buf[i], cmp_val );  
             } 
           } 
         } 
         len = LEN * sizeof (double); 
        break; 
       case 26 /*Alltoall (sendlng=lng)*/: /*TODO*/
         len = LEN * sizeof (double);  break;
       case 27 /*Alltoall (sendlng=lng/n)*/: /*TODO*/
         len = (LEN/coll_size)*coll_size * sizeof (double);  break;
       case 28 /*Alltoall (sendlng=lng/n^2)*/: /*TODO*/
         len = ((LEN/coll_size)/coll_size)*coll_size*coll_size * sizeof (double);  break;
       case 29 /*Allgather (sendlng=lng/n)*/: /*TODO*/
         len = (LEN/coll_size)*coll_size * sizeof (double);  break;
       case 30 /*Allgather (sendlng=lng/n^2)*/: /*TODO*/
         len = ((LEN/coll_size)/coll_size)*coll_size*coll_size * sizeof (double);  break;
       case 32 /*half<->half left/right ISend/Irecv*/: /*TODO*/
         break;
       case 33 /*Scatter (sendlng=lng)*/:  /*TODO*/
         len = LEN * sizeof (double);  break;
       case 34 /*Scatter (sendlng=lng/n)*/:  /*TODO*/
         len = (LEN/coll_size)*coll_size * sizeof (double);  break;
       case 35 /*Gather (sendlng=lng)*/:  /*TODO*/
         len = LEN * sizeof (double);  break;
       case 36 /*Gather (sendlng=lng/n)*/:  /*TODO*/
         len = (LEN/coll_size)*coll_size * sizeof (double);  break;
       case 37 /*Allgather (sendlng=lng)*/:
       case 39 /*MYallgather (sendlng=lng)*/:
       case 40 /*CHallgather (sendlng=lng)*/:
         { int i, coll_r, r ; 
           double cmp_val;
           MPI_Group world_group, coll_group;
           MPI_Comm_group(MPI_COMM_WORLD, &world_group);
           MPI_Comm_group(coll_comm, &coll_group);
           for (coll_r=0; coll_r<coll_size; coll_r++) 
           { 
             MPI_Group_translate_ranks(coll_group, 1, &coll_r,
                                     world_group, &r); 
             for (i=0; i<LEN; i++)
             { 
               cmp_val = i*(r+1);
               if (RECV_Buf[i+r*LEN] != cmp_val)
                 printf("Myrank = %1d, RECV_Buf[%d+%d*%d]=%3.0lf != %3.0lf\n",
                         Mynode,i,r,LEN,RECV_Buf[i+r*LEN],cmp_val);
             } 
           } 
           MPI_Group_free(&world_group); MPI_Group_free(&coll_group); 
         }   
         len = LEN * sizeof (double);  break;
       case 38 /*Allgather (sendlng=lng) Simulation*/:  /*TODO*/
         len = LEN * sizeof (double);  break;
       case 47 /*Allreduce(maxloc)*/: 
       case 48 /*Reduce(maxloc)*/: 
#ifdef WITH_MYREDUCE
       case 49 /*MYallreduce(maxloc)*/: 
       case 50 /*MYreduce(maxloc)*/: 
#endif 
         if ((coll_operation==47) || (coll_operation==49) 
                                  || (root_in_comm_world == Mynode))
         { 
           for (i=0; i<LEN_LOC; i++)
           { double cmp_val; int cmp_idx;
             cmp_val = -1;
             cmp_idx = i % Numnode; 
             if (is_in_coll_comm[cmp_idx]) cmp_val = cmp_idx;
             if ((cmp_val >= 0) 
                 && ((RECV_LOC_Buf[i].v != cmp_val) || (RECV_LOC_Buf[i].i != cmp_idx)) ) 
             {
               printf("Myrank = %1d, RECV_LOC_Buf[%d].v=%3.0lf != %3.0lf or .i=%3d != %3d\n",
                       Mynode, i, RECV_LOC_Buf[i].v, cmp_val, RECV_LOC_Buf[i].i, cmp_idx);  
             } 
           } 
         } 
         /* 
           for (i=-1; i<coll_myrank; i++) MPI_Barrier(coll_comm);
           printf("[%2d] SEND_LOC_Buf.v= ",Mynode);
           for (i=0; i<LEN_LOC; i++) printf(" %5.2f",SEND_LOC_Buf[i].v); printf("\n");
           printf("[%2d] SEND_LOC_Buf.i= ",Mynode);
           for (i=0; i<LEN_LOC; i++) printf(" %5d",SEND_LOC_Buf[i].i); printf("\n");
           for (i=coll_myrank; i<coll_size; i++) MPI_Barrier(coll_comm);
           for (i=-1; i<coll_myrank; i++) MPI_Barrier(coll_comm);
           printf("[%2d] RECV_LOC_Buf.v= ",Mynode);
           for (i=0; i<LEN_LOC; i++) printf(" %5.2f",RECV_LOC_Buf[i].v); printf("\n");
           printf("[%2d] RECV_LOC_Buf.i= ",Mynode);
           for (i=0; i<LEN_LOC; i++) printf(" %5d",RECV_LOC_Buf[i].i); printf("\n");
           for (i=coll_myrank; i<coll_size; i++) MPI_Barrier(coll_comm);
         */
         len = LEN_LOC * sizeof (double_int); 
        break; 
      } /*coll_operation*/ 
#     ifdef PRINT_ROOT_TIME
        /* only measured at root: */ 
        if (Mynode == root_in_comm_world) print_band(len,' ',s_time,max,min,num_tests);
#     endif 
#     ifndef PRINT_NO_MD_TIME
        /* time = maximum of the execution times on all processes */ 
        if (Mynode == root_in_comm_world) print_band(len,'M',md_s_time,md_max,md_min,num_tests);
#     endif 
#     ifdef PRINT_GLOBAL_TIME
        /* time = (max of all end times) - (min of all begin times) */
        if (Mynode == root_in_comm_world) print_band(len,'G',gl_s_time,gl_max,gl_min,num_tests);
#     endif 
    } 
}

static void coll ()
{
  int len;
  if (Mynode == root_in_comm_world) print_head (coll_size); 
  for(len = min_lng; len <= max_lng;
      len = (step_method==1 ? len*step_lng : len+step_lng)) coll_one(len);
}
 
int MYallgather(void* s_buf, int s_cnt, MPI_Datatype s_type, 
            void* R_buf, int r_cnt, MPI_Datatype r_type, MPI_Comm comm) 
{ int dist, partner, size, rank, block, p_block; 
  MPI_Status status; MPI_Aint extent;
  char *r_buf = (char *)R_buf; 
  MPI_Comm_size(comm, &size);  MPI_Comm_rank(comm, &rank);
  MPI_Type_extent(r_type, &extent); 
  dist = 1; while (dist<size) dist=dist*2; 
  if (dist==size) /* i.e. size == some power of two */
  { 
    MPI_Sendrecv( s_buf,                     s_cnt, s_type, rank, 5800, 
                  r_buf + rank*extent*r_cnt, r_cnt, r_type, rank, 5800,
                  comm, &status);
    for (dist=1; dist<size; dist=dist*2 ) 
    { block = rank - (rank % dist); 
      if ((rank % (2*dist)) < dist)
      { /* lower node in the exchange */
        partner = rank + dist;  p_block = block + dist;
      } else
      { /* higher  node in the exchange */
        partner = rank - dist;  p_block = block - dist;
      } 
      MPI_Sendrecv( r_buf + block * extent*r_cnt, r_cnt*dist, r_type, 
                    partner, 5800, 
                    r_buf + p_block*extent*r_cnt, r_cnt*dist, r_type, 
                    partner, 5800, comm, &status);
    } 
  } 
  else
  MPI_Allgather( s_buf, s_cnt, s_type, r_buf, r_cnt, r_type, comm );
} 

int CHallgather(void* s_buf, int s_cnt, MPI_Datatype s_type,
            void* r_buf, int r_cnt, MPI_Datatype r_type, MPI_Comm comm) 
{
  int        size, rank;
  int        mpi_errno = MPI_SUCCESS;
  MPI_Status status;
  MPI_Aint   r_extent;
  int        j, jnext, i, right, left;

  /* Get the size of the communicator */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );

  /* Do a gather for each process in the communicator
     This is the "circular" algorithm for allgather - each process sends to
     its right and receives from its left.  This is faster than simply
     doing size Gathers.
   */

  MPI_Type_extent ( r_type, &r_extent );

  /* First, load the "local" version in the r_buf. */
  mpi_errno = 
      MPI_Sendrecv( s_buf, s_cnt, s_type, rank, 5801,
                    (void *)((char *)r_buf + rank * r_cnt * r_extent),
                           r_cnt, r_type, rank, 5801, comm, &status );
  if (mpi_errno) return mpi_errno;

  /* 
     Now, send left to right.  This fills in the receive area in 
     reverse order.
   */
  left  = (size + rank - 1) % size;
  right = (rank + 1) % size;
  
  j     = rank;
  jnext = left;
  for (i=1; i<size; i++) {
      mpi_errno =
        MPI_Sendrecv((void *)((char *)r_buf +     j * r_cnt * r_extent),
                     r_cnt, r_type, right, 5801,
                     (void *)((char *)r_buf + jnext * r_cnt * r_extent),
                     r_cnt, r_type, left,  5801, comm, &status );
      if (mpi_errno) break;
      j     = jnext;
      jnext = (size + jnext - 1) % size;
      }
  return (mpi_errno);
}

#if defined(SHMEM_PUT) || defined(SHMEM_GET)
int MPISHMEM_Sendrecv(void *sendbuf, int sendcount, MPI_Datatype sendtype,
    int dest, int sendtag, void *recvbuf, int recvcount,
    MPI_Datatype recvtype, int source, int recvtag, MPI_Comm comm,
    MPI_Status *status)
{
#ifdef SHMEM_PUT
        shmem_put(recvbuf, sendbuf, sendcount>>3, dest) ;
        shmem_quiet() ;
#else
        shmem_get(recvbuf, sendbuf, sendcount>>3, source) ;
#endif
        return(0) ;
}
#endif
