/* Copyright: Rolf Rabenseifner, July 2003
 *            High-Performance Computing Center 
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h> 
 
/* ------------------------------------------------------------------- *
 *                                                                     *
 *            Fast MPI_Allreduce and MPI_Reduce algorithms for         *
 *   predefined operations and extensible for used-defined operations  *
 *                                                                     *
 * ------------------------------------------------------------------- *

 * content:
 
    - overview 
    - compile-time options 
    - platform dependent optimization decisions
    - operation handling 
       -- MPI-implementation dependent operation handling
       -- implementation independent default handling 
    - sendrecv handling 
    - 1. algorithm: halving + doubling
    - 2. algorithm: binary blocks based halving + doubling
    - 3. algorithm: pairwise + ring/pttopt
    - 4. algorithm: binary tree (+ bcast)
    - 5. algorithm: recursive doubling (allreduce only)
    - decision between these algorithms and vendors algo.
    - decision - printing main program 

 * CAUTION:
 * ------- 
 
   - MPI_MAXLOC and MPI_MINLOC  are not implemented with these algorithms.
 
   - size==1 (i.e. MPI_(All)Reduce copies the sendbuf into the Recvbuf)
     is not implemented with these algorithms.

   - in both cases, the vendor algorithm is used. 

 * ------------------------------------------------------------------- *
 *                                                                     *
 *                             Overview                                *
 *                                                                     *
 * ------------------------------------------------------------------- *

   MPI_Allreduce and MPI_Reduce are (probably the most) important
   collective MPI routines. In 1997 I developed the 1. algorithm
   to balance the work better than with binary tree algorithms,
   especially for making reduction operations on longer vectors.
  
   Inspired by the Rajeev Thakur's and Bill Gropp's publication at
   Euro PVM/MPI 2003, I have developed more efficient algorithms:
  
   1. algorithm: halving + doubling
   -------------------------------- 
  
      MPI_Allreduce and MPI_Reduce is done in two steps:
        - a special reduce_scatter using recursive halving
        - a special allgather / gather
   
      This first algorithm has a non-optimal load balance for non-power-of-two
      number of processes.

      Computing { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } 
  
   2. algorithm: binary blocks based halving + doubling
   ----------------------------------------------------
  
      Better balancing could be achieved with splitting the number
      of processes in its set of power-of-two numbers, i.e. the binary blocks,
      and implementing recursive halving + doubling mainly inside of these
      blocks. Additional steps are necessary to combine the blocks.
  
      This algorithm is nearly optimal, but non-optimal if the 
      exponents of two adjacent blocks differs to much.

      Computing { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j) + (k+l)] + m } 
  
   3. algorithm: pairwise + ring/pttopt
   ------------------------------------
  
      Optimal load-balance, but non-optimal latency could be achieved
      with a reduce-scatter implemented with pairwise data exchange with all
      other processes. The allgather is implemented by a ring-shift.
      For MPI_Reduce, the gather operation is done by direct messaging.

      Computing 
         (((((a+b)+c)+d)+e)+f)   for buffer segment A
         a + ((((b+c)+d)+e)+f)   for buffer segment B
         (a+b) + (((c+d)+e)+f)   for buffer segment C
         ((a+b)+c) + ((d+e)+f)   for buffer segment D
         (((a+b)+c)+d) + (e+f)   for buffer segment E
         ((((a+b)+c)+d)+e) + f   for buffer segment F 
      
   4. algorithm: binary tree (+ bcast)
   -----------------------------------
  
      Here, a special binary tree algorithm is implemented, to prohibit
      additional messages in the case of non-power-of-two number of processes
      and non-commute (user-defined) operations and to minimize local
      memcpy (only necessary with non-commute operations).
  
      Binary tree implements the worst load-balancing, but the best latency
      due to a minimal number of messages.
  
      In case of MPI_Allreduce, an additional MPI_Bcast must deliver the
      reduction-result to all processes.
 
      Computing { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j) + (k+l)] + m } 
      
   5. algorithm: recursive doubling (allreduce only)
   -------------------------------------------------
  
      Exchange of the full vector pairwise with the neighbors 
      (distance 2**0), computing the reduction on both processes
      of the pairs (redundantly), and then repeating the same procedure
      by recursively doubling the distance. 
      Handling of non-power-of-two number of processes as in
      the 1st algorithm, but with full vectors. 
       
      Computing { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } 

   Compilation
   -----------
 
   Aggressive inlining should be used! 

 * ------------------------------------------------------------------- *
 *                                                                     *
 *                        Compile-time options                         *
 *                                                                     *
 * ------------------------------------------------------------------- *

    -DREDUCE_PROTOCOL    generates a main program that tells on standard 
                         output, which protocol is used in MPI_Reduce();
    -DALLREDUCE_PROTOCOL generates a main program that tells on standard 
                         output, which protocol is used in MPI_Allreduce();
                         the main program must be started as one MPI process; 
                         (-DREDUCE_PROTOCOL and -DALLREDUCE_PROTOCOL
                          are mutually exclusive and  
                          cannot be combined with other switches) 
    -DBYTE_MSG           switches from datatype based communication
                         to MPI_BYTE based communication 
    -DDEBUG=1            writes protocol decision by rank-0-process
    -DDEBUG=2            writes additionally the protocol 
    -DDEBUG=3            writes additionally the buffer content
    -DTEST_NO_COMMUTE    switches to none-commute protocol for all operations
    -DUSE_Irecv          uses Irecv-Send-Wait instead of Sendrecv 
    -DUSE_Isend          uses Isend-Recv-Wait instead of Sendrecv 
                         (-DUSE_Irecv and -DUSE_Isend are mutually exclusive)
    -DUSE_HALF           forces HALF protocol 
    -DUSE_BLCK           forces BLCK protocol 
    -DUSE_PAIR           forces PAIR protocol 
    -DUSE_TREE           forces TREE protocol 
    -DUSE_RDBL           forces RDBL protocol	
*/ 
                        

/* ------------------------------------------------------------------- *
 *                                                                     *
 *              Platform dependent optimization decisions              *
 *                                                                     *
 * ------------------------------------------------------------------- */
 
#ifdef CRAY
#  define SCR_LNG_OPTIM(bytelng)  128 + ((bytelng+127)/256) * 256;
                             /* =  16 + multiple of 32 doubles*/
#  define LOOP_PRAGMA _Pragma("_CRI concurrent")  
#endif 
 

/* try some chache strategy on any platform: */  
#ifndef SCR_LNG_OPTIM
#  define SCR_LNG_OPTIM(bytelng)  128 + ((bytelng+127)/256) * 256;
                             /* =  16 + multiple of 32 doubles*/
#endif 

/* default - no pragma */ 
#ifndef LOOP_PRAGMA
#  define LOOP_PRAGMA /* */
#endif 


/* ------------------------------------------------------------------- *
 *                                                                     *
 *          Operations from left:  b2_rslt := b1 op b2_rslt            *
 *                                                                     *
 * ------------------------------------------------------------------- */
 
/* This operation handling requires, that all predefined operation
   handles are implemented by constants in mpi.h, i.e., by
   cpp macros or be enum constants.
*/
 
#ifdef TEST_NO_COMMUTE
#  define OP_CHECK_COMMUTE  commute=0;
#endif 
 
/* Comments:
    - MPI_REPLACE (MPI-2) is not usable with MPI_REDUCE 
    - This routine may be substituted by a faster implementation dependent macro 
    - Result values (ordered!) 
*/
#define MPI_I_OP_TYPE_NULL   -1  /* MPI_OP_NULL or MPI_REPLACE */ 
#define MPI_I_OP_TYPE_UNIMPL  0  /* predefined, but not implemented operations */
#define MPI_I_OP_TYPE_USERDEF 1  /* user-defined operation */
#define MPI_I_OP_TYPE_PREDEF  2  /* predefined and implemented operations */


 
/* ---------- MPI-implementation-dependent operation handling -------- */
/*            -----------------------------------------------          */
 

/* All 4 algorithms are implemented for commute and non-commute predefined and
   user-defined operations, but without knowledge about the internal structure
   of the operation handles, only predefined operations can be used and implemented
   inside of MPI_I_do_o(). 
   If internal operation handle's information is available, then it can be used
   here, e.g., with
    ifdef platform_xxx 
      define OP_VARIABLES \
        struct MPIR_OP *op_ptr; \
        MPI_User_function *uop;
      define OP_INIT \
        op_ptr = MPIR_GET_OP_PTR(op); \
        MPIR_TEST_MPI_OP(op,op_ptr,comm,(is_all?"MPI_ALLREDUCE":"MPI_REDUCE")); \ 
        uop  = op_ptr->op; 
      ifndef OP_CHECK_COMMUTE
        define OP_CHECK_COMMUTE   commute = op_ptr->commute;
      endif
      define MPI_I_CALL_OP(b1, b2_rslt, count, datatype, op) \
              (*uop)(b1, b2_rslt, &count, &datatype)
      ifdef MPI_REPLACE 
        define MPI_I_get_op_type(op) \
          (op_ptr->permanent ? (op==MPI_REPLACE ? MPI_I_OP_TYPE_NULL : MPI_I_OP_TYPE_PREDEF) : MPI_I_OP_TYPE_USERDEF) 
      else 
        define MPI_I_get_op_type(op) \
          (op_ptr->permanent ? MPI_I_OP_TYPE_PREDEF : MPI_I_OP_TYPE_USERDEF) 
      endif
      define MPI_I_OP_TYPE_IMPL  MPI_I_OP_TYPE_USERDEF /-* and higher *-/
    endif
*/

 
/* ------------ implementation independent default handling ---------- */
/*              -------------------------------------------            */

/* default: only predefined operations are implemented within these 4 algorithms,
            therefore always commute: */
#ifndef OP_CHECK_COMMUTE
#  define OP_CHECK_COMMUTE  commute=1; 
#endif 
 
#ifndef OP_VARIABLES 
#  define OP_VARIABLES /* */
#endif 
 
#ifndef OP_INIT
#  define OP_INIT      /* */
#endif 
 
 
#ifndef MPI_I_CALL_OP
 
# define MPI_I_DO_OP_C_INTEGER(MPI_I_do_op_TYPE,TYPE)               \
  void MPI_I_do_op_TYPE(TYPE* b1, TYPE* b2_rslt, int cnt, MPI_Op op)   \
  { int i;                                                               \
    switch (op) {                                                        \
     case MPI_MAX :                                                                           \
         LOOP_PRAGMA for(i=0;i<cnt;i++) if (b1[i] > b2_rslt[i])  b2_rslt[i] = b1[i]; break;   \
     case MPI_MIN :                                                                           \
         LOOP_PRAGMA for(i=0;i<cnt;i++) if (b1[i] < b2_rslt[i])  b2_rslt[i] = b1[i]; break;   \
     case MPI_SUM : LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] +  b2_rslt[i]; break;   \
     case MPI_PROD: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] *  b2_rslt[i]; break;   \
     case MPI_LAND: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] && b2_rslt[i]; break;   \
     case MPI_LOR : LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] || b2_rslt[i]; break;   \
     case MPI_LXOR: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] != b2_rslt[i]; break;   \
     case MPI_BAND: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] &  b2_rslt[i]; break;   \
     case MPI_BOR : LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] |  b2_rslt[i]; break;   \
     case MPI_BXOR: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] ^  b2_rslt[i]; break;   \
    }                                                                    \
  }
  
# define MPI_I_DO_OP_FP(MPI_I_do_op_TYPE,TYPE)                      \
  void MPI_I_do_op_TYPE(TYPE* b1, TYPE* b2_rslt, int cnt, MPI_Op op)   \
  { int i;                                                               \
    switch (op) {                                                        \
     case MPI_MAX :                                                                           \
         LOOP_PRAGMA for(i=0;i<cnt;i++) if (b1[i] > b2_rslt[i])  b2_rslt[i] = b1[i]; break;   \
     case MPI_MIN :                                                                           \
         LOOP_PRAGMA for(i=0;i<cnt;i++) if (b1[i] < b2_rslt[i])  b2_rslt[i] = b1[i]; break;   \
     case MPI_SUM : LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] +  b2_rslt[i]; break;   \
     case MPI_PROD: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] *  b2_rslt[i]; break;   \
    }                                                                    \
  }
  
# define MPI_I_DO_OP_BYTE(MPI_I_do_op_TYPE,TYPE)                    \
  void MPI_I_do_op_TYPE(TYPE* b1, TYPE* b2_rslt, int cnt, MPI_Op op)   \
  { int i;                                                               \
    switch (op) {                                                        \
     case MPI_BAND: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] &  b2_rslt[i]; break;   \
     case MPI_BOR : LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] |  b2_rslt[i]; break;   \
     case MPI_BXOR: LOOP_PRAGMA for(i=0;i<cnt;i++) b2_rslt[i] = b1[i] ^  b2_rslt[i]; break;   \
    }                                                                    \
  }
  
# define MPI_I_DO_OP_M_LOC(MPI_I_do_op_TYPE,TYPE)                        \
  void MPI_I_do_op_TYPE(TYPE* b1, TYPE* b2_rslt, int cnt, MPI_Op op)     \
  { int i;                                                               \
    switch (op) {                                                        \
     case MPI_MAXLOC : LOOP_PRAGMA                                       \
        for(i=0;i<cnt;i++) if (b1[i].v > b2_rslt[i].v)  b2_rslt[i] = b1[i];      \
                           else if ((b1[i].v == b2_rslt[i].v)                    \
                                    && (b1[i].i < b2_rslt[i].i)) b2_rslt[i].i = b1[i].i; break; \
     case MPI_MINLOC : LOOP_PRAGMA                                       \
        for(i=0;i<cnt;i++) if (b1[i].v < b2_rslt[i].v)  b2_rslt[i] = b1[i];      \
                           else if ((b1[i].v == b2_rslt[i].v)                    \
                                    && (b1[i].i < b2_rslt[i].i)) b2_rslt[i].i = b1[i].i; break; \
    }                                                                    \
  }
  
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_short,  short)
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_int,    int)
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_long,   long)
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_ushort, unsigned short)
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_uint,   unsigned int)
  MPI_I_DO_OP_C_INTEGER( MPI_I_do_op_ulong,  unsigned long)
  MPI_I_DO_OP_FP(        MPI_I_do_op_float,  float)
  MPI_I_DO_OP_FP(        MPI_I_do_op_double, double)
  MPI_I_DO_OP_FP(        MPI_I_do_op_long_double, long double)
  MPI_I_DO_OP_BYTE(      MPI_I_do_op_byte,   char)
  
  typedef struct { float v;       int i;} type_loc_float_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_float_int, type_loc_float_int)
  typedef struct { double v;      int i;} type_loc_double_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_double_int, type_loc_double_int)
  typedef struct { long double v; int i;} type_loc_longdouble_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_longdouble_int, type_loc_longdouble_int)
  typedef struct { long v;        int i;} type_loc_long_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_long_int, type_loc_long_int)
  typedef struct { int v;         int i;} type_loc_int_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_int_int, type_loc_int_int)
  typedef struct { short v;       int i;} type_loc_short_int;
  MPI_I_DO_OP_M_LOC(     MPI_I_do_op_loc_short_int, type_loc_short_int)
  
  
# define MPI_I_DO_OP_CALL(MPI_I_do_op_TYPE,TYPE)                    \
     MPI_I_do_op_TYPE ((TYPE*)b1, (TYPE*)b2_rslt, cnt, op); break;
  
  
  /* this is the main operation routine: */
   
  void MPI_I_do_op(void* b1, void* b2_rslt, int cnt,
                   MPI_Datatype datatype, MPI_Op op)
  {
   switch (datatype) {
    case MPI_SHORT : MPI_I_DO_OP_CALL(MPI_I_do_op_short,  short)
    case MPI_INT   : MPI_I_DO_OP_CALL(MPI_I_do_op_int,    int)
    case MPI_LONG  : MPI_I_DO_OP_CALL(MPI_I_do_op_long,   long)
    case MPI_UNSIGNED_SHORT:
                     MPI_I_DO_OP_CALL(MPI_I_do_op_ushort, unsigned short)
    case MPI_UNSIGNED:
                     MPI_I_DO_OP_CALL(MPI_I_do_op_uint,   unsigned int)
    case MPI_UNSIGNED_LONG:
                     MPI_I_DO_OP_CALL(MPI_I_do_op_ulong,  unsigned long)
    case MPI_FLOAT : MPI_I_DO_OP_CALL(MPI_I_do_op_float,  float)
    case MPI_DOUBLE: MPI_I_DO_OP_CALL(MPI_I_do_op_double, double)
    case MPI_LONG_DOUBLE: MPI_I_DO_OP_CALL(MPI_I_do_op_long_double, long double)
    case MPI_BYTE  : MPI_I_DO_OP_CALL(MPI_I_do_op_byte,   char)
    case MPI_FLOAT_INT       : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_float_int, type_loc_float_int)
    case MPI_DOUBLE_INT      : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_double_int, type_loc_double_int)
    case MPI_LONG_DOUBLE_INT : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_longdouble_int, type_loc_longdouble_int)
    case MPI_LONG_INT        : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_long_int, type_loc_long_int)
    case MPI_2INT            : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_int_int, type_loc_int_int)
    case MPI_SHORT_INT       : MPI_I_DO_OP_CALL(MPI_I_do_op_loc_short_int, type_loc_short_int)
  
   }
  }

 
# define MPI_I_CALL_OP(b1, b2_rslt, count, datatype, op) \
            MPI_I_do_op(b1, b2_rslt, count, datatype, op)
 
#endif /*ifndef MPI_I_CALL_OP*/


#ifndef MPI_I_OP_TYPE_IMPL 
#  define MPI_I_OP_TYPE_IMPL    MPI_I_OP_TYPE_PREDEF /* defines, which category and higher
                                                        is implemented in MPI_I_CALL_OP() */
#endif 
 
#ifndef MPI_I_get_op_type 
  int MPI_I_get_op_type(MPI_Op op)
  { switch (op) {
       case MPI_OP_NULL: 
       case MPI_REPLACE:  /* this line must be removed on MPI-1-only implementations */
         return(MPI_I_OP_TYPE_NULL);
       case MPI_MAX:    case MPI_MIN:  case MPI_SUM:  case MPI_PROD:
       case MPI_BAND:   case MPI_BOR:  case MPI_BXOR:
       case MPI_LAND:   case MPI_LOR:  case MPI_LXOR:
       case MPI_MAXLOC: case MPI_MINLOC:
         return(MPI_I_OP_TYPE_PREDEF);
    } 
    return(MPI_I_OP_TYPE_USERDEF); 
  } 
#endif
 
/* ------------------------------------------------------------------- *
 *   The following code is not needed for -DALLREDUCE_PROTOCOL         *
 *   nor for -DREDUCE_PROTOCOL                                         *
 * ------------------------------------------------------------------- */

#if !(defined(REDUCE_PROTOCOL) || defined(ALLREDUCE_PROTOCOL)) 

 
/* ------------------------------------------------------------------- *
 *                                                                     *
 *                         Sendrecv handling                           *
 *                                                                     *
 * ------------------------------------------------------------------- */

/* BYTE_MSG option requires, that a local variable "typelng" exists
   and contains the extend of the datatype, which itself must be contigues. */
 
#ifdef BYTE_MSG 
# define  MPI_I_Send(sb,sc,sd,dest,st,comm)                            \
             MPI_Send(sb,(sc)*typelng,MPI_BYTE,dest,st,comm)
# define  MPI_I_Recv(rb,rc,rd,source,rt,comm,stat)                     \
             MPI_Recv(rb,(rc)*typelng,MPI_BYTE,source,rt,comm,stat)
# define  MPI_I_Bcast(b,c,d,rt,comm)  MPI_Bcast(b,(c)*typelng,MPI_BYTE,rt,comm)
#else
# define  MPI_I_Send(sb,sc,sd,dest,st,comm)                            \
             MPI_Send(sb,sc,sd,dest,st,comm)
# define  MPI_I_Recv(rb,rc,rd,source,rt,comm,stat)                     \
             MPI_Recv(rb,rc,rd,source,rt,comm,stat)
# define  MPI_I_Bcast(b,c,d,rt,comm)  MPI_Bcast(b,c,d,rt,comm)
#endif

#ifdef USE_Irecv
# define  MPI_I_Sendrecv(sb,sc,sd,dest,st,rb,rc,rd,source,rt,comm,stat)\
           { MPI_Request req;                                          \
             MPI_Irecv(rb,rc,rd,source,rt,comm,&req);                  \
             MPI_I_Send(sb,sc,sd,dest,st,comm);                          \
             MPI_Wait(&req,stat);                                      \
           }
#else
#ifdef USE_Isend
# define  MPI_I_Sendrecv(sb,sc,sd,dest,st,rb,rc,rd,source,rt,comm,stat)\
           { MPI_Request req;                                          \
             MPI_Isend(sb,sc,sd,dest,st,comm,&req);                    \
             MPI_I_Recv(rb,rc,rd,source,rt,comm,stat);                   \
             MPI_Wait(&req,stat);                                      \
           }
#else
#ifdef BYTE_MSG 
# define  MPI_I_Sendrecv(sb,sc,sd,dest,st,rb,rc,rd,source,rt,comm,stat)\
            MPI_Sendrecv(sb,(sc)*typelng,MPI_BYTE,dest,st,rb,(rc)*typelng,MPI_BYTE,source,rt,comm,stat)
#else 
# define  MPI_I_Sendrecv(sb,sc,sd,dest,st,rb,rc,rd,source,rt,comm,stat) \
            MPI_Sendrecv(sb,sc,sd,dest,st,rb,rc,rd,source,rt,comm,stat)
#endif
#endif
#endif
 

/* ------------------------------------------------------------------- *
 *                                                                     *
 *                  1. algorithm: halving + doubling                   *
 *                                                                     *
 * ------------------------------------------------------------------- */
 
/* Copyright: Rolf Rabenseifner, September 1997 and July 2003
 *            High-Performance Computing Center 
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 


/* Fast reduce and allreduce algorithm for longer buffers and predefined
   operations.

   This algorithm is explaned with the example of 13 nodes.
   The nodes are numbered   0, 1, 2, ... 12.
   The sendbuf content is   a, b, c, ...  m.
   The buffer array is notated with ABCDEFGH, this means that
   e.g. 'C' is the third 1/8 of the buffer.

   The algorithm computes

   { [(a+b)+(c+d)] + [(e+f)+(g+h)] }  +  { [(i+j)+k] + [l+m] }

   This is equivalent to the mostly used binary tree algorithm
   (e.g. used in mpich).

   size := number of nodes in the communicator.
   2**n := the power of 2 that is next smaller or equal to the size.
   r    := size - 2**n

   Exa.: size=13 ==> n=3, r=5  (i.e. size == 13 == 2**n+r ==  2**3 + 5)

   The algoritm needs for the execution of one MPI_Reduce

   - for r==0
     exec_time = n*(L1+L2)     + buf_lng * (1-1/2**n) * (T1 + T2 + O/d)

   - for r>0
     exec_time = (n+1)*(L1+L2) + buf_lng               * T1
                               + buf_lng*(1+1/2-1/2**n)*     (T2 + O/d)

   with L1 = latency of a message transfer e.g. by send+recv
        L2 = latency of a message exchange e.g. by sendrecv
        T1 = additional time to transfer 1 byte (= 1/bandwidth)
        T2 = additional time to exchange 1 byte in both directions.
        O  = time for one operation
        d  = size of the datatype

        On a MPP system it is expected that T1==T2.

   In Comparison with the binary tree algorithm that needs:

   - for r==0
     exec_time_bin_tree =  n * L1   +  buf_lng * n * (T1 + O/d)

   - for r>0
     exec_time_bin_tree = (n+1)*L1  +  buf_lng*(n+1)*(T1 + O/d)

   the new algorithm is faster if (assuming T1=T2)

    for n>2:
     for r==0:  buf_lng  >  L2 *   n   / [ (n-2) * T1 + (n-1) * O/d ]
     for r>0:   buf_lng  >  L2 * (n+1) / [ (n-1.5)*T1 + (n-0.5)*O/d ]

    for size = 5, 6, and 7:
                buf_lng  >  L2 / [0.25 * T1  +  0.58 * O/d ]

    for size = 4:
                buf_lng  >  L2 / [0.25 * T1  +  0.62 * O/d ]
    for size = 2 and 3:
                buf_lng  >  L2 / [  0.5 * O/d ]

   and for O/d >> T1 we can summarize:

    The new algorithm is faster for about   buf_lng > 2 * L2 * d / O

   Example L1 = L2 = 50 us,
           bandwidth=300MB/s, i.e. T1 = T2 = 3.3 ns/byte
           and 10 MFLOP/s,    i.e. O  = 100 ns/FLOP
           for double,        i.e. d  =   8 byte/FLOP

           ==> the new algorithm is faster for about buf_lng>8000 bytes
               i.e count > 1000 doubles !
Step 1)

   compute n and r

Step 2)

   if  myrank < 2*r

     split the buffer into ABCD and EFGH

     even myrank: send    buffer EFGH to   myrank+1 (buf_lng/2 exchange)
                  receive buffer ABCD from myrank+1
                  compute op for ABCD                 (count/2 ops)
                  receive result EFGH               (buf_lng/2 transfer)
     odd  myrank: send    buffer ABCD to   myrank+1
                  receive buffer EFGH from myrank+1
                  compute op for EFGH
                  send    result EFGH

   Result: node:     0    2    4    6    8  10  11  12
           value:  a+b  c+d  e+f  g+h  i+j   k   l   m

Step 3)

   The following algorithm uses only the nodes with

   (myrank is even  &&  myrank < 2*r) || (myrank >= 2*r)

Step 4)

   define NEWRANK(old) := (old < 2*r ? old/2 : old-r)
   define OLDRANK(new) := (new < r   ? new*2 : new+r)

   Result:
       old ranks:    0    2    4    6    8  10  11  12
       new ranks:    0    1    2    3    4   5   6   7
           value:  a+b  c+d  e+f  g+h  i+j   k   l   m

Step 5.1)

   Split the buffer (ABCDEFGH) in the middle,
   the lower half (ABCD) is computed on even (new) ranks,
   the upper half (EFGH) is computed on odd  (new) ranks.

   exchange: ABCD from 1 to 0, from 3 to 2, from 5 to 4 and from 7 to 6
             EFGH from 0 to 1, from 2 to 3, from 4 to 5 and from 6 to 7
                                               (i.e. buf_lng/2 exchange)
   compute op in each node on its half:        (i.e.   count/2 ops)

   Result: node 0: (a+b)+(c+d)   for  ABCD
           node 1: (a+b)+(c+d)   for  EFGH
           node 2: (e+f)+(g+h)   for  ABCD
           node 3: (e+f)+(g+h)   for  EFGH
           node 4: (i+j)+ k      for  ABCD
           node 5: (i+j)+ k      for  EFGH
           node 6:    l + m      for  ABCD
           node 7:    l + m      for  EFGH

Step 5.2)

   Same with double distance and oncemore the half of the buffer.
                                               (i.e. buf_lng/4 exchange)
                                               (i.e.   count/4 ops)

   Result: node 0: [(a+b)+(c+d)] + [(e+f)+(g+h)]  for  AB
           node 1: [(a+b)+(c+d)] + [(e+f)+(g+h)]  for  EF
           node 2: [(a+b)+(c+d)] + [(e+f)+(g+h)]  for  CD
           node 3: [(a+b)+(c+d)] + [(e+f)+(g+h)]  for  GH
           node 4: [(i+j)+ k   ] + [   l + m   ]  for  AB
           node 5: [(i+j)+ k   ] + [   l + m   ]  for  EF
           node 6: [(i+j)+ k   ] + [   l + m   ]  for  CD
           node 7: [(i+j)+ k   ] + [   l + m   ]  for  GH

...
Step 5.n)

   Same with double distance and oncemore the half of the buffer.
                                            (i.e. buf_lng/2**n exchange)
                                            (i.e.   count/2**n ops)

   Result:
     0: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for A
     1: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for E
     2: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for C
     3: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for G
     4: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for B
     5: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for F
     6: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for D
     7: { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] } for H


For MPI_Allreduce:
------------------

Step 6.1)

   Exchange on the last distance (2*n) the last result
                                            (i.e. buf_lng/2**n exchange)

Step 6.2)

   Same with half the distance and double of the result
                                        (i.e. buf_lng/2**(n-1) exchange)

...
Step 6.n)

   Same with distance 1 and double of the result (== half of the
   original buffer)                            (i.e. buf_lng/2 exchange)

   Result after 6.1     6.2         6.n

    on node 0:   AB    ABCD    ABCDEFGH
    on node 1:   EF    EFGH    ABCDEFGH
    on node 2:   CD    ABCD    ABCDEFGH
    on node 3:   GH    EFGH    ABCDEFGH
    on node 4:   AB    ABCD    ABCDEFGH
    on node 5:   EF    EFGH    ABCDEFGH
    on node 6:   CD    ABCD    ABCDEFGH
    on node 7:   GH    EFGH    ABCDEFGH

Step 7)

   If r > 0
     transfer the result from the even nodes with old rank < 2*r
                           to the odd  nodes with old rank < 2*r
                                                 (i.e. buf_lng transfer)
   Result:
     { [(a+b)+(c+d)] + [(e+f)+(g+h)] } + { [(i+j)+k] + [l+m] }
     for ABCDEFGH
     on all nodes 0..12


For MPI_Reduce:
---------------

Step 6.0)

   If root node not in the list of the nodes with newranks
   (see steps 3+4) then

     send last result from node 0 to the root    (buf_lng/2**n transfer)
     and replace the role of node 0 by the root node.

Step 6.1)

   Send on the last distance (2**(n-1)) the last result
   from node with bit '2**(n-1)' in the 'new rank' unequal to that of
   root's new rank  to the node with same '2**(n-1)' bit.
                                            (i.e. buf_lng/2**n transfer)

Step 6.2)

   Same with half the distance and double of the result
   and bit '2**(n-2)'
                                        (i.e. buf_lng/2**(n-1) transfer)

...
Step 6.n)

   Same with distance 1 and double of the result (== half of the
   original buffer) and bit '2**0'             (i.e. buf_lng/2 transfer)

   Example: roots old rank: 10
            roots new rank:  5

   Results:          6.1               6.2                 6.n
                action result     action result       action result

    on node 0:  send A
    on node 1:  send E
    on node 2:  send C
    on node 3:  send G
    on node 4:  recv A => AB     recv CD => ABCD   send ABCD
    on node 5:  recv E => EF     recv GH => EFGH   recv ABCD => ABCDEFGH
    on node 6:  recv C => CD     send CD
    on node 7:  recv G => GH     send GH


Buffer handling:  After each iteration, scr1 (or recvbuf) contains the result. 
----------------  scr2 is a helper buffer.

   lower half:
   ------------
   [local process ......]      [partner process ................................]
                           msg 
                     scr2 <--- latest result          (stored in sendbuf or scr1)
   latest result --> scr2      from partner process
   (locally)     op

             scr1 := scr2
   (i.e. exchange role of scr2
    and scr1 (double-buffering))

   higher half:
   ------------- 
                [partner process...]       [local process .....................]
   if commute                         msg
                latest result         ---> scr2
                from partner process       scr2 <-- latest result (sendbuf or scr1)
                                                 op (locally)
              
                Exchange role of scr2 and scr1 (double-buffering)

   if not commute                     msg           if scr1 does not contain latest
                latest result         ---> scr2     result: 
                from partner process                scr1 <--- sendbuf
                                           scr2 --> scr1
                                                op

 
Benchmark results on CRAY T3E with version 26 Sep 1997
-----------------------------

   uname -a: (sn6307 hwwt3e 1.6.1.51 unicosmk CRAY T3E)
   MPI:      /opt/ctl/mpt/1.1.0.3
   datatype: MPI_DOUBLE
   Ldb[][] = {{ 896,1728, 576, 736},{ 448,1280, 512, 512}}
   env: export MPI_BUFFER_MAX=4099 
   compiled with: cc -c -O3 -h restrict=f

   old = binary tree protocol of the vendor
   new = the new protocol and its implementation
   mixed = 'new' is used if count > limit(datatype, communicator size)

   REDUCE:
                                          communicator size
   measurement count prot. unit    2    3    4    6    8   12   16   24
   --------------------------------------------------------------------
   latency         1 mixed  us  20.7 35.1 35.6 49.1 49.2 61.8 62.4 74.8
                       old  us  19.0 32.9 33.8 47.1 47.2 61.7 62.1 73.2
   --------------------------------------------------------------------
   bandwidth    128k mixed MB/s 75.4 34.9 49.1 27.3 41.1 24.6 38.0 23.7
   (=buf_lng/time)     old MB/s 28.8 16.2 16.3 11.6 11.6  8.8  8.8  7.2
   ration = mixed/old            2.6  2.1  3.0  2.4  3.5  2.8  4.3  3.3
   --------------------------------------------------------------------
   limit                doubles  896 1536  576  736  576  736  576  736
   bandwidth   limit mixed MB/s 35.9 20.5 18.6 12.3 13.5  9.2 10.0  8.6
                       old MB/s 35.9 20.3 17.8 13.0 12.5  9.6  9.2  7.8
   ratio = mixed/old            1.00 1.01 1.04 0.95 1.08 0.96 1.09 1.10

                                          communicator size
   measurement count prot. unit   32   48   64   96  128  192  256
   ---------------------------------------------------------------
   latency         1 mixed  us  77.8 88.5 90.6  102                 1)
                       old  us  78.6 87.2 90.1 99.7  108  119  120
   ---------------------------------------------------------------
   bandwidth    128k mixed MB/s 35.1 23.3 34.1 22.8 34.4 22.4 33.9
   (=buf_lng/time)     old MB/s  6.0  6.0  6.0  5.2  5.2  4.6  4.6
   ration = mixed/old            5.8  3.9  5.7  4.4  6.6  4.8  7.4  5)
   ---------------------------------------------------------------
   limit                doubles  576  736  576  736  576  736  576  2)
   bandwidth   limit mixed MB/s  9.7  7.5  8.4  6.5  6.9  5.5  5.1  3)
                       old MB/s  7.7  6.4  6.4  5.7  5.5  4.9  4.7  3)
   ratio = mixed/old            1.26 1.17 1.31 1.14 1.26 1.12 1.08  4)

   ALLREDUCE:
                                          communicator size
   measurement count prot. unit    2    3    4    6    8   12   16   24
   --------------------------------------------------------------------
   latency         1 mixed  us  28.2 51.0 44.5 74.4 59.9  102 74.2  133
                       old  us  26.9 48.3 42.4 69.8 57.6 96.0 75.7  126
   --------------------------------------------------------------------
   bandwidth    128k mixed MB/s 74.0 29.4 42.4 23.3 35.0 20.9 32.8 19.7
   (=buf_lng/time)     old MB/s 20.9 14.4 13.2  9.7  9.6  7.3  7.4  5.8
   ration = mixed/old            3.5  2.0  3.2  2.4  3.6  2.9  4.4  3.4
   --------------------------------------------------------------------
   limit                doubles  448 1280  512  512  512  512  512  512
   bandwidth   limit mixed MB/s 26.4 15.1 16.2  8.2 12.4  7.2 10.8  5.7
                       old MB/s 26.1 14.9 15.0  9.1 10.5  6.7  7.3  5.3
   ratio = mixed/old            1.01 1.01 1.08 0.90 1.18 1.07 1.48 1.08

                                          communicator size
   measurement count prot. unit   32   48   64   96  128  192  256
   ---------------------------------------------------------------
   latency         1 mixed  us  90.3  162  109  190
                       old  us  92.7  152  104  179  122  225  135
   ---------------------------------------------------------------
   bandwidth    128k mixed MB/s 31.1 19.7 30.1 19.2 29.8 18.7 29.0
   (=buf_lng/time)     old MB/s  5.9  4.8  5.0  4.1  4.4  3.4  3.8
   ration = mixed/old            5.3  4.1  6.0  4.7  6.8  5.5  7.7
   ---------------------------------------------------------------
   limit                doubles  512  512  512  512  512  512  512
   bandwidth   limit mixed MB/s  6.6  5.6  5.7  3.5  4.4  3.2  3.8
                       old MB/s  6.3  4.2  5.4  3.6  4.4  3.1
   ratio = mixed/old            1.05 1.33 1.06 0.97 1.00 1.03

   Footnotes:
   1) This line shows that the overhead to decide which protocol
      should be used can be ignored.
   2) This line shows the limit for the count argument.
      If count < limit then the vendor protocol is used,
      otherwise the new protocol is used (see variable Ldb).
   3) These lines show the bandwidth (=bufer length / execution time)
      for both protocols.
   4) This line shows that the limit is choosen well if the ratio is
      between 0.95 (loosing 5% for buffer length near and >=limit)
      and 1.10 (not gaining 10% for buffer length near and <limit).
   5) This line shows that the new protocol is 2..7 times faster
      for long counts.

*/

int MPI_I_anyReduce_half(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm, int is_all)
{
  int commute;
  char *scr_all, *scr1buf, *scr2buf, *xxx, *sendbuf, *recvbuf, *r_buf;
  int myrank, size, x_base, x_size, computed, idx;
  int x_start, x_count, r, n, mynewrank, newroot, partner;
  int start_even[40], start_odd[40], count_even[40], count_odd[40];
  MPI_Aint typelng;
  MPI_Status status;
  size_t scrlng;
  OP_VARIABLES

  if (count == 0) return (MPI_SUCCESS);

  /* init */ 
    OP_INIT 
    OP_CHECK_COMMUTE 
 
    sendbuf = (char*) Sendbuf;
    recvbuf = (char*) Recvbuf;
    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &myrank);
    MPI_Type_extent(datatype, &typelng);
 
    scrlng  = typelng * count;
#  ifdef SCR_LNG_OPTIM
    scrlng = SCR_LNG_OPTIM(scrlng);
#  endif
    scr_all = malloc(2*scrlng); 
    scr1buf = scr_all;
    scr2buf = scr1buf + scrlng; 
 
    computed = 0;
    if (is_all) root = myrank; /* for correct recvbuf handling */

  /*...step 1 */

#   ifdef DEBUG
     printf("[%2d] step 1 begin\n",myrank); fflush(stdout);
#   endif
    n = 0; x_size = 1;
    while (2*x_size <= size) { n++; x_size = x_size * 2; }
    /* x_sixe == 2**n */
    r = size - x_size;

  /*...step 2 */

#   ifdef DEBUG
    printf("[%2d] step 2 begin n=%d, r=%d\n",myrank,n,r);fflush(stdout);
#   endif
    if (myrank < 2*r)
    {
      if ((myrank % 2) == 0 /*even*/)
      {
        MPI_I_Sendrecv(sendbuf + (count/2)*typelng,
                       count - count/2, datatype, myrank+1, 1220,
                       scr1buf, count/2,datatype, myrank+1, 1221,
                       comm, &status);
        MPI_I_CALL_OP(sendbuf, scr1buf,
                      count/2, datatype, op);
        MPI_I_Recv(scr1buf + (count/2)*typelng, count - count/2,
                   datatype, myrank+1, 1223, comm, &status);
        computed = 1;
#       ifdef DEBUG
        { int i; printf("[%2d] after step 2: val=",
                        myrank);
          for (i=0; i<count; i++)
           printf(" %5.0lf",((double*)scr1buf)[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      }
      else /*odd*/
      {
        if (commute)
        { 
          MPI_I_Sendrecv(sendbuf, count/2,datatype, myrank-1, 1221,
                         scr1buf + (count/2)*typelng,
                         count - count/2, datatype, myrank-1, 1220,
                         comm, &status);
          MPI_I_CALL_OP(sendbuf + (count/2)*typelng,
                        scr1buf + (count/2)*typelng,
                        count - count/2, datatype, op);
          MPI_I_Send(scr1buf + (count/2)*typelng, count - count/2,
                     datatype, myrank-1, 1223, comm);
        } else
        { /* not commute */ 
          MPI_I_Sendrecv(sendbuf, count/2,datatype, myrank-1, 1221,
                         scr2buf + (count/2)*typelng,
                         count - count/2, datatype, myrank-1, 1220,
                         comm, &status);
          memcpy(scr1buf + (count/2)*typelng, sendbuf + (count/2)*typelng, 
                 (count - count/2) * typelng); 
          MPI_I_CALL_OP(scr2buf + (count/2)*typelng,
                        scr1buf + (count/2)*typelng,
                        count - count/2, datatype, op);
          MPI_I_Send(scr1buf + (count/2)*typelng, count - count/2,
                     datatype, myrank-1, 1223, comm);
        } 
      }
    }

  /*...step 3+4 */

#   ifdef DEBUG
     printf("[%2d] step 3+4 begin\n",myrank); fflush(stdout);
#   endif
    if ((myrank >= 2*r) || ((myrank%2 == 0)  &&  (myrank < 2*r)))
         mynewrank = (myrank < 2*r ? myrank/2 : myrank-r);
    else mynewrank = -1;

    if (mynewrank >= 0)
    { /* begin -- only for nodes with new rank */

#     define OLDRANK(new)   ((new) < r ? (new)*2 : (new)+r)

  /*...step 5 */

      x_start = 0;
      x_count = count;
      for (idx=0, x_base=1; idx<n; idx++, x_base=x_base*2)
      {
        start_even[idx] = x_start;
        count_even[idx] = x_count / 2;
        start_odd [idx] = x_start + count_even[idx];
        count_odd [idx] = x_count - count_even[idx];
        if (((mynewrank/x_base) % 2) == 0 /*even*/)
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 5.%d begin even c=%1d\n",
                   myrank,mynewrank,idx+1,computed); fflush(stdout);
#         endif
          x_start = start_even[idx];
          x_count = count_even[idx];
          r_buf = ((root==myrank) && (idx==(n-1)) ? recvbuf : scr2buf ); 
          MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                         + start_odd[idx]*typelng, count_odd[idx],
                         datatype, OLDRANK(mynewrank+x_base), 1231,
                         r_buf + x_start*typelng, x_count,
                         datatype, OLDRANK(mynewrank+x_base), 1232,
                         comm, &status);
          MPI_I_CALL_OP((computed?scr1buf:sendbuf) + x_start*typelng,
                        r_buf + x_start*typelng, x_count, datatype, op);
          xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx;
          /* i.e., intermediate result in scr1buf, final result in r_buf */
        }
        else /*odd*/
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 5.%d begin  odd c=%1d\n",
                   myrank,mynewrank,idx+1,computed); fflush(stdout);
#         endif
          x_start = start_odd[idx];
          x_count = count_odd[idx];
          if (commute)
          { 
            r_buf = ((root==myrank) && (idx==(n-1)) ? recvbuf : scr2buf ); 
            MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                           +start_even[idx]*typelng, count_even[idx],
                           datatype, OLDRANK(mynewrank-x_base), 1232,
                           r_buf + x_start*typelng, x_count,
                           datatype, OLDRANK(mynewrank-x_base), 1231,
                           comm, &status);
            MPI_I_CALL_OP((computed?scr1buf:sendbuf) + x_start*typelng,
                          r_buf + x_start*typelng, x_count, datatype, op);
            xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx;
            /* i.e., intermediate result in scr1buf, final result in r_buf */
          } else
          { /* not commute */
            MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                           +start_even[idx]*typelng, count_even[idx],
                           datatype, OLDRANK(mynewrank-x_base), 1232,
                           scr2buf + x_start*typelng, x_count,
                           datatype, OLDRANK(mynewrank-x_base), 1231,
                           comm, &status);
            if (!computed)
            { memcpy(scr1buf+x_start*typelng, sendbuf+x_start*typelng, x_count*typelng);
            }
            MPI_I_CALL_OP(scr2buf + x_start*typelng,
                          scr1buf + x_start*typelng, x_count, datatype, op);
            r_buf = scr1buf;
            if ((root==myrank) && (idx==(n-1)))
            { r_buf = recvbuf;
              memcpy(recvbuf+x_start*typelng, scr1buf+x_start*typelng, x_count*typelng);
            }
          } 
        }
        computed = 1;
#       ifdef DEBUG
        { int i; printf("[%2d](%2d) after step 5.%d   end: start=%2d  count=%2d  val=",
                        myrank,mynewrank,idx+1,x_start,x_count);
          for (i=0; i<x_count; i++)
           printf(" %5.0lf",((double*)(r_buf + x_start*typelng))[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      } /*for*/

#     undef OLDRANK


    } /* end -- only for nodes with new rank */

    if (is_all)
    {
      /*...steps 6.1 to 6.n */

      if (mynewrank >= 0)
      { /* begin -- only for nodes with new rank */

#       define OLDRANK(new)   ((new) < r ? (new)*2 : (new)+r)

        for(idx=n-1, x_base=x_size/2; idx>=0; idx--, x_base=x_base/2)
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 6.%d begin\n",myrank,mynewrank,n-idx); fflush(stdout);
#         endif
          if (((mynewrank/x_base) % 2) == 0 /*even*/)
          {
            MPI_I_Sendrecv(recvbuf + start_even[idx]*typelng,
                                     count_even[idx],
                           datatype, OLDRANK(mynewrank+x_base), 1241,
                           recvbuf + start_odd[idx]*typelng,
                                     count_odd[idx],
                           datatype, OLDRANK(mynewrank+x_base), 1242,
                           comm, &status);
#           ifdef DEBUG
              x_start = start_odd[idx];
              x_count = count_odd[idx];
#           endif
          }
          else /*odd*/
          {
            MPI_I_Sendrecv(recvbuf + start_odd[idx]*typelng,
                                     count_odd[idx],
                           datatype, OLDRANK(mynewrank-x_base), 1242,
                           recvbuf + start_even[idx]*typelng,
                                     count_even[idx],
                           datatype, OLDRANK(mynewrank-x_base), 1241,
                           comm, &status);
#           ifdef DEBUG
              x_start = start_even[idx];
              x_count = count_even[idx];
#           endif
          }
#         ifdef DEBUG
          { int i; printf("[%2d](%2d) after step 6.%d   end: start=%2d  count=%2d  val=",
                          myrank,mynewrank,n-idx,x_start,x_count);
            for (i=0; i<x_count; i++)
             printf(" %5.0lf",((double*)(recvbuf + x_start*typelng))[i] );
            printf("\n"); fflush(stdout);
          }
#         endif
        } /*for*/

#       undef OLDRANK

      } /* end -- only for nodes with new rank */

      /*...step 7 */

      if (myrank < 2*r)
      {
#       ifdef DEBUG
          printf("[%2d] step 7 begin\n",myrank); fflush(stdout);
#       endif
        if (myrank%2 == 0 /*even*/)
          MPI_I_Send(recvbuf, count, datatype, myrank+1, 1253, comm);
        else /*odd*/
          MPI_I_Recv(recvbuf, count, datatype, myrank-1, 1253, comm, &status);
      }

    }
    else /* not is_all, i.e. Reduce */
    {

    /*...step 6.0 */

      if ((root < 2*r) && (root%2 == 1))
      {
#       ifdef DEBUG
          printf("[%2d] step 6.0 begin\n",myrank); fflush(stdout);
#       endif
        if (myrank == 0) /* then mynewrank==0, x_start==0
                                 x_count == count/x_size  */
        {
          MPI_I_Send(scr1buf,x_count,datatype,root,1241,comm);
          mynewrank = -1;
        }

        if (myrank == root)
        {
          mynewrank = 0;
          x_start = 0;
          x_count = count;
          for (idx=0, x_base=1; idx<n; idx++, x_base=x_base*2)
          {
            start_even[idx] = x_start;
            count_even[idx] = x_count / 2;
            start_odd [idx] = x_start + count_even[idx];
            count_odd [idx] = x_count - count_even[idx];
            /* code for always even in each bit of mynewrank: */
            x_start = start_even[idx];
            x_count = count_even[idx];
          }
          MPI_I_Recv(recvbuf,x_count,datatype,0,1241,comm,&status);
        }
        newroot = 0;
      }
      else
      {
        newroot = (root < 2*r ? root/2 : root-r);
      }

    /*...steps 6.1 to 6.n */

      if (mynewrank >= 0)
      { /* begin -- only for nodes with new rank */

#       define OLDRANK(new) ((new)==newroot ? root                     \
                             : ((new)<r ? (new)*2 : (new)+r) )

        for(idx=n-1, x_base=x_size/2; idx>=0; idx--, x_base=x_base/2)
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 6.%d begin\n",myrank,mynewrank,n-idx); fflush(stdout);
#         endif
          if ((mynewrank & x_base) != (newroot & x_base))
          {
            if (((mynewrank/x_base) % 2) == 0 /*even*/)
            { x_start = start_even[idx]; x_count = count_even[idx];
              partner = mynewrank+x_base; }
            else
            { x_start = start_odd[idx]; x_count = count_odd[idx];
              partner = mynewrank-x_base; }
            MPI_I_Send(scr1buf + x_start*typelng, x_count, datatype,
                       OLDRANK(partner), 1244, comm);
          }
          else /*odd*/
          {
            if (((mynewrank/x_base) % 2) == 0 /*even*/)
            { x_start = start_odd[idx]; x_count = count_odd[idx];
              partner = mynewrank+x_base; }
            else
            { x_start = start_even[idx]; x_count = count_even[idx];
              partner = mynewrank-x_base; }
            MPI_I_Recv((myrank==root ? recvbuf : scr1buf)
                       + x_start*typelng, x_count, datatype,
                       OLDRANK(partner), 1244, comm, &status);
#           ifdef DEBUG
            { int i; printf("[%2d](%2d) after step 6.%d   end: start=%2d  count=%2d  val=",
                            myrank,mynewrank,n-idx,x_start,x_count);
              for (i=0; i<x_count; i++)
               printf(" %5.0lf",((double*)(root==myrank
                                 ? recvbuf + x_start*typelng
                                 : scr1buf + x_start*typelng))[i] );
              printf("\n"); fflush(stdout);
            }
#           endif
          }
        } /*for*/

#       undef OLDRANK

      } /* end -- only for nodes with new rank */
    }

    free(scr_all);
    return(MPI_SUCCESS);
}

/* ------------------------------------------------------------------- *
 *                                                                     *
 *         2. algorithm: binary blocks based halving + doubling        *
 *                                                                     *
 * ------------------------------------------------------------------- */

/* Copyright: Rolf Rabenseifner, July 2003
 *            High-Performance Computing Center 
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 

 
/* Allreduce algorithm implemented with
 * binary blocks based halving reduce_scatter + doubling allgather 
 * ---------------------------------------------------------------
 * 
 * Copyright: Rolf Rabenseifner, July 2003
 *            Computing Center University of Stuttgart
 *            rabenseifner@rus.uni-stuttgart.de 
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 


/* Fast reduce and allreduce algorithm for longer buffers and predefined
   operations.

   Algorithm:
   ----------
 
   This algorithm is explaned with the example of 13 nodes.
   The nodes are numbered   0, 1, 2, ... 12.
   The sendbuf content is   a, b, c, ...  m.
   The buffer array is notated with ABCDEFGH, this means that
   e.g. 'C' is the third 1/8 of the buffer.

   The algorithm computes

   { [(a+b)+(c+d)] + [(e+f)+(g+h)] }  +  { [(i+j)+(k+l)] + m }

   This is equivalent to "a" binary tree algorithm.
   It is not equivalent to the parenthesizing in Rabenseifner's algorithm
   from Nov. 1997 (another halving and doubling scheme!).

    { [(a+b)+(c+d)] + [(e+f)+(g+h)] }  +  { [(i+j) + (k+l)] +    m    }
    | ------- block '2**3' ---------|       |block '2**2' |   |'2**0'|   <<-- independently
                                                                              calculated
                                          | reduce blocks '2**2'+tail |
   |          reduce blocks '2**3'+tail                                |

   Inside of each block, the same halving algorithm is described in
   myreduce.c (Rolf Rabenseifner, Nov. 1997), steps 5.1 to 5.n

   The processes are grouped into blocks.
   The vectors (send_buf, recv_buf, and all scratch buffers) are segmented
   into segments. The number of segments is identical to the number of
   processes in the first block.

   Buffer handling:  After each iteration, scr1 (or recvbuf) contains the result.
   ----------------  scr2 is a helper buffer.

   lower half: 
   ------------
   [local process ......]      [partner process ................................]
                           msg 
                     scr2 <--- latest result          (stored in sendbuf or scr1)
   latest result --> scr2      from partner process 
   (locally)     op

             scr1 := scr2
   (i.e. exchange role of scr2 
    and scr1 (double-buffering))

   higher half:
   ------------- 
                [partner process...]       [local process .....................]
   if commute                         msg
                latest result         ---> scr2  
                from partner process       scr2 <-- latest result (sendbuf or scr1)
                                                 op (locally)
             
                Exchange role of scr2 and scr1 (double-buffering)

   if not commute                     msg           if scr1 does not contain latest
                latest result         ---> scr2     result: 
                from partner process                scr1 <--- sendbuf
                                           scr2 --> scr1
                                                op  
*/

 
void MPI_I_send_to_all_lower(char *buf, int start, int count, 
              MPI_Datatype datatype, MPI_Aint typelng, 
#             ifdef DEBUG
                 int myrank, int my_block_rank, 
#             endif
              int dest_rank, 
              int process_stride, int diff_expos, MPI_Comm comm)
{ if (diff_expos > 0)
  { 
    MPI_I_send_to_all_lower( buf, start, count/2, datatype, typelng,
#             ifdef DEBUG
                 myrank, my_block_rank, 
#             endif
              dest_rank + 0, process_stride*2, diff_expos-1, comm);
    MPI_I_send_to_all_lower( buf, start+count/2, count-count/2, datatype, typelng,
#             ifdef DEBUG
                 myrank, my_block_rank, 
#             endif
              dest_rank + process_stride, process_stride*2, diff_expos-1, comm);
  } else 
  {
#      ifdef DEBUG
         printf("[%2d](%2d) sending to rank=%2d, start=%2d, count=%2d \n",  
                myrank,my_block_rank, dest_rank, 
                  start,count); fflush(stdout);
#      endif
    MPI_I_Send(buf+start*typelng, count, datatype, dest_rank, 1333, comm);
  } 
} 

int MPI_I_anyReduce_blck(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm, int is_all, int block_count)
{
  int commute;
  int myrank, size;  
  MPI_Aint typelng;
  size_t scrlng;
  char *scr_all, *scr1buf, *scr2buf, *xxx, *sendbuf, *recvbuf, *r_buf;
  int computed; /* 0 = argument from send_buf,  1 = argument from scratch buffer */
  int idx, x_base, x_start, x_count;
#define MAX_BLOCKS 64
  int my_block_num, my_block_base_rank, my_block_rank, my_block_expo, my_block_size;
  struct {   int expo; 
             int size;
             int base_rank;
         } block[MAX_BLOCKS]; 
  int start_even[MAX_BLOCKS], start_odd[MAX_BLOCKS];
  int count_even[MAX_BLOCKS], count_odd[MAX_BLOCKS];
  int newroot, partner; 
  MPI_Status status; 
  OP_VARIABLES

  if (count == 0) return (MPI_SUCCESS);
 
  MPI_Comm_size(comm, &size); 

  /* the following optimization is __not__ allowed, because 
   * it is wrong for user defined operations with non-contiguous used-defined
   * datatypes
   *  if (size == 1)
   *  { MPI_Type_extent(datatype, &typelng);
   *    memcpy(Recvbuf, Sendbuf, typelng * count);
   *    return (MPI_SUCCESS);
   *  }    
   */ 

  OP_INIT 
  OP_CHECK_COMMUTE 
 
  sendbuf = (char*) Sendbuf;
  recvbuf = (char*) Recvbuf;
  MPI_Comm_rank(comm, &myrank);
  if (is_all) root = myrank; /* for correct recvbuf handling in Allreduce */
  if (myrank != root) recvbuf = (char*) -1; /* to verfy correct recvbuf handling */
  MPI_Type_extent(datatype, &typelng);
  
  /* analysing the binary block structure of "size" */
  { 
    int i, upper_bits_of_size, block_expo, block_size, block_base_rank;
    /* .. counting bits, i.e., calculating the number of blocks*/
    if (block_count < 0)
    { 
      upper_bits_of_size = size;
      block_count = 0; 
      while (upper_bits_of_size) 
      { if (upper_bits_of_size & 1 /*check lowest bit*/) block_count++;
        upper_bits_of_size = upper_bits_of_size >> 1; 
      } 
    } 
 
#ifdef DEBUG 
    if (myrank == 0) printf("BLCK: block_count=%d\n",block_count);fflush(stdout);
#endif 
 
    /* .. calculating the block structure and store reverse */
    upper_bits_of_size = size;
    i=block_count;  block_expo=0;   block_size=1 /* := 2**block_expo */;
    block_base_rank = size; 
    while (upper_bits_of_size) 
    {
      if (upper_bits_of_size & 1 /*check lowest bit*/)
      { /* block found */ 
        i--; 
        block_base_rank = block_base_rank - block_size; 
        block[i].expo = block_expo;
        block[i].size = block_size;
        block[i].base_rank = block_base_rank; 
        if ((myrank >= block_base_rank) && (myrank < block_base_rank+block_size)) 
        { my_block_num = i;
          my_block_base_rank = block_base_rank;
          my_block_rank = myrank - my_block_base_rank;
          my_block_expo = block_expo;
          my_block_size = block_size;
        }
      }   
      block_expo++;
      block_size = block_size << 1 /* := block_size*2 */;
      upper_bits_of_size = upper_bits_of_size >> 1; 
    } 
  } 
# ifdef DEBUG
  { int i; MPI_Barrier(comm);
    for (i=0; i<size; i++) { if (i==myrank) { 
        printf("[%2d] my_block_.. num=%2d, base=%2d, rank=%2d, expo=%2d, size=%2d\n",
               myrank, my_block_num, my_block_base_rank, my_block_rank, my_block_expo,
               my_block_size);fflush(stdout);
        } MPI_Barrier(comm);
  } }
# endif
 
  scrlng  = typelng * count;
#ifdef SCR_LNG_OPTIM
  scrlng = SCR_LNG_OPTIM(scrlng);
#endif
  scr_all = malloc(2*scrlng); 
  scr1buf = scr_all;
  scr2buf = scr1buf + scrlng;

  /* reduce_scatter inside of each block */ 
  { 
 
    computed = 0; 
 
    x_start = 0;
    x_count = count;
    r_buf = sendbuf; 
    for (idx=0, x_base=1; idx<my_block_expo; idx++, x_base=x_base*2)
    {
      start_even[idx] = x_start;
      count_even[idx] = x_count / 2;
      start_odd [idx] = x_start + count_even[idx];
      count_odd [idx] = x_count - count_even[idx];
      if (((my_block_rank/x_base) % 2) == 0 /*even*/)
      { /* lower half */
#       ifdef DEBUG
          printf("[%2d](%2d) step 5.%d begin even c=%1d\n",
                 myrank,my_block_rank,idx+1,computed); fflush(stdout);
#       endif
        x_start = start_even[idx];
        x_count = count_even[idx];
        r_buf = ((block_count==1) && (idx==(my_block_expo-1)) && (myrank == root)
                  ? recvbuf : scr2buf);
        MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                       + start_odd[idx]*typelng, count_odd[idx],
                       datatype, my_block_base_rank+my_block_rank+x_base, 1331,
                       r_buf + x_start*typelng, x_count,
                       datatype, my_block_base_rank+my_block_rank+x_base, 1332,
                       comm, &status);
        MPI_I_CALL_OP((computed?scr1buf:sendbuf) + x_start*typelng, 
                      r_buf + x_start*typelng, x_count, datatype, op);
        xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx; 
        /* i.e., intermediate result in scr1buf, final result in r_buf */
      }
      else /*odd*/
      { /* higher half */
#       ifdef DEBUG
          printf("[%2d](%2d) step 5.%d begin  odd c=%1d\n",
                 myrank,my_block_rank,idx+1,computed); fflush(stdout);
#       endif
        x_start = start_odd[idx];
        x_count = count_odd[idx];
        if (commute) 
        { 
          r_buf = ((block_count==1) && (idx==(my_block_expo-1)) && (myrank == root)
                    ? recvbuf : scr2buf);
          MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                         +start_even[idx]*typelng, count_even[idx],
                         datatype, my_block_base_rank+my_block_rank-x_base, 1332,
                         r_buf + x_start*typelng, x_count,
                         datatype, my_block_base_rank+my_block_rank-x_base, 1331,
                         comm, &status);
          /* use commute law to compute r_buf := r_buf op scr1buf */ 
          MPI_I_CALL_OP((computed?scr1buf:sendbuf) + x_start*typelng, 
                        r_buf + x_start*typelng, x_count, datatype, op);
          xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx; 
          /* i.e., intermediate result in scr1buf, final result in r_buf */
        } else 
        { /* not commute */ 
          MPI_I_Sendrecv((computed ? scr1buf : sendbuf)
                         +start_even[idx]*typelng, count_even[idx],
                         datatype, my_block_base_rank+my_block_rank-x_base, 1332,
                         scr2buf + x_start*typelng, x_count,
                         datatype, my_block_base_rank+my_block_rank-x_base, 1331,
                         comm, &status);
          if (!computed)
          { memcpy(scr1buf+x_start*typelng, sendbuf+x_start*typelng, x_count*typelng);
          } 
          MPI_I_CALL_OP(scr2buf + x_start*typelng, 
                        scr1buf + x_start*typelng, x_count, datatype, op);
          r_buf = scr1buf;
          if ((block_count==1) && (idx==(my_block_expo-1)) && (myrank == root))
          { r_buf = recvbuf;
            memcpy(recvbuf+x_start*typelng, scr1buf+x_start*typelng, x_count*typelng);
          } 
        } 
      }
      computed = 1;
#     ifdef DEBUG
      { int i; printf("[%2d](%2d) after step 5.%d   end: start=%2d  count=%2d  val=",
                      myrank,my_block_rank,idx+1,x_start,x_count);
        for (i=0; i<x_count; i++)
         printf(" %5.0lf",((double*)(r_buf + x_start*typelng))[i] );
        printf("\n"); fflush(stdout);
      }
#     endif
    } /*for*/
  } 
  /* end of reduce_scatter inside of each block */ 
 
# ifdef DEBUG
  { int i; MPI_Barrier(comm);
    for (i=0; i<size; i++) { if (i==myrank) { 
      { int i; printf("[%2d](%2d) before block exchange: start=%2d  count=%2d  val=",
                      myrank,my_block_rank,x_start,x_count);
        for (i=0; i<x_count; i++)
         printf(" %5.0lf",((double*)(r_buf + x_start*typelng))[i] );
        printf("\n"); fflush(stdout);
      }
        } MPI_Barrier(comm);
  } }
# endif
  
  /* receive from higher block and reduce */ 
  if (my_block_num < block_count-1) 
  { /* lower half */ 
#    ifdef DEBUG
       printf("[%2d](%2d) receiving from rank=%2d, x_start=%2d, x_count=%2d \n",  
              myrank,my_block_rank, block[my_block_num+1].base_rank 
                + my_block_rank % block[my_block_num+1].size, x_start,x_count); fflush(stdout);
#    endif
     r_buf = ((my_block_num == 0) && (myrank == root) ? recvbuf : scr2buf);
     MPI_I_Recv(r_buf + x_start*typelng, x_count, datatype, 
                block[my_block_num+1].base_rank 
                 + my_block_rank % block[my_block_num+1].size, 
                1333, comm, &status);
    
     MPI_I_CALL_OP(scr1buf + x_start*typelng,
                   r_buf + x_start*typelng, x_count, datatype, op);
     xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx; 
     /* i.e., intermediate result in scr1buf, final result in r_buf */
  } 
  /* end of receive from higher block and reduce */    
 
  /* send to lower block */
  if (my_block_num > 0) 
  {
    MPI_I_send_to_all_lower( (computed?scr1buf:sendbuf), x_start, x_count,
            datatype, typelng, 
#           ifdef DEBUG
               myrank, my_block_rank, 
#           endif
            /* destination rank: */
                block[my_block_num-1].base_rank + my_block_rank,
            /* process' stride: */ 
                my_block_size,
            /* difference of the expos: */ 
                block[my_block_num-1].expo - my_block_expo,
            comm);

    /* ---- defined above (begin) ----------------------- 
 
     * void MPI_I_send_to_all_lower(char *buf, int start, int count, 
     *               MPI_Datatype datatype, MPI_Aint typelng, 
     *               int dest_rank, 
     *               int process_stride, int diff_expos, MPI_Comm comm)
     * { if (diff_expos > 0)
     *   { 
     *     MPI_I_send_to_all_lower( buf, start, count/2, datatype, typelng,
     *               dest_rank + 0, process_stride*2, diff_expos-1, comm);
     *     MPI_I_send_to_all_lower( buf, start+count/2, count-count/2, datatype, typelng,
     *               dest_rank + process_stride, process_stride*2, diff_expos-1, comm);
     *   } else 
     *   {
     *     MPI_I_Send(buf+start*typelng, count, datatype, dest_rank, 1333, comm);
     *   } 
     * } 
     * ------- defined above (end) ------------------------ */ 
  } 
  /* end of send to lower block */
 
# ifdef DEBUG
  { int i; MPI_Barrier(comm);
    for (i=0; i<size; i++) { if (i==myrank) { 
      { int i; printf("[%2d](%2d) after block exchange: start=%2d  count=%2d  %s=",
                      myrank,my_block_rank,x_start,x_count,
                      (r_buf==sendbuf ? "sendbuf" : 
                       (r_buf==recvbuf ? "recvbuf" :
                        (r_buf==scr1buf ? "scr1buf" : "UNKNOWN BUFFER"))) );
        for (i=0; i<x_count; i++)
         printf(" %5.0lf",((double*)(r_buf + x_start*typelng))[i] );
        printf("\n"); fflush(stdout);
      }
        } MPI_Barrier(comm);
  } }
# endif

  if (is_all)
  {

    /* inside of each block: allgather with doubling algorithm */ 
    { 
      x_base = my_block_size/2; 
      if (my_block_num > 0)
      { /* receive from lower block with two messages */
#       ifdef DEBUG
          printf("[%2d](%2d) receive from %2d, start=%2d, count=%2d\n",
                 myrank,my_block_rank,myrank - 2 * my_block_size,
                 x_start, x_count/2); fflush(stdout);
#       endif
        MPI_I_Recv(recvbuf+ x_start*typelng, x_count/2, datatype, 
                   myrank - 2 * my_block_size, 1341, comm, &status);
#       ifdef DEBUG
          printf("[%2d](%2d) receive from %2d, start=%2d, count=%2d\n",
                 myrank,my_block_rank,myrank - my_block_size,
                 x_start+x_count/2, x_count-x_count/2); fflush(stdout);
#       endif
        MPI_I_Recv(recvbuf+(x_start+x_count/2)*typelng, x_count-x_count/2, datatype,
                   myrank - my_block_size, 1341, comm, &status);
      } 

      for(idx=my_block_expo-1; idx>=0; idx--, x_base=x_base/2)
      {
#       ifdef DEBUG
          printf("[%2d](%2d) step 6.%d begin\n",myrank,my_block_rank,my_block_expo-idx); fflush(stdout);
#       endif

        if (( my_block_num < (block_count-1) ) 
            && ( (my_block_expo-idx) == (my_block_expo-block[my_block_num+1].expo) )
            && ( (my_block_size-my_block_rank) <= (2*block[my_block_num+1].size) )
           )
        { /* send to higher block from two processes */
          if (((my_block_rank/x_base) % 2) == 0 /*even*/)
               { x_start = start_even[idx]; x_count = count_even[idx]; }
          else { x_start = start_odd[idx];  x_count = count_odd[idx];  }
#         ifdef DEBUG
            printf("[%2d](%2d) send to %2d, start=%2d, count=%2d\n",
                   myrank,my_block_rank,block[my_block_num+1].base_rank
                    + my_block_rank % block[my_block_num+1].size,
                   x_start, x_count); fflush(stdout);
#         endif
          MPI_I_Send(recvbuf+ x_start*typelng, x_count, datatype, 
                     block[my_block_num+1].base_rank 
                      + my_block_rank % block[my_block_num+1].size,  1341, comm);
        }
 
        if (((my_block_rank/x_base) % 2) == 0 /*even*/)
        {
          MPI_I_Sendrecv(recvbuf + start_even[idx]*typelng,
                                   count_even[idx],
                         datatype, my_block_base_rank+my_block_rank+x_base, 1342,
                         recvbuf + start_odd[idx]*typelng,
                                   count_odd[idx],
                         datatype, my_block_base_rank+my_block_rank+x_base, 1343,
                         comm, &status);
#         ifdef DEBUG
            x_start = start_odd[idx];
            x_count = count_odd[idx];
#         endif
        }
        else /*odd*/
        {
          MPI_I_Sendrecv(recvbuf + start_odd[idx]*typelng,
                                   count_odd[idx],
                         datatype, my_block_base_rank+my_block_rank-x_base, 1343,
                         recvbuf + start_even[idx]*typelng,
                                   count_even[idx],
                         datatype, my_block_base_rank+my_block_rank-x_base, 1342,
                         comm, &status);
#         ifdef DEBUG
            x_start = start_even[idx];
            x_count = count_even[idx];
#         endif
        }
#       ifdef DEBUG
        { int i; printf("[%2d](%2d) after step 6.%d   end: start=%2d  count=%2d  val=",
                        myrank,my_block_rank,my_block_expo-idx,x_start,x_count);
          for (i=0; i<x_count; i++)
           printf(" %5.0lf",((double*)(recvbuf + x_start*typelng))[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      } /*for*/
 
    }
    /* end of inside of each block: allgather with doubling algorithm */ 
 
#   ifdef DEBUG
    { int i; MPI_Barrier(comm);
      for (i=0; i<size; i++) { if (i==myrank) { 
        { int i; printf("[%2d](%2d) after allgather: val=",
                        myrank,my_block_rank);
          for (i=0; i<count; i++) printf(" %5.0lf",((double*)(recvbuf))[i] );
          printf("\n"); fflush(stdout);
        }
          } MPI_Barrier(comm);
    } }
#   endif

  }
  else /* not is_all, i.e. Reduce */
  { 
  /*...step 6.0 */

    my_block_rank = (myrank >= block[0].size ? -1 : myrank);
    my_block_expo = block[0].expo; 
    my_block_size = block[0].size; 
 
    if (root >= block[0].size)
    {
#     ifdef DEBUG
        printf("[%2d] step 6.0 begin\n",myrank); fflush(stdout);
#     endif
      if (myrank == 0) /* then my_block_rank==0, x_start==0
                               x_count == count/my_block_size */
      {
        MPI_I_Send(scr1buf,x_count,datatype,root,1340,comm);
        my_block_rank = -1;
      }

      if (myrank == root)
      {
        my_block_rank = 0;
        x_start = 0;
        x_count = count;
        for (idx=0, x_base=1; idx<my_block_expo; idx++, x_base=x_base*2)
        {
          start_even[idx] = x_start;
          count_even[idx] = x_count / 2;
          start_odd [idx] = x_start + count_even[idx];
          count_odd [idx] = x_count - count_even[idx];
          /* code for always even in each bit of my_block_rank: */
          x_start = start_even[idx];
          x_count = count_even[idx];
        }
        MPI_I_Recv(recvbuf,x_count,datatype,0,1340,comm,&status);
      }
      newroot = 0;
    }
    else
    {
      newroot = root;
    }

  /*...steps 6.1 to 6.n */

    if (my_block_rank >= 0)
    { /* begin -- only for nodes with new rank */

#     define OLDRANK(new) ((new)==newroot ? root : (new))

      for(idx=my_block_expo-1, x_base=my_block_size/2; idx>=0; idx--, x_base=x_base/2)
      {
#       ifdef DEBUG
          printf("[%2d](%2d) step 6.%d begin\n",myrank,my_block_rank,my_block_expo-idx); fflush(stdout);
#       endif
        if ((my_block_rank & x_base) != (newroot & x_base))
        {
          if (((my_block_rank/x_base) % 2) == 0 /*even*/)
          { x_start = start_even[idx]; x_count = count_even[idx];
            partner = my_block_rank+x_base; }
          else
          { x_start = start_odd[idx]; x_count = count_odd[idx];
            partner = my_block_rank-x_base; }
          MPI_I_Send(scr1buf + x_start*typelng, x_count, datatype,
                     OLDRANK(partner), 1344, comm);
        }
        else /*odd*/
        {
          if (((my_block_rank/x_base) % 2) == 0 /*even*/)
          { x_start = start_odd[idx]; x_count = count_odd[idx];
            partner = my_block_rank+x_base; }
          else
          { x_start = start_even[idx]; x_count = count_even[idx];
            partner = my_block_rank-x_base; }
          MPI_I_Recv((myrank==root ? recvbuf : scr1buf)
                     + x_start*typelng, x_count, datatype,
                     OLDRANK(partner), 1344, comm, &status);
#         ifdef DEBUG
          { int i; printf("[%2d](%2d) after step 6.%d   end: start=%2d  count=%2d  val=",
                          myrank,my_block_rank,my_block_expo-idx,x_start,x_count);
            for (i=0; i<x_count; i++)
             printf(" %5.0lf",((double*)(root==myrank
                               ? recvbuf + x_start*typelng
                               : scr1buf + x_start*typelng))[i] );
            printf("\n"); fflush(stdout);
          }
#         endif
        }
      } /*for*/

#     undef OLDRANK

    } /* end -- only for nodes with new rank */
  } 

  free(scr_all); 
  return(MPI_SUCCESS);
}

/* ------------------------------------------------------------------- *
 *                                                                     *
 *                3. algorithm: pairwise + ring/pttopt                 *
 *                                                                     *
 * ------------------------------------------------------------------- */

/* Copyright: Rolf Rabenseifner, July 2003
 *            High-Performance Computing Center 
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 

/* Allreduce algorithm implemented with                     
 * pairwise reduce_scatter algorithm + ring-based allgather
 * --------------------------------------------------------
 *
 * Copyright: Rolf Rabenseifner, July 2003
 *            Computing Center University of Stuttgart
 *            rabenseifner@rus.uni-stuttgart.de 
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 


/* Fast allreduce algorithm for longer buffers and predefined
   operations.

   This algorithm is explaned with the example of 6 nodes.
   The nodes are numbered   0, 1, 2, ... 5.
   The sendbuf content is   a, b, c, ... f.
   The buffer array is notated with ABCDEF, this means that
   e.g. 'C' is the third 1/6 of the buffer.

   The algorithm computes

     (((((a+b)+c)+d)+e)+f)   for buffer segment A
     a + ((((b+c)+d)+e)+f)   for buffer segment B
     (a+b) + (((c+d)+e)+f)   for buffer segment C
     ((a+b)+c) + ((d+e)+f)   for buffer segment D
     (((a+b)+c)+d) + (e+f)   for buffer segment E
     ((((a+b)+c)+d)+e) + f   for buffer segment F 
     

   This is __not__ equivalent to the mostly used binary tree algorithm
   (e.g. used in mpich), which would compute
 
     [(a+b)+(c+d)] + [e+f]  for all buffer segments

   size := number of nodes in the communicator.

   The allreduce algorithm is done in two steps:
    - first a reduce_scatter implemented with pairwise exchange in 
      size-1 iterations: In the i-th step (i = 1 .. size-1),
      Each process receives from (myrank+i)%size the data needed to 
      compute the reduction on its own segment
    - with an allgather, all processes receive all computed segments:
      the segments a send in a ring through all processes 
 
*/


int MPI_I_anyReduce_pair(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm, int is_all)
{
  int myrank, size, num_full_segments, i, source, dest, send_datarank, recv_datarank;
  int commute; 
  int my_seg_count, send_cnt, recv_cnt;
  MPI_Aint data_lng, seg_lng, lastseg_lng, typelng; 
  char *scr_all, *sendbuf, *recvbuf, *scr1buf, *scr2buf, *scr3buf, *xxx;
  char *left_vec, *rslt_vec, *accu_vec;
  MPI_Status status;
  size_t scrlng;
  OP_VARIABLES
 
  if (count == 0) return (MPI_SUCCESS);
 
  MPI_Comm_size(comm, &size);

  /* the following optimization is __not__ allowed, because
   * it is wrong for user defined operations with non-contiguous used-defined
   * datatypes
   *  if (size == 1)
   *  { MPI_Type_extent(datatype, &typelng);
   *    data_lng = typelng * count;
   *    memcpy(Recvbuf, Sendbuf, data_lng);
   *    return (MPI_SUCCESS);
   *  } 
   */ 
 
  MPI_Type_extent(datatype, &typelng);
  seg_lng = ((count - 1)/size + 1) * typelng; 
 
  OP_INIT 
  OP_CHECK_COMMUTE 
 
  sendbuf = (char*) Sendbuf;
  recvbuf = (char*) Recvbuf;
  MPI_Comm_rank(comm, &myrank);
  if (is_all) root = myrank; /* for correct recvbuf handling in Allreduce */ 
  data_lng = typelng * count;
  num_full_segments = data_lng / seg_lng;
  lastseg_lng = data_lng - num_full_segments*seg_lng;  
 
# define SEGMENT_SIZE(rank) \
        ((rank)<num_full_segments ? seg_lng : ((rank)==num_full_segments ? lastseg_lng : 0) )
# define SEGMENT_BEGIN(rank) \
        ((rank)<=num_full_segments ? (rank) * seg_lng : 0) 
# define RANK(cnt,rank) \
        ((cnt)>0 ? (rank) : MPI_PROC_NULL) 
 
  my_seg_count = SEGMENT_SIZE(myrank) / typelng; 
  recv_cnt = my_seg_count;
  if (recv_cnt == 0) recv_cnt = 1; /* zero recv_cnt is not allowed, although  */
 
# ifdef DEBUG
   printf("MPI_(All)Reduce: myrank=%1d, data_lng=%1d, num_full_segments=%1d, lastseg_lng=%1d, SEGMENT_SIZE(myrank)=%1d, SEGMENT_BEGIN(myrank)=%1d\n",
                            myrank,     data_lng,     num_full_segments,     lastseg_lng,     SEGMENT_SIZE(myrank),     SEGMENT_BEGIN(myrank));

   { double *snd; int j; 
     printf("MPI_(All)Reduce: myrank=%1d, send_buf=",myrank);
     snd = (double *) Sendbuf;
#    if (DEBUG > 2) 
      for (j=0; j<count; j++) printf(" %5.0f",snd[j]);
#    endif 
     printf("\n"); 
   }
# endif
  
  scrlng  = seg_lng;
#ifdef SCR_LNG_OPTIM
  scrlng = SCR_LNG_OPTIM(scrlng);
#endif    
  scr_all = malloc(3*scrlng);
  scr1buf = scr_all;
  scr2buf = scr1buf + scrlng;
  scr3buf = scr2buf + scrlng;
   
  /* ----------------------------------------------
     Step 1: reduce_scatter with pairwise algorithm 
     
     Pseudo-code:

       i=1 
       neighbor myrank+i --> scr3buf
       sendbuf+SEGMENT_BEGIN(myrank)
       + scr3buf         --> scr3buf       (accumulating 2nd part of result)
       scr2buf := scr3buf                  (pointer exchange, double buffering)
       
       i++
       neighbor myrank+i --> scr3buf
       scr2buf
       + scr3buf         --> scr3buf       (accumulating 2nd part of result)
       scr2buf := scr3buf                  (pointer exchange, double buffering)
 
       ...
       
       i++
       neighbor size-1   --> scr3buf (or recvbuf, if myrank==root)
       scr2buf
       + scr3/recv       --> scr3/recv     (accumulating 2nd part of result)
       scr2buf := scr3buf                  (pointer exchange, double buffering)

       i++ 
       neighbor 0        --> scr3buf       (accumulating 1st part of result)
       scr1buf := scr3buf                  (pointer exchange, double buffering)
       
       i++ 
       neighbor 1        --> scr3buf
       scr1buf
       + scr3buf         --> scr3buf       (accumulating 1st part of result)
       scr1buf := scr3buf                  (pointer exchange, double buffering)
 
       ...

       i++  (i.e. i == size-1)
       neighbor myrank-1 --> scr3buf 
       scr1buf
       + scr3buf         --> scr3buf       (accumulating 1st part of result)
       scr1buf := scr3buf                  (pointer exchange, double buffering)

       (after the loop) 
       scr1buf + recvbuf --> recvbuf       (if myrank == root)
       scr1buf + scr2buf --> scr2buf       (if myrank != root)


       Exceptions: 
        - if myrank == 0 
            ==> scr1buf will not be used
                i.e. if i==size-1, the first part "accumulating to scr2buf" 
                                   is still in use, i.e.,
                                   "accumulating 1st part" is not done, and
                                   "after the loop" is not done.
        - if myrank == size-1  
            ==> scr2buf will not be used, i.e., "accumulating 2nd part" is not done,
                and "i==size-1" and "after loop" is substituted by
                  a) if commute:
                       i++  (i.e. i == size-1)
                       neighbor myrank-1 --> recvbuf 
                       scr1buf
                       + scr3buf         --> recvbuf
                       sendbuf + recvbuf --> recvbuf (using commute law!) 
                  b) if not commute 
                       i++  (i.e. i == size-1)
                       neighbor myrank-1 --> scr3buf
                       scr1buf
                       + scr3buf         --> src3buf
                       recvbuf := sendbuf (memcpy) 
                       scr3buf + recvbuf --> recvbuf
  */ 
 
  for (i=1; i<size; i++)
  {
    source = (myrank + i) % size; 
    dest   = (myrank - i + size) % size; 
    rslt_vec = (source < myrank ? scr1buf : scr2buf); 
    left_vec = (i==1 ? sendbuf+SEGMENT_BEGIN(myrank) : rslt_vec);
 
    /* recv from source into scr3buf */
    send_cnt = SEGMENT_SIZE(dest)/typelng; 
#   ifdef DEBUG
     printf("MPI_(All)Reduce: myrank=%1d, i=%1d, dest=%1d, sendsize=%1d, source=%1d, recvsize=%1d\n",
                              myrank, i, RANK(send_cnt,dest), send_cnt, RANK(my_seg_count,source), my_seg_count);
#   endif
    accu_vec = ((myrank==root) && 
                (   (source==(size-1)) 
                 || ( (myrank==(size-1)) && commute && (i==(size-1)) )
                ) ? recvbuf+SEGMENT_BEGIN(myrank) : scr3buf); 
    MPI_I_Sendrecv(sendbuf+SEGMENT_BEGIN(dest), 
                     send_cnt ,datatype, RANK(send_cnt,dest), 1220,
                   accu_vec,
                     recv_cnt, datatype, RANK(my_seg_count,source), 1220,
                   comm, &status);
    if (source != 0)
    { /* reduction operation   scr3buf := left_vec op scr3buf */
      MPI_I_CALL_OP(left_vec, accu_vec, my_seg_count, datatype, op);
    }
    xxx = rslt_vec; /* helper for exchange of pointers */
    if (rslt_vec==scr1buf) scr1buf = scr3buf; 
    else                   scr2buf = scr3buf;
    scr3buf = xxx;
#   ifdef DEBUG
#     define BUF_TEST(buf)  (buf==(recvbuf+SEGMENT_BEGIN(myrank))?"recv" :""),\
        (buf==scr1buf?"scr1":""), (buf==scr2buf?"scr2":""), (buf==scr3buf?"scr3":"")
      printf("MPI_(All)Reduce: myrank=%1d i=%1d after loop accu_vec=%s%s%s%s rslt=%s%s%s%s scr1=%s%s%s%s scr2=%s%s%s%s scr3=%s%s%s%s\n",
        myrank, i, BUF_TEST(accu_vec), BUF_TEST(rslt_vec), 
        BUF_TEST(scr1buf), BUF_TEST(scr2buf), BUF_TEST(scr3buf) );
#   endif
  } 
  /* "after the loop" */ 
  if (myrank == 0)
  { 
    rslt_vec = accu_vec;
  } 
  else if (myrank == (size-1))
  {
    if (commute)
    { rslt_vec = accu_vec;
      MPI_I_CALL_OP(sendbuf+SEGMENT_BEGIN(myrank), rslt_vec, my_seg_count, datatype, op);
    } else /* not commute */
    { rslt_vec = (myrank==root ? recvbuf+SEGMENT_BEGIN(myrank) : scr2buf);
#     ifdef DEBUG
        printf("MPI_(All)Reduce: [%2d] memcpy\n", myrank ); fflush(stdout);
#     endif
      memcpy(rslt_vec, sendbuf+SEGMENT_BEGIN(myrank), my_seg_count*typelng); 
      MPI_I_CALL_OP(scr1buf, rslt_vec, my_seg_count, datatype, op);
    }
  } 
  else /* (myrank != 0 or size-1) */
  {
    rslt_vec = (myrank==root ? recvbuf+SEGMENT_BEGIN(myrank) : scr2buf); 
    /* reduction operation   rslt_vec := left_vec op rslt_vec */
    MPI_I_CALL_OP(scr1buf, rslt_vec, my_seg_count, datatype, op);
  } 
 
# ifdef DEBUG
   { double *rcv; int j; 
     printf("MPI_(All)Reduce: myrank=%1d, %s%s%s%s=",myrank,BUF_TEST(rslt_vec));
     rcv = (double *) rslt_vec;
#    if (DEBUG > 2) 
      for (j=0; j<my_seg_count; j++) printf(" %5.0f",rcv[j]);
#    endif 
     printf("\n"); 
   }
# endif
 
  if (is_all)
  { /*Allreduce*/ 
 
    /* ----------------------------------------------
       Step 2: all_gather with ring sending
    */ 
 
    dest     = (myrank + 1) % size; 
    source   = (myrank - 1 + size) % size; 
    for (i=0; i<size-1; i++)
    {
      send_datarank = (myrank - i     + size) % size;
      recv_datarank = (myrank - i - 1 + size) % size;
      send_cnt = SEGMENT_SIZE(send_datarank) / typelng; 
      recv_cnt = SEGMENT_SIZE(recv_datarank) / typelng; 
#     ifdef DEBUG
       printf("MPI_(All)Reduce: myrank=%1d, dest=%1d, sendsize=%1d, source=%1d, recvsize=%1d\n",
                                myrank, RANK(send_cnt,dest), send_cnt, RANK(recv_cnt,source), recv_cnt);
#     endif
      MPI_I_Sendrecv(recvbuf+SEGMENT_BEGIN(send_datarank), 
                       send_cnt ,datatype, RANK(send_cnt,dest), 1221,
                     recvbuf+SEGMENT_BEGIN(recv_datarank),
                       (recv_cnt==0 ? 1 : recv_cnt), datatype, RANK(recv_cnt,source), 1221,
                     comm, &status);
    } 
  } else 
  { /*Reduce*/ 
    if (myrank == root)
    {
      for (i=0; i<size; i++) if (i != root)
      { 
        recv_cnt = SEGMENT_SIZE(i) / typelng;
        MPI_I_Recv(recvbuf+SEGMENT_BEGIN(i), (recv_cnt==0 ? 1 : recv_cnt), 
                   datatype, RANK(recv_cnt,i), 1222, comm, &status);
      }  
    } else 
    {
      MPI_I_Send(rslt_vec, my_seg_count, datatype, RANK(my_seg_count,root), 1222, comm);
    } 
  } 
 
# ifdef DEBUG
    if (is_all || (myrank==root)) 
    { double *rcv; int j; 
      printf("MPI_(All)Reduce: myrank=%1d, recv_buf=",myrank);
      rcv = (double *) Recvbuf;
#     if (DEBUG > 2) 
       for (j=0; j<count; j++) printf(" %5.0f",rcv[j]);
#     endif 
      printf("\n"); 
    }
# endif
   
  free(scr_all); 
  return(MPI_SUCCESS);
}


/* ------------------------------------------------------------------- *
 *                                                                     *
 *                 4. algorithm: binary tree (+ bcast)                 *
 *                                                                     *
 * ------------------------------------------------------------------- */
 
/* Copyright: Rolf Rabenseifner, July 2003
 *            High-Performance Computing Center 
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
 *
 * The usage of this software is free, 
 * but this header must not be removed.
 */ 

 
int MPI_I_anyReduce_tree(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm, int is_all)
{
  int commute;
  int myrank, size, dest, src;  
  int mask, inv, old_root; 
  MPI_Aint typelng;
  size_t scrlng;
  char *scr_all, *scr1buf, *scr2buf, *sendbuf, *recvbuf, *r_buf;
  int computed; /* 0 = argument from send_buf,  1 = argument from scratch buffer */
  MPI_Status status; 
  OP_VARIABLES

  if (count == 0) return (MPI_SUCCESS);
 
  MPI_Comm_size(comm, &size); 

  /* the following optimization is __not__ allowed, because 
   * it is wrong for user defined operations with non-contiguous used-defined
   * datatypes
   *  if (size == 1)
   *  { MPI_Type_extent(datatype, &typelng);
   *    memcpy(Recvbuf, Sendbuf, typelng * count);
   *    return (MPI_SUCCESS);
   *  }    
   */ 
 
  OP_INIT 
  OP_CHECK_COMMUTE 
 
  sendbuf = (char*) Sendbuf;
  recvbuf = (char*) Recvbuf;
  MPI_Comm_rank(comm, &myrank);
  if (is_all) root = 0; /* for correct recvbuf handling in Allreduce */
  if ((!is_all) && (myrank != root)) recvbuf = (char*) -1; /* to verfy correct recvbuf handling */
  MPI_Type_extent(datatype, &typelng);
 
# ifdef DEBUG
  { int i; MPI_Barrier(comm);
    for (i=0; i<size; i++) { if (i==myrank) {int j; 
        printf("[%2d] binary tree begin, is_all=%d, root=%d, count=%d, sbuf=",
               myrank, is_all, root, count);
        for (j=0; j<count; j++)
         printf(" %5.0lf",((double*)(sendbuf))[j] );
        printf("\n"); fflush(stdout);
        } MPI_Barrier(comm);
  } }
# endif
  
  scrlng  = typelng * count;
#ifdef SCR_LNG_OPTIM
  scrlng = SCR_LNG_OPTIM(scrlng);
#endif    
  scr_all = malloc(2*scrlng);
  scr1buf = scr_all;
  scr2buf = scr1buf + scrlng;

  /* Binary tree based on root */
  computed = 0; 
  mask = 1;
  old_root = root;
  while (mask < size)
  {
    old_root = old_root & (~mask);
    inv = (root & mask) && ((myrank == root) || (myrank == old_root));
    if (!inv) 
    { /* usual binary tree */ 
      if (myrank & mask)
      { /* sender on this level of the bin tree */
        dest = myrank - mask;
        if (dest == old_root) dest = root;
#       ifdef DEBUG
          printf("[%2d] bin.tree norm send mask=%d, dest=%2d\n", myrank, mask, dest); fflush(stdout);
#       endif
        MPI_I_Send( (computed ? scr1buf : sendbuf), count, datatype, dest, 1401, comm);
        /* this process has done its work, therefore */ mask=size;  
      } else
      { /* receiver and executing the operation on this level of the bin tree */
        r_buf = ( (mask << 1) >= size ? /* last iteration */ recvbuf : scr2buf);
        src = (myrank==root ? old_root+mask : myrank+mask); 
        if ( src < size )
        {
#         ifdef DEBUG
            printf("[%2d] bin.tree norm recv mask=%d, from=%2d\n", myrank, mask, src); fflush(stdout);
#         endif
          MPI_I_Recv( r_buf, count, datatype, src, 1401, comm, &status);
          MPI_I_CALL_OP((computed ? scr1buf : sendbuf), r_buf, count, datatype, op);
          computed = 1;
          scr2buf = scr1buf;  /* scratch in the next iteration */
          scr1buf = r_buf;    /* current result */
        } 
      } 
    } else
    { /* root and old_root have to change their roles */ 
      if (myrank == old_root) 
      { /* sender on this level of the bin tree */
#       ifdef DEBUG
          printf("[%2d] bin.tree inv. send mask=%d, dest=%2d\n", myrank, mask, root); fflush(stdout);
#       endif
        MPI_I_Send( (computed ? scr1buf : sendbuf), count, datatype, root, 1401, comm);
        /* this process has done its work, therefore */ mask=size;  
      } else
      { /* receiver and executing the operation on this level of the bin tree */
        if (commute)
        {
          r_buf = ( (mask << 1) >= size ? /* last iteration */ recvbuf : scr2buf);
#         ifdef DEBUG
            printf("[%2d] bin.tree inv. recv mask=%d, from=%2d\n", myrank, mask, old_root); fflush(stdout);
#         endif
          MPI_I_Recv( r_buf, count, datatype, old_root, 1401, comm, &status);
          /* todo: r_buf := r_buf + scr1buf;  done via commute law: */
          MPI_I_CALL_OP((computed ? scr1buf : sendbuf), r_buf, count, datatype, op);
          computed = 1;
          scr2buf = scr1buf;  /* scratch in the next iteration */
          scr1buf = r_buf;    /* current result */
        } else
        { /* not commute */
#         ifdef DEBUG
            printf("[%2d] bin.tree i.r.^commute m=%d, from=%2d\n", myrank, mask, old_root); fflush(stdout);
#         endif
          MPI_I_Recv( scr2buf, count, datatype, old_root, 1401, comm, &status);
          r_buf = scr1buf;
          if (!computed) 
          { if ((mask << 1) >= size) r_buf = recvbuf; 
#           ifdef DEBUG
              printf("[%2d] bin.tree %s := sendbuf, mask=%d\n", myrank, (r_buf==recvbuf?"recvbuf":"scr1buf"), mask); fflush(stdout);
#           endif
            memcpy(r_buf, sendbuf, count*typelng);
            computed = 1;
          }
          MPI_I_CALL_OP(scr2buf, r_buf, count, datatype, op);
        } 
      } 
    }
    mask = mask << 1;
  } 
  
# ifdef DEBUG
    if ((myrank==root) && (r_buf != recvbuf))
       printf("[%2d] bin.tree inv. recvbuf:=memcpy\n", myrank); fflush(stdout);
# endif
  if ((myrank==root) && (r_buf != recvbuf)) memcpy(recvbuf, r_buf, count*typelng);
 
# ifdef DEBUG
  { MPI_Barrier(comm);
    if (myrank==root) {int j; 
        printf("[%2d] binary tree end of reduce, is_all=%d, root=%d, count=%d, rbuf=",
               myrank, is_all, root, count);
        for (j=0; j<count; j++)
         printf(" %5.0lf",((double*)(recvbuf))[j] );
        printf("\n"); fflush(stdout);
    } MPI_Barrier(comm);
  } 
# endif
      
  if (is_all) 
  { 
#   ifdef USE_BCAST  
      MPI_I_Bcast(recvbuf, count, datatype, root, comm);
#   else
      /* this algorithm requires that root==0 */ 
      mask = 1;
      while (mask < size)
      { 
        if (myrank & mask)
        { /* level reached on which this process receives its data */
          src = myrank - mask; 
          MPI_I_Recv( recvbuf, count, datatype, src, 1402, comm, &status);
          break; 
        } 
        mask <<= 1; 
      } 
 
      mask >>= 1; 
 
      while (mask > 0)
      {
        if ( (myrank + mask ) < size )
        { 
          MPI_I_Send( recvbuf, count, datatype, myrank+mask, 1402, comm); 
        } 
        mask >>= 1; 
      } 
#   endif 
  } 

  free(scr_all); 
  return(MPI_SUCCESS);
}
/* ------------------------------------------------------------------- *
 *                                                                     *
 *                5. algorithm: recursive doubling                     *
 *                                                                     *
 * ------------------------------------------------------------------- */

/* Copyright: Rolf Rabenseifner, September 1997 and July 2003
 *            High-Performance Computing Center
 *            University of Stuttgart
 *            www.hlrs.de/people/rabenseifner
 *            rabenseifner@hlrs.de
*
 * The usage of this software is free,
 * but this header must not be removed.
 */
int MPI_I_allReduce_rdbl(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{ 
  int commute;
  char *scr_all, *scr1buf, *scr2buf, *xxx, *sendbuf, *recvbuf, *r_buf;
  int myrank, size, x_base, x_size, computed, idx;
  int r, n, mynewrank;
  MPI_Aint typelng;
  MPI_Status status;
  size_t scrlng;
  OP_VARIABLES

  if (count == 0) return (MPI_SUCCESS);

  /* init */
    OP_INIT
    OP_CHECK_COMMUTE
    sendbuf = (char*) Sendbuf;
    recvbuf = (char*) Recvbuf;
    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &myrank);
    MPI_Type_extent(datatype, &typelng);
    /* printf("[%d] Sendbuf= %d\n",myrank,Sendbuf);fflush(stdout); */
    /* printf("[%d] RECVbuf= %d\n",myrank,Recvbuf);fflush(stdout); */
    scrlng  = typelng * count;
#   ifdef SCR_LNG_OPTIM
      scrlng = SCR_LNG_OPTIM(scrlng);
#   endif
    scr_all = malloc(2*scrlng);
    scr1buf = scr_all;
    scr2buf = scr1buf + scrlng;

    computed = 0;

  /*...step 1 */

#   ifdef DEBUG
     printf("[%2d] step 1 begin\n",myrank); fflush(stdout);
#   endif
    n = 0; x_size = 1;
    while (2*x_size <= size) { n++; x_size = x_size * 2; }
    /* x_sixe == 2**n */
    r = size - x_size;

  /*...step 2 */
#   ifdef DEBUG
      printf("[%2d] step 2 begin n=%d, r=%d\n",myrank,n,r);fflush(stdout);
#   endif
    if (myrank < 2*r)
    {
      if ((myrank % 2) == 0 /*even*/)
      {
        MPI_I_Recv(scr1buf, count, datatype, myrank+1, 1223, comm, &status);
        /* printf("step 2(even_recv): [%d] scr1buf= %d\n",myrank,scr1buf); */
        MPI_I_CALL_OP(sendbuf, scr1buf, count, datatype, op);
        /* printf("step 2(even_op):   [%d] scr1buf= %d, sendbuf= %d\n",myrank,scr1buf,sendbuf); */
        computed = 1;

#       ifdef DEBUG
        { int i; printf("[%2d] after step 2: val=",
                        myrank);
          for (i=0; i<count; i++)
           printf(" %5.0lf",((double*)scr1buf)[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      }
      else /*odd*/
      {
        if (commute)
        {
          MPI_I_Send(sendbuf, count, datatype, myrank-1, 1223, comm);
          /* printf("step 2(odd):       [%d] scr1buf= %d\n",myrank,scr1buf); */

        } else
        { /* not commute */
          MPI_I_Send(sendbuf,count,datatype, myrank-1, 1223, comm);
          /* printf("step 2(odd):     [%d] sendbuf= %d, scr1buf= %d, scr2buf= %d\n",myrank,sendbuf,scr1buf,scr2buf); */
          scr2buf=scr1buf;
      
        }
      }      /* else odd*/
    }        /* if (myrank < 2*r)*/

  /*...step 3+4 */

#   ifdef DEBUG
     printf("[%2d] step 3+4 begin\n",myrank); fflush(stdout);
#   endif

    if ((myrank >= 2*r) || ((myrank%2 == 0)  &&  (myrank < 2*r)))
         mynewrank = (myrank < 2*r ? myrank/2 : myrank-r);
    else mynewrank = -1;
    
    if (mynewrank >= 0)
    { /* begin -- only for nodes with new rank */
#     define OLDRANK(new)   ((new) < r ? (new)*2 : (new)+r)

/****************************************************************************************************/
/*...step 5 */

      for (idx=0, x_base=1; idx<n; idx++, x_base=x_base*2)
      {
        if (((mynewrank/x_base) % 2) == 0 /*even*/)
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 5.%d begin even c=%1d\n",
                   myrank,mynewrank,idx+1,computed); fflush(stdout);
#         endif
          r_buf = (idx==(n-1) ? recvbuf : scr2buf );
          MPI_I_Sendrecv((computed ? scr1buf : sendbuf), count,
                         datatype, OLDRANK(mynewrank+x_base), 1231,
                         r_buf, count,
                         datatype, OLDRANK(mynewrank+x_base), 1232,
                         comm, &status);

          MPI_I_CALL_OP((computed?scr1buf:sendbuf), r_buf, count, datatype, op);

          xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx;
          /* i.e., intermediate result in scr1buf, final result in r_buf */
        }
        else /*odd*/
        {
#         ifdef DEBUG
            printf("[%2d](%2d) step 5.%d begin  odd c=%1d\n",
                   myrank,mynewrank,idx+1,computed); fflush(stdout);
#         endif
          if (commute)
          {
            r_buf = (idx==(n-1) ? recvbuf : scr2buf );
            MPI_I_Sendrecv((computed ? scr1buf : sendbuf),
                           count,
                           datatype, OLDRANK(mynewrank-x_base), 1232,
                           r_buf, 
                           count,
                           datatype, OLDRANK(mynewrank-x_base), 1231,
                           comm, &status);
            MPI_I_CALL_OP((computed?scr1buf:sendbuf), r_buf, count, datatype, op);
            xxx = scr1buf; scr1buf = scr2buf; scr2buf = xxx;
            /* i.e., intermediate result in scr1buf, final result in r_buf */
          } else
          { /* not commute */
            MPI_I_Sendrecv((computed ? scr1buf : sendbuf), 
                           count,
                           datatype, OLDRANK(mynewrank-x_base), 1232,
                           scr2buf, 
                           count,
                           datatype, OLDRANK(mynewrank-x_base), 1231,
                           comm, &status); 
            if (!computed)
            { 
              memcpy(scr1buf, sendbuf, count*typelng); 
            }
            MPI_I_CALL_OP(scr2buf, scr1buf, count, datatype, op); 

            r_buf = scr1buf;
            if (idx==(n-1))
            { r_buf = recvbuf;
              memcpy(recvbuf, scr1buf, count*typelng); 
            }
          }
        }
        computed = 1;
#       ifdef DEBUG
        { int i; printf("[%2d](%2d) after step 5.%d   end: count=%2d  val=",
                        myrank,mynewrank,idx+1,count);
          for (i=0; i<count; i++)
           printf(" %5.0lf", ((double*)(r_buf))[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      } /*for*/
#     undef OLDRANK
    } /* end -- only for nodes with new rank */
/***********************************************************************************************/
/* ...step 6 */
#   ifdef DEBUG
       printf("[%2d] step 6 begin n=%d, r=%d\n",myrank,n,r);fflush(stdout);
#   endif
    if (myrank < 2*r)
    {
      if ((myrank % 2) == 0 /*even*/)
      {
        MPI_I_Send(recvbuf, count, datatype, myrank+1, 1243, comm);    /*anstatt recvbuf -> scr1buf*/
        /* printf("step 6:            [%d] scr1buf= %d\n",myrank,scr1buf); */

#       ifdef DEBUG
        { int i; printf("[%2d] after step 6: val=",
                        myrank);
          for (i=0; i<count; i++)
           printf(" %5.0lf",((double*)scr1buf)[i] );
          printf("\n"); fflush(stdout);
        }
#       endif
      }
      else /*odd*/
      {
        MPI_I_Recv(recvbuf, count, datatype, myrank-1, 1243, comm, &status);
        /* printf("step 6:            [%d] recvbuf= %d\n",myrank,recvbuf); */
      }      /* else odd*/
    }        /* if (myrank < 2*r)*/  
/*********************************************************************************************/
    free(scr_all);
    return(MPI_SUCCESS);
}

/* ------------------------------------------------------------------- */

#endif /* !(defined(REDUCE_PROTOCOL) || defined(ALLREDUCE_PROTOCOL)) */
 
/* ------------------------------------------------------------------- *
 *                                                                     *
 *                Decision - Which protocol is the best                *
 *                                                                     *
 * ------------------------------------------------------------------- */

#define PROT_UNDEF -1
#define DFLT  0 
#define HALF  1 
#define BLCK  2 
#define PAIR  3 
#define TREE  4 
#define RDBL  5

#define pairwise_allowed 1  /* allow the pairwis algorithm */ 

/* -------------------------- Allreduce ------------------------------ */
/*                            ---------                                */

#ifndef REDUCE_PROTOCOL 
 
#ifdef ALLREDUCE_PROTOCOL 
 int MPI_MYallreduce_prot(int size, MPI_Aint lng, MPI_Datatype datatype, MPI_Op op)
#else 
 int MPI_MYallreduce(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
#endif 
{ int op_type, prot; 
#ifndef ALLREDUCE_PROTOCOL 
  int size;
  MPI_Aint lng;
#endif 
  MPI_Aint typelng;
  int block_count, max_delta_expo;

  MPI_Type_extent(datatype, &typelng);
#ifndef ALLREDUCE_PROTOCOL 
  lng = count * typelng; 
  MPI_Comm_size(comm, &size);
#endif 
  op_type = MPI_I_get_op_type(op);
 
  block_count = -1; /* to prohibit, that the number of blocks is calculated twice */
  prot=PROT_UNDEF;

# ifdef PROT_DEFINED
#   undef PROT_DEFINED
# endif 

# ifdef USE_HALF
#   define PROT_DEFINED 
    if (size>1) prot=HALF; 
# endif

# ifdef USE_BLCK
#   define PROT_DEFINED 
    if (size>1) prot=BLCK; 
# endif

# ifdef USE_PAIR
#   define PROT_DEFINED 
    if (size>1) prot=PAIR; 
# endif

# ifdef USE_TREE
#   define PROT_DEFINED 
    if (size>1) prot=TREE; 
# endif

# ifdef USE_RDBL
#   define PROT_DEFINED
    if (size>1) prot=RDBL;
# endif
 
/* ------------------------------------------------------------------- */

/* Here, platform dependent decisions can be included
 * Rules: - it must test PROT_DEFINED to prevent overwriting values from above
 *        - it must set PROT_DEFINED to prevent default handling 
 * ... */    

 
/* ------------------------------------------------------------------- */

/* default handling: */ 
 
# ifndef PROT_DEFINED 
#   define PROT_DEFINED 
    /* the following default decision was developed on a Cray T3E 900 */ 
    if (op_type < MPI_I_OP_TYPE_IMPL) prot=DFLT;
    else
      if (lng <= (size<=32 ? 2*1024 : 1*1024) )  
        if ((op==MPI_MAXLOC) || (op==MPI_MINLOC)) prot=TREE;
        else 
        { register int upper_bits;
          upper_bits = size; 
          while ( ! (upper_bits & 1) ) upper_bits = upper_bits >> 1; 
          if (upper_bits == 1) prot=DFLT;
          else prot=TREE;
        } 
      else 
        if (size <= 32) /* look at out_t3e_myallreduce_bytes.eps, 2-32 CPUs */ 
          switch(size) {
            case 1:
              prot=DFLT; 
              break; 
            case 3: case 5:
              if (pairwise_allowed) prot=PAIR; else prot=BLCK;
              break; 
            case 7:
              if (pairwise_allowed && lng >= 8*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 9: case 17:
              if (lng < 8*1024) prot=HALF; 
              else if (pairwise_allowed) prot=PAIR; else prot=BLCK;
              break; 
            case 10: case 11:
              if (pairwise_allowed && lng >=  32*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 13: case 15:
              if (pairwise_allowed && lng >=  64*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 19: case 21: case 23: case 25:
              if (pairwise_allowed && lng >= 128*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 26: case 27: case 29: case 31:
              if (pairwise_allowed && lng >= 512*1024) prot=PAIR; else prot=BLCK;
              break; 
            default:
              prot=BLCK;
              break; 
          } /*switch*/
      
        else /*(size > 32)*/   /* see out_t3e_myallreduce_cpus_0-130.eps */
                               /* and out_t3e_myallreduce_cpus_120-260.eps */
          if (lng < 16*1024) prot=HALF;
          else 
          { int upper_bits_of_size, block_expo, last_block_expo;
            /* .. counting bits, i.e., calculating the number of blocks*/
            upper_bits_of_size = size;
            block_count = 0; block_expo = 0; max_delta_expo = 0;
            while (upper_bits_of_size) 
            {
              if (upper_bits_of_size & 1 /*check lowest bit*/) 
              { if (block_count > 0)
                { if (max_delta_expo < (block_expo - last_block_expo))
                       max_delta_expo = block_expo - last_block_expo;
                }
                last_block_expo = block_expo; 
                block_count++;
              } 
              upper_bits_of_size = upper_bits_of_size >> 1; 
              block_expo++; 
            } 
            if (max_delta_expo >= (log(lng)/log(2) / 2. - 2.5) ) prot=HALF;
            else prot=BLCK;
          } 
        /*endif size*/
      /*endif lng*/
    /*endif op_type*/
# endif /*not PROT_DEFINED*/ 
 
/* ------------------------------------------------------------------- */
 
/* Here, decisions base on run-time options can be included */ 
/* Rules: - they can overwrite the compile-time decisions   */
/* ... */ 
 
/* ------------------------------------------------------------------- */

#ifdef ALLREDUCE_PROTOCOL
  return(prot); 
#else  
  switch(prot) {
    case HALF:  return( MPI_I_anyReduce_half(Sendbuf, Recvbuf, count, datatype, op, -1, comm, 1) );
    case BLCK:  return( MPI_I_anyReduce_blck(Sendbuf, Recvbuf, count, datatype, op, -1, comm, 1, block_count) );
    case PAIR:  return( MPI_I_anyReduce_pair(Sendbuf, Recvbuf, count, datatype, op, -1, comm, 1) );
    case TREE:  return( MPI_I_anyReduce_tree(Sendbuf, Recvbuf, count, datatype, op, -1, comm, 1) );
    case RDBL:  return( MPI_I_allReduce_rdbl(Sendbuf, Recvbuf, count, datatype, op, comm) );
    default:    return( MPI_Allreduce(Sendbuf, Recvbuf, count, datatype, op, comm) );
  } /*switch*/
#endif 
}

#endif /*ifndef REDUCE_PROTOCOL*/ 

/* ---------------------------- Reduce ------------------------------- */
/*                              ------                                 */
 
#ifndef ALLREDUCE_PROTOCOL 
 
#ifdef REDUCE_PROTOCOL 
 int MPI_MYreduce_prot(int size, MPI_Aint lng, MPI_Datatype datatype, MPI_Op op)
#else 
 int MPI_MYreduce(void* Sendbuf, void* Recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
#endif 
{ int op_type, prot; 
#ifndef REDUCE_PROTOCOL 
  int size;
  MPI_Aint lng;
#endif 
  MPI_Aint typelng;
  int block_count, max_delta_expo;

  MPI_Type_extent(datatype, &typelng);
#ifndef REDUCE_PROTOCOL 
  lng = count * typelng; 
  MPI_Comm_size(comm, &size);
#endif 
  op_type = MPI_I_get_op_type(op);
 
  block_count = -1; /* to prohibit, that the number of blocks is calculated twice */
  prot=PROT_UNDEF;

# ifdef PROT_DEFINED
#   undef PROT_DEFINED
# endif 

# ifdef USE_HALF
#   define PROT_DEFINED 
    if (size>1) prot=HALF; 
# endif

# ifdef USE_BLCK
#   define PROT_DEFINED 
    if (size>1) prot=BLCK; 
# endif

# ifdef USE_PAIR
#   define PROT_DEFINED 
    if (size>1) prot=PAIR; 
# endif

# ifdef USE_TREE
#   define PROT_DEFINED 
    if (size>1) prot=TREE; 
# endif
 
/* ------------------------------------------------------------------- */

/* Here, platform dependent decisions can be included 
 * Rules: - it must test PROT_DEFINED to prevent overwriting values from above
 *        - it must set PROT_DEFINED to prevent default handling 
 * ... */    

 
/* ------------------------------------------------------------------- */

/* default handling: */ 
 
# ifndef PROT_DEFINED 
#   define PROT_DEFINED 
    /* the following default decision was developed on a Cray T3E 900 */ 
    if (op_type < MPI_I_OP_TYPE_IMPL) prot=DFLT;
    else
      if (lng <= (size<=32 ? 4*1024 : 2*1024) ) 
        if (size==1) prot=DFLT;
        else prot=TREE; 
      else 
        if (size <= 32) /* look at out_t3e_myallreduce_bytes.eps, 2-32 CPUs */ 
          switch(size) {
            case 1:
              prot=DFLT; 
              break; 
            case 2:
              if (lng < 8*1024) prot=TREE; else prot=BLCK; 
              break; 
            case 3: case 5:
              if (pairwise_allowed) prot=PAIR; else prot=BLCK;
              break; 
            case 9:
              if (lng < 8*1024) prot=HALF; 
              else if (pairwise_allowed) prot=PAIR; else prot=BLCK;
              break; 
            case 17:
              if (lng < 16*1024) prot=HALF; 
              else prot=BLCK;
              break; 
            case 7:
              if (pairwise_allowed && lng >=  256*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 10: case 11: case 13:
              if (pairwise_allowed && lng >=  512*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 15: case 23:
              if (pairwise_allowed && lng >= 2048*1024) prot=PAIR; else prot=BLCK;
              break; 
            case 25:
              if (pairwise_allowed && lng >= 8192*1024) prot=PAIR; else prot=BLCK;
              break; 
            default:
              prot=BLCK;
              break; 
          } /*switch*/
      
        else /*(size > 32)*/   /* see out_t3e_myallreduce_cpus_0-130.eps */
                               /* and out_t3e_myallreduce_cpus_120-260.eps */
          if (lng < 16*1024) prot=HALF;
          else 
          { int upper_bits_of_size, block_expo, last_block_expo;
            /* .. counting bits, i.e., calculating the number of blocks*/
            upper_bits_of_size = size;
            block_count = 0; block_expo = 0; max_delta_expo = 0;
            while (upper_bits_of_size) 
            {
              if (upper_bits_of_size & 1 /*check lowest bit*/) 
              { if (block_count > 0)
                { if (max_delta_expo < (block_expo - last_block_expo))
                       max_delta_expo = block_expo - last_block_expo;
                }
                last_block_expo = block_expo; 
                block_count++;
              } 
              upper_bits_of_size = upper_bits_of_size >> 1; 
              block_expo++; 
            } 
            if (max_delta_expo >= (log(lng)/log(2) / 2. - 2.5) ) prot=HALF;
            else prot=BLCK;
          } 
        /*endif size*/
      /*endif lng*/
    /*endif op_type*/
# endif /*not PROT_DEFINED*/ 
 
/* ------------------------------------------------------------------- */
 
/* Here, decisions base on run-time options can be included */ 
/* Rules: - they can overwrite the compile-time decisions   */ 
/* ... */ 
 
/* ------------------------------------------------------------------- */

#ifdef REDUCE_PROTOCOL
  return(prot); 
#else  
  switch(prot) {
    case HALF:  return( MPI_I_anyReduce_half(Sendbuf, Recvbuf, count, datatype, op, root, comm, 0) );
    case BLCK:  return( MPI_I_anyReduce_blck(Sendbuf, Recvbuf, count, datatype, op, root, comm, 0, block_count) );
    case PAIR:  return( MPI_I_anyReduce_pair(Sendbuf, Recvbuf, count, datatype, op, root, comm, 0) );
    case TREE:  return( MPI_I_anyReduce_tree(Sendbuf, Recvbuf, count, datatype, op, root, comm, 0) );
    default:    return( MPI_Reduce(Sendbuf, Recvbuf, count, datatype, op, root, comm) );
  } /*switch*/
#endif
} /* int MPI_MYallreduce() */

#endif /*ifndef ALLREDUCE_PROTOCOL*/ 
 
/* ------------------------------------------------------------------- *
 *                                                                     *
 *                Decision - printing main program                     *
 *                                                                     *
 * ------------------------------------------------------------------- */

#if defined(ALLREDUCE_PROTOCOL) || defined(REDUCE_PROTOCOL)

int main ( int argc, char *argv[] )
{
  int is_all, myrank, expo, size, prot;
  char c; 
  MPI_Aint lng; 
  MPI_Datatype datatype; 
  MPI_Op op; 
 
  MPI_Init ( &argc, &argv ); 
  MPI_Comm_rank ( MPI_COMM_WORLD, &myrank);
  if (myrank == 0)
  {
#   ifdef ALLREDUCE_PROTOCOL
      is_all = 1;
#   else
      is_all = 0; 
#   endif 
    printf("%s\n", (is_all ? "MPI_Allreduce" : "MPI_Reduce") ); 
    printf("op = MPI_SUM\n"); op = MPI_SUM;
    printf("datatype = MPI_DOUBLE\n"); datatype = MPI_DOUBLE; 
    printf("HALF: o   BLCK: *   PAIR: p   TREE: ^   RDBL: r   DFLT: X\n");
    printf("\nbuffer size = 2**expo\n");
    printf("    expo=");
    for (expo = 3; expo <= 23; expo++) printf(" %2d",expo); printf("\n");
    for (size = 1; size <= 257; size++) 
    {
      printf("%4d PEs:", size);
      for (expo = 3, lng=8; expo <= 23; expo++, lng=lng*2) 
      { 
#       ifdef ALLREDUCE_PROTOCOL
          prot = MPI_MYallreduce_prot(size, lng, datatype, op);
#       else
          prot = MPI_MYreduce_prot(size, lng, datatype, op);
#       endif 
        switch(prot) {
          case HALF:  c='o'; break;
          case BLCK:  c='*'; break;
          case PAIR:  c='-'; break;
          case TREE:  c='^'; break;
          case RDBL:  c='r'; break;
          default:    c='X'; break;
        } /*switch*/
        printf("  %c",c);
      } 
      printf("\n");
    }
  }
  MPI_Finalize ();
  return ( 0 ); 
} 
 
#endif /*defined(ALLREDUCE_PROTOCOL) || defined(REDUCE_PROTOCOL)*/
